/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	26-Jul-2005
 */

#include "usrbin.h"

#include <sys/stat.h>

#include "math/eps.h"
#include "tool/fortran.h"

#define TOLERANCE 1e-7

const char* Usrbin::_typeStr[] = {
		"XYZ_point",
		"RPhiZ_point",
		"region_point",
		"|X|_point",
		"|Y|_point",
		"|Z|_point",
		"|XYZ|_point",
		"RPhi|Z|_point",
		"special_point",
		"",
		"XYZ",
		"RPhiZ",
		"region",
		"|X|",
		"|Y|",
		"|Z|",
		"|XYZ|",
		"RPhi|Z|",
		"special",
		""
	};

// ================================== Usrbin ===================================
/** cleanup */
void Usrbin::cleanup()
{
	filename.clear();
	if (data) delete [] data;
	data      = NULL;
	_detector = 0;
	nx = ny = nz = 0;
	x0 = y0  = 0.0;
	xofs = yofs = zofs = 0.0;
	runtitle[0] = 0;
	runtime[0]  = 0;
	name[0]     = 0;
	_norm       = 1.0;
	_logscale   = false;
	_checker    = false;
	_mtime      = 0;
	_hasMatrix  = false;

	lntzer = 0;
	bkusbn = 0.0;
	b2usbn = 0.0;
	tcusbn = 0.0;

	weipri = 0.0;
	ncase  = 0;
	mcase  = 0;
	nbatch = 0;
//	_total = 0.0;
} // cleanup

/** matrix - set transformation matrix */
void Usrbin::matrix(const Matrix4& M)
{
	if (!_hasMatrix) {
		// First time set the matrix
		_hasMatrix = true;
		_matrix.copy(M);
	} else {
		// Multiple with the existing matrix
		// _matrix = M * _matrix
		Matrix4 N;
		N.multiply(M, _matrix);
		_matrix.copy(N);
	}
	_matrix.fix();
} // matrix

/** load
 * @return	true on success
 *		false on failure
 */
bool Usrbin::load(const char* fn, int det)
{
	char buffer[512];
	int  i, length;
	int  totSize = 0;
	struct stat st;

	// Check if information is valid
	if (stat(fn, &st)) {
		cleanup();
		return false;
	}
	if (!filename.compare(fn) &&
	     st.st_mtime==_mtime && det==_detector) return true;

	cleanup();

	FortranFile f(fn,"rb");
	if (!f) return false;

	filename = fn;

	// Remember modification time
	_mtime = st.st_mtime;

	FortranParser parser(buffer, sizeof(buffer));
	length = f.read(parser);
	parser.read(runtitle, USRBIN_RUNTIT);
	parser.read(runtime,  USRBIN_RUNTIM);
	weipri = parser.readFloat();
	ncase  = parser.readInt();
	if (parser.more()) mcase  = parser.readInt();
	if (parser.more()) nbatch = parser.readInt();

	for (i=0; i<det; i++) {
		/* Read detector information */
		length = f.read(parser);
		if (length!=86) break;
		if (length<0) goto ERROR;

		parser.skipInt();
		parser.read(name, USRBIN_TITUSB);
		_type  = (Type)parser.readInt();
		_score = parser.readInt();

		xlow  = (double)parser.readFloat();
		xhigh = (double)parser.readFloat();
		nx    =         parser.readInt();
		dx    = (double)parser.readFloat();

		ylow  = (double)parser.readFloat();
		yhigh = (double)parser.readFloat();
		ny    =         parser.readInt();
		dy    = (double)parser.readFloat();

		zlow  = (double)parser.readFloat();
		zhigh = (double)parser.readFloat();
		nz    =         parser.readInt();
		dz    = (double)parser.readFloat();

		if (_type!=Type::region_point  && _type!=Type::region &&
		    _type!=Type::special_point && _type!=Type::special) {
			// Re-normalize dx,dy,dz to get more precision from float
			dx  = (xhigh-xlow) / (double)nx;
			dy  = (yhigh-ylow) / (double)ny;
			dz  = (zhigh-zlow) / (double)nz;
		}

		lntzer = parser.readInt();
		bkusbn = parser.readFloat();
		b2usbn = parser.readFloat();
		tcusbn = parser.readFloat();

		nynx = ny*nx;
		totSize = nynx*nz;

		/* skip detector */
		if (i<det-1) f.seek(8+totSize*sizeof(float), SEEK_CUR);
	}

	if (det==i && totSize>0) {
		data   = new float[totSize];
		length = totSize*sizeof(float);
		if (f.read(data, length) != length) goto ERROR;

		scanMinMax();

		printf("Usrbin: %d %s\n",det, name);
		printf("\tType:%d [%s]\tScore:%d\n",(int)_type,_typeStr[(int)_type],_score);
		printf("\tX: [%g .. %g] x %d (%g)\n",xlow, xhigh, nx, dx);
		printf("\tY: [%g .. %g] x %d (%g)\n",ylow, yhigh, ny, dy);
		printf("\tZ: [%g .. %g] x %d (%g)\n",zlow, zhigh, nz, dz);
		printf("\tMin=%-10.5g\t\tMax=%-10.5g\n",min,max);
		printf("\tLogMin=%-10.5g\tLogMax=%-10.5g\n",log10(min),log10(max));
	}
	_detector = det;
	return true;

ERROR:
	fprintf(stderr,"ERROR reading usrbin file \"%s\"\n",filename.c_str());
	cleanup();
	return true;
} // load

/** save usrbin data to file
 * @param	filename
 * @return	true on success
 *		false on failure
 */
bool Usrbin::save(const char* fn)
{
	char buffer[512];
	if (fn) filename = fn;

	FortranFile f(filename.c_str(), "wb");
	if (!f) return false;

	FortranParser parser(buffer, sizeof(buffer));

	// Write title
	parser.write(runtitle, USRBIN_RUNTIT);
	parser.write(runtime,  USRBIN_RUNTIM);
	parser.write(weipri);
	parser.write(ncase);
	parser.write(mcase);
	parser.write(nbatch);
	f.write(parser);

	// Write detector header
	parser.reset();
	parser.write(_detector);		// detector number
	parser.write(name, USRBIN_TITUSB);	// usrbin detector title
	parser.write((int)_type);		// type
	parser.write(_score);			// score

	parser.write((float)xlow);
	parser.write((float)xhigh);
	parser.write(nx);
	parser.write((float)dx);

	parser.write((float)ylow);
	parser.write((float)yhigh);
	parser.write(ny);
	parser.write((float)dy);

	parser.write((float)zlow);
	parser.write((float)zhigh);
	parser.write(nz);
	parser.write((float)dz);

	parser.write(lntzer);
	parser.write(bkusbn);
	parser.write(b2usbn);
	parser.write(tcusbn);
	assert(parser.length() == 86);
	f.write(parser);

	// Write data
	int len = nynx*nz*sizeof(float);
	if (f.write(data, len) != len)
		fprintf(stderr,"ERROR writing usrbin file \"%s\"\n",filename.c_str());

	return true;
} // save

/** create a new usrbin and initialize matrix to size ax*ay*az
 */
void Usrbin::create(int det, Type atype, int ascore,
			const double axlow, const double axhigh, const int anx,
			const double aylow, const double ayhigh, const int any,
			const double azlow, const double azhigh, const int anz)
{
	cleanup();

	_detector = det;
	_type  = atype;
	_score = ascore;

	nx    = Max(1,anx);
	ny    = Max(1,any);
	nz    = Max(1,anz);
	nynx  = nx*ny;
	data  = new float[nynx*nz];

	xlow  = axlow;
	ylow  = aylow;
	zlow  = azlow;

	xhigh = axhigh;
	yhigh = ayhigh;
	zhigh = azhigh;

	dx  = (xhigh-xlow) / (double)nx;
	dy  = (yhigh-ylow) / (double)ny;
	dz  = (zhigh-zlow) / (double)nz;

	assert(xlow < xhigh);
	assert(ylow < yhigh);
	assert(zlow < zhigh);
} // create

/** checker */
void Usrbin::checker(const Type atype,
			double axlow, double axhigh, const int anx,
			double aylow, double ayhigh, const int any,
			double azlow, double azhigh, const int anz)
{
	cleanup();

	_checker = true;

	if (axlow > axhigh) Swap(axlow, axhigh);

	_type = atype;
	xlow  = axlow;
	xhigh = axhigh;
	nx    = Max(1,anx);
	ny    = Max(1,any);

	dx = (xhigh - xlow)/(double)nx;

	if (_type==Type::RPhiZ_point    || _type==Type::RPhiZ ||
	    _type==Type::RPhiZsym_point || _type==Type::RPhiZsym) {
		// For R-Z-Phi use -PI to PI as limits
		ylow  = -PI;
		yhigh =  PI;
		x0    = aylow;
		y0    = ayhigh;
	} else {
		if (aylow > ayhigh) Swap(aylow, ayhigh);
		ylow  = aylow;
		yhigh = ayhigh;
		x0    = 0.0;
		y0    = 0.0;
	}
	dy = (yhigh - ylow)/(double)ny;

	if (azlow > azhigh) Swap(azlow, azhigh);
	zlow  = azlow;
	zhigh = azhigh;
	nz    = Max(1,anz);
	dz    = (zhigh - zlow)/(double)nz;

#if _DEBUG>1
	printf("Usrbin::setChecker it=%d\n",(int)_type);
	printf("\tx:[%g..%g]x%d\n",xlow,xhigh,nx);
	printf("\ty:[%g..%g]x%d\n",ylow,yhigh,ny);
	printf("\tz:[%g..%g]x%d\n",zlow,zhigh,nz);
#endif

	if (data) {
		delete [] data;
		data = NULL;
	}
} // setChecker

/** offset */
void Usrbin::offset(const Point& ofs)
{
	xofs = ofs.x;
	yofs = ofs.y;
	zofs = ofs.z;
} // offset

/** scanMinMax */
void Usrbin::scanMinMax()
{
	/* scan for min and maximum */
	min =  1.0e30;
	max = -1.0e30;
	for (int j=0; j<nynx*nz; j++) {
		if (data[j]<min) min = data[j];
		if (data[j]>max) max = data[j];
	}
} // scanMinMax

/** norm */
void Usrbin::norm(const double n)
{
	if (_logscale)
		_norm = log10(n);
	else
		_norm = n;
} // norm

/** normalize usrbin by multiplying by n */
void Usrbin::normalize(double n)
{
	assert(!_checker && data);

	int totSize = nynx*nz;
	if (_logscale) {
		n = log10(n);
		for (int i=0; i<totSize; i++)
			data[i] += (float)n;
	} else
		for (int i=0; i<totSize; i++)
			data[i] *= (float)n;
} // normalize

/** convert */
bool Usrbin::convert(bool tolog)
{
	if (tolog == _logscale) return false;
	if (data==NULL) return true;
	bool err = false;
	int totSize = nynx*nz;
	for (int i=0; i<totSize; i++)
		if (tolog) {
			if (data[i]>0.0)
				data[i] = log10(data[i]);
			else {
				data[i] = -1E+30f;
				err = true;
			}
		} else
			data[i] = exp10(data[i]);
	if (tolog)
		_norm = log10(_norm);
	else
		_norm = exp10(_norm);
	_logscale = tolog;
	return err;
} // convert

/** convert x,y,z coordinates to i,j,k in usrbin
 * @param x,y,z	location to return value
 * @return true if inside USRBIN, false otherwise
 */
inline bool Usrbin::xyz2ijk(double x, double y, double z, int* i, int* j, int* k) const
{
	double	r, phi;

	if (!_checker && !data) return false;

	if (hasMatrix()) matrix().transform(&x,&y,&z);
	x -= xofs;
	y -= yofs;
	z -= zofs;

	switch (_type) {
		case Type::XYZ_point:		// Cartesian no symmetry
		case Type::XYZ:
			break;

		case Type::RPhiZ_point:	// R-Phi-Z
		case Type::RPhiZ:
			x  -= x0;
			y  -= y0;
			r   = hypot(x,y);
			phi = atan2(y,x);
			if (phi < ylow)		// Check if in another period it fits
				phi += PI2;
			else
			if (phi > yhigh)
				phi -= PI2;
			x   = r;
			y   = phi;
			break;

		case Type::Xsym_point:		// X-symmetry
		case Type::Xsym:
			if (x<0.0) x = -x;
			break;

		case Type::Ysym_point:		// Y-symmetry
		case Type::Ysym:
			if (y<0.0) y = -y;
			break;

		case Type::Zsym_point:		// Z-symmetry
		case Type::Zsym:
			if (z<0.0) z = -z;
			break;

		case Type::XYZsym_point:	// X,Y,Z-symmetry
			if (x<0.0) x = -x;
			if (y<0.0) y = -y;
			if (z<0.0) z = -z;
			break;

		case Type::RPhiZsym_point:	// R-Phi-Z, Z-symmetry
		case Type::RPhiZsym:
			x  -= x0;
			y  -= y0;
			r   = hypot(x,y);
			phi = atan2(y,x);
			if (phi < ylow)		// Check if in another period it fits
				phi += PI2;
			else
			if (phi > yhigh)
				phi -= PI2;
			if (z<0.0) z = -z;
			x   = r;
			y   = phi;
			break;

		default:	// special (ignored)
			return 0.0;
	}

	// be more flexible on the limits
	if (Abs(x-xlow)<Abs(x)*TOLERANCE)
		*i = 0;
	else
	if (Abs(x-xhigh)<Abs(x)*TOLERANCE)
		*i = nx-1;
	else {
		*i = usrbini(x);
		if (*i<0 || *i>=nx)
			return false;
	}

	if (Abs(y-ylow)<Abs(y)*TOLERANCE)
		*j = 0;
	else
	if (Abs(y-yhigh)<Abs(y)*TOLERANCE)
		*j = ny-1;
	else {
		*j = usrbinj(y);
		if (*j<0 || *j>=ny)
			return false;
	}

	if (Abs(z-zlow)<Abs(z)*TOLERANCE)
		*k = 0;
	else
	if (Abs(z-zhigh)<Abs(z)*TOLERANCE)
		*k = nz-1;
	else {
		*k = usrbink(z);
		if (*k<0 || *k>=nz)
			return false;
	}
	return true;
} // xyz2ijk

/** get stored value (log10 in case of log)
 * @param x,y,z	location to return value
 * @return value at (x,y,z) location
 */
bool Usrbin::getData(const double x, const double y, const double z, double* value) const
{
	int i,j,k;

	if (xyz2ijk(x,y,z, &i, &j, &k)) {
		if (_checker)
			*value = ((i^j^k)&1? 0.0 : 1.0);
		else
			*value = _normValue((double)data[k*nynx + j*nx + i]);
		return true;
	} else {
		*value = 0.0;
		return false;
	}
} // getData

/** get stored value (log10 in case of log) by region (ONLY for binning per region)
 * @param region
 * @return value for region
 */
bool Usrbin::getData(const int region, double* value) const
{
	if (!data) return false;

	if (_type!=Type::region_point && _type!=Type::region) return false;

	// First set...
	int rstart = (int)xlow  - 1;
	int rend   = (int)xhigh - 1;
	int rstep  = (int)dx;
	//int rlen = (int)nx;

	if (region < rstart || region > rend) return false;

	if ((region - rstart) % rstep) return false;

	*value = _normValue((double)data[(region - rstart) / rstep]);
	return true;
} // getData

/** get stored value (log10 in case of log)
 * @param r,phi,z	location to return value
 * @return value at (r,phi,z) location
 */
bool Usrbin::getDataRPhiZ(double r, double phi, double z, double* value) const
{
	assert( _type==Type::RPhiZ_point    || _type==Type::RPhiZ ||
		_type==Type::RPhiZsym_point || _type==Type::RPhiZsym);
	assert(!_checker && data);

	if (_type==Type::RPhiZsym_point || _type==Type::RPhiZsym)
		if (z<0.0) z = -z;

	int i, j, k;
	// be more flexible on the limits
	if (Abs(r-xlow)<Abs(r)*TOLERANCE)
		i = 0;
	else
	if (Abs(r-xhigh)<Abs(r)*TOLERANCE)
		i = nx-1;
	else {
		i = Int((r-xlow)/dx);
		if (i<0 || i>=nx)
			return false;
	}

	if (Abs(phi-ylow)<Abs(phi)*TOLERANCE)
		j = 0;
	else
	if (Abs(phi-yhigh)<Abs(phi)*TOLERANCE)
		j = ny-1;
	else {
		j = Int((phi-ylow)/dy);
		if (j<0 || j>=ny)
			return false;
	}

	if (Abs(z-zlow)<Abs(z)*TOLERANCE)
		k = 0;
	else
	if (Abs(z-zhigh)<Abs(z)*TOLERANCE)
		k = nz-1;
	else {
		k = Int((z-zlow)/dz);
		if (k<0 || k>=nz)
			return false;
	}

	*value = _normValue(data[k*nynx + j*nx + i]);
	return true;
} // getDataRPhiZ

/** intersect a "ray" given by a point and direction with a Cartesian usrbin boundary
 * @param P	I=starting point, O=point on boundary of usrbin
 * @param D	direction of ray
 * @return	distance traveled >=0. <0 if no interaction is found
 */
inline double Usrbin::intersectXYZ(Point& P, const Vector& D) const
{
	// move to boundary
	double t = 0;
	if (P.x < xmin()) {
		if (D.x<=0.0) return -1.0;		// going out ignore
		double f = (xmin() - P.x) / D.x;	// distance to travel
		t += f;					// increase distance traveled
		P.x  = xmin();				// new position
		P.y += f*D.y;
		P.z += f*D.z;
	}
	if (P.y < ymin()) {
		if (D.y<=0.0) return -1.0;
		double f = (ymin() - P.y) / D.y;
		t += f;
		P.x += f*D.x;
		P.y  = ymin();
		P.z += f*D.z;
	}
	if (P.z < zmin()) {
		if (D.z<=0.0) return -1.0;
		double f = (zmin() - P.z) / D.z;
		t += f;
		P.x += f*D.x;
		P.y += f*D.y;
		P.z  = zmin();
	}

	if (P.x > xmax()) {
		if (D.x>=0.0) return -1.0;
		double f = (xmax() - P.x) / D.x;
		t += f;
		P.x  = xmax();
		P.y += f*D.y;
		P.z += f*D.z;
	}
	if (P.y > ymax()) {
		if (D.y>=0.0) return -1.0;
		double f = (ymax() - P.y) / D.y;
		t += f;
		P.x += f*D.x;
		P.y  = ymax();
		P.z += f*D.z;
	}
	if (P.z > zmax()) {
		if (D.z>=0.0) return -1.0;
		double f = (zmax() - P.z) / D.z;
		t += f;
		P.x += f*D.x;
		P.y += f*D.y;
		P.z  = zmax();
	}
	return t;
} // intersectXYZ

/** initialize step in Cartesian mesh for one direction
 * @param pos	position component of this direction x,y or z
 * @param dir	direction cosine on this direction
 * @param low	if inside the voxel lower direction value of the current bin
 * @param high	if inside the voxel higher direction value of the current bin
 *
 * @param t	intersection distance on his direction
 * @param dt	step increment when traversing subsequent steps
 * @param step	step index increase -1/0/+1 to the next bin
 * @param out	out bin index
 */
void Usrbin::_xyzStepInit(double pos, double dir, double low, double high,
			double* t, double* dt, int* step, int* out) const
{
	if (dir > SMALL) {
		*t    = (high - pos) / dir;
		*dt   = dx / dir;
		*step = 1;
		*out  = nx;
	} else
	if (dir < -SMALL) {
		*t    = (low - pos) / dir;
		*dt   = -dx / dir;
		*step = -1;
		*out  = -1;
	} else {
		*t    = INFINITE;
		*dt   = 0.0;
		*step = 0;
		*out  = 0;
	}
} // _xyzStepInit

/** distribute amount 'value' to all the usrbin cells that the line start-end will cross.
 * The amount scored per cell is proportional to the distance travelled in the cell.
 * @param A, B		starting and ending points of the line to distribute the amount value.
 *			warning points can lie also outside the usrbin
 * @param value		amount to score (times the weight of the particle)
 * @param tracklength	if true score the amount as tracklength (do not divide by the total length)
 * @return the amount that was scored
 */
double Usrbin::add(Point A, Point B, double value, bool tracklength)
{
	assert(!_logscale && !_checker && data);

	if (hasMatrix()) {
		// Convert to usrbin coordinates
		matrix().transform(A);
		matrix().transform(B);
	}
	A.x -= xofs;				// FIXME convert offset to Point
	A.y -= yofs;
	A.z -= zofs;

	B.x -= xofs;
	B.y -= yofs;
	B.z -= zofs;

	Vector dir = B - A;
	double totallength = dir.normalize();
	if (!tracklength) value /= totallength;		// already divide by length

	//cout << "total length=" << totallength << endl;

	// for the moment it works only with XYZ
	assert(_type==Type::XYZ_point || _type==Type::XYZ);

	double t = intersectXYZ(A, dir);		// distance travelled to far
	if (t<0.0) return 0;
	//cout << "starting length=" << t << endl;

	// Find starting location
	int i,j,k;
	if (!xyz2ijk(A.x, A.y, A.z, &i, &j, &k))
		return 0;

	// Moving variables
	int    ptr = k*nynx + j*nx + i;
	double tx,    ty,    tz;		// distance to next intersection with x/y/z plane
	double tdx,   tdy,   tdz;		// distance increase for every x/y/z plane
	int    stepx, stepy, stepz;		// step direction -1/0/+1
	#define dptrx stepx
	int          dptry, dptrz;		// ptr change dptrx=stepx
	int    outx,  outy,  outz;		// ending condition

	_xyzStepInit(A.x, dir.x, usrbinx(i), usrbinx(i+1), &tx, &tdx, &stepx, &outx);
	_xyzStepInit(A.y, dir.y, usrbiny(j), usrbiny(j+1), &ty, &tdy, &stepy, &outy);
	_xyzStepInit(A.z, dir.z, usrbinz(k), usrbinz(k+1), &tz, &tdz, &stepz, &outz);
	dptry = stepy * nx;
	dptrz = stepz * nynx;

	// increase by initial starting length
	tx += t;
	ty += t;
	tz += t;
	double amount = 0.0;	// amount to score
	// Loop forever until we run out of length
	while (true) {
		//cout << "pos=" << i << ", " << j << ", " << k
		//	<< " t=" << t << " total=" << totallength << endl;
		if (tx<ty && tx<tz) {
			i += stepx;
			if (i==outx) return amount;
			ptr += dptrx;
			if (tx < totallength) {
				double dt = tx - t;		// length to score
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				t = tx;
			} else {
				double dt = totallength - t;
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				return amount;
			}
			tx += tdx;
		} else
		if (ty<tz) {
			j += stepy;
			if (j==outy) return amount;
			ptr += dptry;
			if (ty < totallength) {
				double dt = ty - t;		// length to score
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				t = ty;
			} else {
				double dt = totallength - t;
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				return amount;
			}
			ty += tdy;
		} else {
			k += stepz;
			if (k==outz) return amount;
			ptr += dptrz;
			if (tz < totallength) {
				double dt = tz - t;		// length to score
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				t = tz;
			} else {
				double dt = totallength - t;
				data[ptr] += (float)(dt*value);
				amount    += (float)(dt*value);
				return amount;
			}
			tz += tdz;
		}
	}
} // add

/** @return integral on a straight path from A to B, during time, assuming constant speed */
/*          NOTE: Returns -1 in case integral could not be calculated (e.g. A or B outside)*/
double Usrbin::integrate(Point A, Point B, double time) const
{
	assert(!_checker && data && time > 0);
	if (_checker || !data || time<=0) return 0;

	if (hasMatrix()) {
		matrix().transform(A);		// Convert to usrbin coordinates
		matrix().transform(B);
	}
	A.x -= xofs;				// FIXME convert offset to Point
	A.y -= yofs;
	A.z -= zofs;

	B.x -= xofs;
	B.y -= yofs;
	B.z -= zofs;

	Vector dir = B - A;
	double totallength = dir.normalize();
	if (Eq0(totallength,1e-10))
		return get(A) * time;		// stationary

	// for the moment it works only with XYZ
	assert(_type==Type::XYZ_point || _type==Type::XYZ);

	double speed = totallength / time;	// moving speed
	double t = intersectXYZ(A, dir);	// distance travelled to far
	if (t < 0)				// point outside the usrbin
		return -1;

	// Find starting location
	int i,j,k;
	if (!xyz2ijk(A.x, A.y, A.z, &i, &j, &k))
		return false;

	// Moving variables
	int    ptr = k*nynx + j*nx + i;
	assert(ptr >= 0);
	double tx,    ty,    tz;		// distance to next intersection with x/y/z plane
	double tdx,   tdy,   tdz;		// distance increase for every x/y/z plane
	int    stepx, stepy, stepz;		// step direction -1/0/+1
	#define dptrx stepx
	int          dptry, dptrz;		// dptr change dptrx=stepx
	int    outx,  outy,  outz;		// ending condition

	// safety catches
	assert(i+1 < nx && j+1 < ny && k+1 < nz);

	_xyzStepInit(A.x, dir.x, usrbinx(i), usrbinx(i+1), &tx, &tdx, &stepx, &outx);
	_xyzStepInit(A.y, dir.y, usrbiny(j), usrbiny(j+1), &ty, &tdy, &stepy, &outy);
	_xyzStepInit(A.z, dir.z, usrbinz(k), usrbinz(k+1), &tz, &tdz, &stepz, &outz);
	dptry = stepy * nx;
	dptrz = stepz * nynx;

	// increase by initial starting length
	tx += t;
	ty += t;
	tz += t;

	double integral = 0.0;
	// Loop forever until we run out of length
	while (true) {
		//cout << "pos=" << i << ", " << j << ", " << k
		//	<< " t=" << t << " total=" << totallength << endl;
		if (tx<ty && tx<tz) {
			i += stepx;
			if (i==outx) return integral;
			ptr += dptrx;
			double value = get(ptr);
			if (tx < totallength) {
				integral += (tx-t)/speed * value;	// tx-t = length
				t = tx;
			} else {
				integral += (totallength-t)/speed * value;
				return integral;
			}
			tx += tdx;
		} else
		if (ty<tz) {
			j += stepy;
			if (j==outy) return integral;
			ptr += dptry;
			double value = get(ptr);
			if (ty < totallength) {
				integral += (ty-t)/speed * value;	// ty-t = length
				t = ty;
			} else {
				integral += (totallength-t)/speed * value;
				return integral;
			}
			ty += tdy;
		} else {
			k += stepz;
			if (k==outz) return integral;
			ptr += dptrz;
			double value = get(ptr);
			if (tz < totallength) {
				integral += (tz-t)/speed * value;	// tz-t = length
				t = tz;
			} else {
				integral += (totallength-t)/speed * value;
				return integral;
			}
			tz += tdz;
		}
	}

	// FIXME integral can be negative if we integrate over neutron balance
	assert(integral >= 0.0);
	return integral;
} // integrate
