////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "anisotropicmicrodist.h"

CAnisotropicMicroDist::CAnisotropicMicroDist(double X, double Y) : m_ExpX( X ), m_ExpY( Y )
{
  m_ExpX = Range(1.0, m_ExpX, 1000.0);
	m_ExpY = Range( 1.0, m_ExpX, 1000.0 );
}

CAnisotropicMicroDist::~CAnisotropicMicroDist(void)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CAnisotropicMicroDist::Sample_f( const Vector& Out, Vector& In, const Vector& Normal, double u1, double u2, float& Pdf ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // Sample from first quadrant and remap to hemisphere to sample \wh
  double Phi, CosTheta;
  if( u1 < .25) {
    SampleFirstQuadrant( 4. * u1, u2, Phi, CosTheta);
  }
  else if( u1 < .5) {
    u1 = 4. * (.5 - u1);
    SampleFirstQuadrant(u1, u2, Phi, CosTheta);
    Phi = PI - Phi;
  }
  else if( u1 < .75) {
    u1 = 4. * (u1 - .5);
    SampleFirstQuadrant(u1, u2, Phi, CosTheta);
    Phi += PI;
  }
  else {
    u1 = 4.0 * (1. - u1);
    SampleFirstQuadrant(u1, u2, Phi, CosTheta);
    Phi = 2.0 * PI - Phi;
  }

  double SinTheta = sqrt( std::max( 0.0, 1.0 - CosTheta * CosTheta ) );

	Vector HalfVec = SphericalDirection( SinTheta, CosTheta, Phi );
	Vector vX, vY, vZ( Normal );
	Matrix3 uvwBase;

	if( fabs( vZ.x ) >= fabs( vZ.y ) ) {
		double invLength = 1.0 / sqrt( vZ.x * vZ.x + vZ.z * vZ.z );
		vX.x = +vZ.z * invLength;
		vX.y = 0.0;
		vX.z = -vZ.x * invLength;
	}
	else {
		double invLength = 1.0 / sqrt( vZ.y * vZ.y + vZ.z * vZ.z );
		vX.x = 0.0;
		vX.y = +vZ.z * invLength;
		vX.z = -vZ.y * invLength;
	}

	vY = vZ.cross( vX );

	uvwBase.make( vX, vY, vZ );
	uvwBase.transpose();

	HalfVec = uvwBase * HalfVec;

	double DotOutHalfVec = Out.dot( HalfVec );
  if(DotOutHalfVec < 0.0) {
		HalfVec = -HalfVec;
    DotOutHalfVec *= -1;
  }

	// Compute incident direction by reflecting about $\wh$
	In = -Out + HalfVec * 2.0 * DotOutHalfVec;
	In.normalize();

  // Compute PDF for \wi from Anisotropic distribution
  Pdf = D( HalfVec, Normal ) / (4.0 * DotOutHalfVec);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CAnisotropicMicroDist::SampleFirstQuadrant(double u1, double u2, double& Phi, double& CosTheta) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  if( m_ExpX == m_ExpY )
    Phi = PI * u1 * 0.5;
  else
    Phi = atan( sqrt((m_ExpX + 1) / (m_ExpY + 1)) * tan(PI * u1 * 0.5));

  double CosPhi = cos(Phi);
  double SinPhi = sin(Phi);

  CosTheta = pow(u2, 1.0/( m_ExpX * CosPhi * CosPhi + m_ExpY * SinPhi * SinPhi + 1));
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

float CAnisotropicMicroDist::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector HalfVec = Out + In;
  HalfVec.normalize();

  // Compute PDF for \wi from Anisotropic distribution
  return ( D(HalfVec, Normal) / (4.0 * Out.dot( HalfVec )) );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Trowbridge Reitz - microfacets

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////

float CTrowbridgeReitzDist::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector HalfVec = Out + In;
	HalfVec.normalize();

	if( m_bSampleVisArea )
	  return D( HalfVec, Normal ) * G1( Out, In, HalfVec, Normal ) * fabs( Out.dot( HalfVec ) ) / fabs(Normal.dot( Out ) );

	return D( HalfVec, Normal ) * fabs( Normal.dot( HalfVec ) );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CTrowbridgeReitzDist::Sample_f( const Vector& Out, Vector& In, const Vector& Normal,
                                     double u1, double u2, float& Pdf ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	Vector HalfVec;
	double DotOutHalfVec = 0.0;

	if( !m_bSampleVisArea ) {
		double cosTheta = 0, Phi = (2 * PI) * u2;

		if( m_AlphaX == m_AlphaY) {
			double tanTheta2 = m_AlphaX * m_AlphaX * u1 / (1.0 - u1);
			cosTheta = 1 / std::sqrt( 1 + tanTheta2 );
		}
		else {
			Phi = std::atan( m_AlphaY / m_AlphaX * std::tan( 2 * PI * u2 + .5 * PI ) );
			if(u2 > .5)
				Phi += PI;

			double sinPhi = std::sin( Phi ), cosPhi = std::cos( Phi );
			const double AlphaX2 = m_AlphaX * m_AlphaX, AlphaY2 = m_AlphaY * m_AlphaY;
			const double Alpha2 = 1 / (cosPhi * cosPhi / AlphaX2 + sinPhi * sinPhi / AlphaY2);
			double tanTheta2 = Alpha2 * u1 / (1 - u1);
			cosTheta = 1 / std::sqrt( 1 + tanTheta2 );
		}

		double sinTheta = std::sqrt( std::max( 0., 1. - cosTheta * cosTheta ) );
		HalfVec = SphericalDirection( sinTheta, cosTheta, Phi );
		HalfVec = CBaseSampler::canonicalToBasis( HalfVec, Normal );

		DotOutHalfVec = Out.dot( HalfVec );
		if(DotOutHalfVec < 0.0) {
			HalfVec = -HalfVec;
			DotOutHalfVec *= -1;
		}
	}
	else {
		bool bFlip = CBaseSampler::sameHemisphere( Out, Normal );
		HalfVec = TrowbridgeReitzSample( bFlip ? -Out : Out, Normal, m_AlphaX, m_AlphaY, u1, u2 );

		if(bFlip) {
			HalfVec = -HalfVec;
			DotOutHalfVec = -1;
		}
	}

	// Compute incident direction by reflecting about $\wh$
	In = -Out + HalfVec * 2.0 * DotOutHalfVec;
	In.normalize();

	Pdf = GetPDF( Out, In, Normal );
}
// vim:ts=2
