////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "blinnmicrodist.h"

CBlinnMicroDist::CBlinnMicroDist( double Exp) : m_Exponent( Exp )
{
  m_Exponent = std::max( 0.0, std::min( 1000.0, m_Exponent ) );
}

CBlinnMicroDist::~CBlinnMicroDist(void)
{
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CBlinnMicroDist::Sample_f( const Vector& Out, Vector& In, const Vector& Normal, double u1, double u2, float& Pdf ) const

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  // Compute sampled half-angle vector $\wh$ for Blinn distribution
  double CosTheta = pow( u1, 1.0 / (m_Exponent + 1));
  double SinTheta = sqrt( std::max( 0.0, 1.0 - CosTheta * CosTheta ) );
  double Phi = u2 * 2.0 * PI;

  Vector HalfVec = SphericalDirection( SinTheta, CosTheta, Phi);
  Vector vX, vY, vZ( Normal );
  Matrix3 uvwBase;

	if( fabs( vZ.x ) >= fabs( vZ.y ) ) {
		double invLength = 1.0 / sqrt( vZ.x * vZ.x + vZ.z * vZ.z );
		vX.x = +vZ.z * invLength;
		vX.y = 0.0;
		vX.z = -vZ.x * invLength;
	}
	else {
		double invLength = 1.0 / sqrt( vZ.y * vZ.y + vZ.z * vZ.z );
		vX.x = 0.0;
		vX.y = +vZ.z * invLength;
		vX.z = -vZ.y * invLength;
	}

	vY = vZ.cross( vX );

	uvwBase.make( vX, vY, vZ );
	uvwBase.transpose();

  HalfVec = uvwBase * HalfVec;

  double DotOutHalfVec = Out.dot( HalfVec);
  if ( DotOutHalfVec < 0.0)
    HalfVec = -HalfVec;

  // Compute incident direction by reflecting about $\wh$
  In = -Out + HalfVec * 2.0 * DotOutHalfVec;
  In.normalize();

  // Compute PDF for \wi from Blinn distribution
  Pdf = ((m_Exponent + 2.0) * pow( CosTheta, m_Exponent )) / (2.0 * PI * 4.0 * DotOutHalfVec );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

float CBlinnMicroDist::GetPDF(const Vector& Out, const Vector& In, const Vector& Normal ) const

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector HalfVec( Out + In);
  HalfVec.normalize();

  double   CosTheta = fabs( Normal.dot( HalfVec) );

  // Compute PDF for \wi from Blinn distribution
  float  BlinnPDF = (( m_Exponent + 2.f) * powf( CosTheta, m_Exponent)) / (2.0 * PI * 4.0 * Out.dot( HalfVec ) );
  return BlinnPDF;
}
// vim:ts=2
