////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdflafortune.h"

#include <cmath>
#include <climits>

CBxDFLaFortune::CBxDFLaFortune( const Color& Diffuse, unsigned int NrOfLobes, const Color* pX, const Color* pY, const Color* pZ, const Color* pExp, CBxDFBase::BxDFType Type) :
                                CBxDFBase( Type ), m_NrOfLobes( NrOfLobes ), m_Diffuse( Diffuse ),  m_pX( pX ), m_pY( pY ), m_pZ( pZ ), m_pExp( pExp )
{
}

CBxDFLaFortune::~CBxDFLaFortune(void)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFLaFortune::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( LightColor ); UNUSED( Power ); UNUSED( Shade ); UNUSED( RenderMat );

  double          Rho = 1.0 / PI;
  Vector          InDir(  In.dir() );
  Vector          OutDir(-Out.dir() );
  Color           TmpCol;

  double Dot = Normal.dot( InDir );
  if(Dot < SMALL3D )
    return 0.0;

  for( unsigned int i = 0; i < m_NrOfLobes; ++i) {
    // Add i_th Phong lobe
    Color CurLobe = m_pX[i] * ( OutDir.x * InDir.x) + m_pY[i] * ( OutDir.y * InDir.y) + m_pZ[i] * ( OutDir.z * InDir.z );

    TmpCol(0) = pow( CurLobe.red(),   m_pExp[i].red() );
		TmpCol(1) = pow( CurLobe.green(), m_pExp[i].green() );
		TmpCol(2) = pow( CurLobe.blue(),  m_pExp[i].blue()  );

    // convert from spectral info to RGB -> luminance (ideally it would return the full spectral info as a color array)
    Rho += TmpCol.spectralToRGB().luminance();
  }

  return Rho;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFLaFortune::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power ); UNUSED( Type );

  Vector InDir, OutDir, TmpVec;
  Pdf = 1.0;

  OutDir = -Out.dir();

  // Cosine-sample the hemisphere, flipping the direction if necessary
  // set up orthogonal sampling basis

  unsigned int CurLobe = (int)(pRandNr[0] * INT_MAX) % (m_NrOfLobes);
  if( CurLobe == m_NrOfLobes - 1) {
    InDir = CBaseSampler::cosineSampleHemisphere( Normal, pRandNr[0], pRandNr[1] );

    // reflect the outgoing ray if Out & In are on opposite sides
    if( OutDir.dot( InDir ) < 0 ) {
      InDir = -InDir;
    }
  }
   else {
    // Sample lobe CurLobe for Lafortune BRDF
    double XLum = m_pX[CurLobe].luminance();
    double YLum = m_pY[CurLobe].luminance();
    double ZLum = m_pZ[CurLobe].luminance();
    double CosTheta = pow( pRandNr[0], 1.0 / (0.8 * m_pExp[CurLobe].luminance() + 1));
    double SinTheta = sqrt( std::max( 0.0, 1.0 - CosTheta * CosTheta ) );
    double Phi = pRandNr[1] * 2.0 * PI;

    // create an orthogonal basis of the lobe centered around vZ
    Vector vZ( XLum * (OutDir.x), YLum * (OutDir.y), ZLum * (OutDir.z) );
    Vector vX, vY;
    vZ.normalize();

		if( fabs( vZ.x ) >= fabs( vZ.y ) ) {
			double InvLength = 1.0 / sqrt( vZ.x * vZ.x + vZ.z * vZ.z );
			vX.x = +vZ.z * InvLength;
			vX.y = 0.0;
			vX.z = -vZ.x * InvLength;
		}
		else {
			double InvLength = 1.0 / sqrt( vZ.y * vZ.y + vZ.z * vZ.z );
			vX.x = 0.0;
			vX.y = +vZ.z * InvLength;
			vX.z = -vZ.y * InvLength;
		}

		vY = vZ.cross( vX );

    InDir = SphericalDirection( SinTheta, CosTheta, Phi, vX, vY, vZ);
  }

  // sample incoming ray
	In = Out;
	In.init();
	In.incDepth();	// to avoid infinite loops
	In.push( RaySegment( Out.hit( -SMALL3D4 ), InDir, Out.prevZone() ) );

  Pdf = GetPDF( OutDir, InDir, Normal);
	if( Pdf == 0 ) {
		return 0;
	}

  return f( Out, In, Normal, RenderMat, LightColor, Shade);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFLaFortune::GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const

// ATTENTION: Out should be directed away from in!!!
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Normal );

  if( !CBaseSampler::sameHemisphere( Out, In))
    return 0.f;

  double PdfSum = fabs( In.z ) / PI;
  for( unsigned int i = 0; i < m_NrOfLobes; ++i ) {
    double XLum = m_pX[i].luminance();
    double YLum = m_pY[i].luminance();
    double ZLum = m_pZ[i].luminance();
    Vector LobeCenter( XLum * (Out.x), YLum * (Out.y), ZLum * (Out.z) );
    double Exp = 0.8 * m_pExp[i].luminance();

    PdfSum += ( Exp + 1.0 ) * pow( std::max( 0.0, In.dot( LobeCenter ) ), Exp );
  }

  return  fabs(PdfSum / (1.0 + m_NrOfLobes));
}
// vim:ts=2
