////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdftranslucent.h"
#include "bxdffresnel.h"

CBxDFTranslucent::CBxDFTranslucent() : CBxDFBase( static_cast<BxDFType>( BSDF_TRANSMISSION | BSDF_SPECULAR) ),
																			 m_pFresnelObj( new CFresnelDiel )
{
}

CBxDFTranslucent::~CBxDFTranslucent(void)
{
  delete m_pFresnelObj;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFTranslucent::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power ); UNUSED( LightColor ); UNUSED( Shade ); UNUSED( Type );

	Vector RefrNormal( Normal );
	Vector OutDir = Out.dir();

	// Figure out which $\eta$ is incident and which is transmitted
	double cosI = -RefrNormal.dot( OutDir );
	bool   bEntering = cosI > 0.;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:
	double EtaT = RenderMat.ior();
	double n = 0.0;

	// check if the ray is incident, otherwise reverse the object normal
	if( !bEntering ) {
		std::swap( EtaI, EtaT );
		EtaT = 1.0;                     // as a first assumption we go from the dense medium into air/vacuum
		RefrNormal = RefrNormal * -1.0;
	}

	n = EtaI / EtaT;
	Pdf = 1.0;

	double cosT2 = 1.0 - n * n * (1.0 - cosI * cosI);
	if( cosT2 > 0.0f ) {
		Vector T = (OutDir * n) + RefrNormal * (n * cosI - sqrt( cosT2 )); // create refraction ray
		// TODO: removed the slow the calculation of the square root
		// Vector T = ( OutDir + RefrNormal * cosI) * n - RefrNormal * cosI;

		// distort our refraction ray
		T = CBaseSampler::uniformSampleCone( T, std::max( 1.0, RenderMat.fuzz()), pRandNr[0], pRandNr[1]);

		In = Out;
		In.init();
		In.incDepth();	// to avoid infinite loops
		In.push( RaySegment( Out.hit( SMALL3D4 ), T, Out.prevZone() ) );

	}
	else {  // total reflection
		In.depth = 32000;                                  // flag ray to be terminated
		return 0.0;
	}

	// Fresnel transmission,
	double Rho = 1.0 / fabs( cosI ) * (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - m_pFresnelObj->Evaluate( cosI, EtaI, EtaT ));
	return Rho;

}
// vim:ts=2
