////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "matbluepaint.h"

CMatBluePaint::CMatBluePaint( CBaseIntegrator* pInt ) : CPhysMatInterface( pInt, static_cast<CBxDFBase::BxDFType>( CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR  | CBxDFBase::BSDF_REFLECTION ), true /* false = no MC sampling required */ )
{
  static Color Diffuse( 0.3094f,    0.39667f,   0.70837f );
  static float xy0[3] =     {  0.870567f,  0.857255f,  0.670982f };
  static float z0[3] =      {  0.803624f,  0.774290f,  0.586674f };
  static float e0[3] =      { 21.820103f, 18.597755f,  7.472717f };
  static float xy1[3] =     { -0.451218f, -0.406681f, -0.477976f };
  static float z1[3] =      {  0.023123f,  0.017625f,  0.227295f };
  static float e1[3] =      {  2.774499f,  2.581499f,  3.677653f };
  static float xy2[3] =     { -1.031545f, -1.029426f, -1.026588f };
  static float z2[3] =      {  0.706734f,  0.696530f,  0.687715f };
  static float e2[3] =      { 66.899060f, 63.767912f, 57.489181f };

  static Color XY[3] = { Color( xy0[0], xy0[1], xy0[2]), Color( xy1[0], xy1[1], xy1[2]), Color( xy2[0], xy2[1], xy2[2]) };
  static Color  Z[3] = { Color( z0[0], z0[1], z0[2]), Color( z1[0], z1[1], z1[2]), Color( z2[0], z2[1], z2[2]) };
  static Color  E[3] = { Color( e0[0], e0[1], e0[2]), Color( e1[0], e1[1], e1[2]), Color( e2[0], e2[1], e2[2]) };

  m_pBxDFDiffuse		  = new CBxDFLaFortune( Diffuse, 3, XY, XY, Z, E, static_cast<CBxDFBase::BxDFType>( CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_REFLECTION ) );
  m_pFresnelSpecular	= new CFresnelPreDefConductor( 0.0, 1.0);
  m_pBxDFSpecular     = new CBxDFSpecReflection( m_pFresnelSpecular );
}

CMatBluePaint::~CMatBluePaint(void)
{
  delete m_pFresnelSpecular;
  delete m_pBxDFDiffuse;
  delete m_pBxDFSpecular;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatBluePaint::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  double Diffuse, Specular;

  Diffuse   = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );
  Specular  = m_pBxDFSpecular->f( Out, In, Normal, RenderMat, LightColor, Shade );

  return Diffuse + Specular * RenderMat.specular();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatBluePaint::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

	// choose between specular & diffuse reflection
	double Rho = 0;
	double TotalPDF = 0.0;

	if(pRandNr[0] < 0.5) {
		Rho = m_pBxDFSpecular->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
		TotalPDF += Pdf;
		TotalPDF += m_pBxDFDiffuse->GetPDF( -Out.dir(), In.dir(), Normal );
	}
	else {
		Rho = m_pBxDFDiffuse->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
		TotalPDF += Pdf;
		TotalPDF += m_pBxDFSpecular->GetPDF( -Out.dir(), In.dir(), Normal );
	}

	// renormalize Pdf
	Pdf /= TotalPDF;

	return Rho;

}
// vim:ts=2
