////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "matshinymetal.h"

CMatShinyMetal::CMatShinyMetal( CBaseIntegrator* pInt, const float cRoughness /*= 1.0 */, const float cAbsorption /* = 0.0 */) : CPhysMatInterface(pInt, static_cast<CBxDFBase::BxDFType>(CBxDFBase::BSDF_DIFFUSE | CBxDFBase::BSDF_SPECULAR | CBxDFBase::BSDF_GLOSSY | CBxDFBase::BSDF_REFLECTION), true /* true = MC sampling required for the microfacet distributions for example */),
                                m_Absorption(cAbsorption)
{
  m_pFresnelReflection = new CFresnelPreDefConductor( m_Absorption, 1.0);
  m_pFresnelSpecular   = new CFresnelPreDefConductor( m_Absorption, 1.0);
  m_pBlinnMicroDist    = new CBlinnMicroDist( cRoughness );

  m_pBxDFDiffuse        = new CBxDFPhong;
  m_pDiffuseMicroFacets = new CBxDFMicrofacet( m_pFresnelReflection, m_pBlinnMicroDist );      // microfacet diffuse reflector
  m_pBxDFSpecular       = new CBxDFSpecReflection( m_pFresnelSpecular );
}

CMatShinyMetal::~CMatShinyMetal(void)
{
  delete m_pFresnelReflection;
  delete m_pFresnelSpecular;
  delete m_pBxDFDiffuse;      // microfacet diffuse reflector
  delete m_pBxDFSpecular;     // specular reflection
  delete m_pBlinnMicroDist;
  delete m_pDiffuseMicroFacets;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatShinyMetal::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );

  double Diffuse, Specular, DiffuseMicroFacets;

  m_pBlinnMicroDist->SetExp( std::max( 1.0 / 12.0, RenderMat.shine() / 12.0 ) );
  // approximate eta
  m_pFresnelReflection->SetEta( CFresnelHelpers::FresnelApproxEta( RenderMat.specular() ) );
  m_pFresnelSpecular->SetEta( CFresnelHelpers::FresnelApproxEta( RenderMat.shine()) );
  // set absorption k --> done in ctor
  //   m_pFresnelReflection->SetAbsorption( m_Absorption );
  //   m_pFresnelSpecular->SetAbsorption( m_Absorption );

  Diffuse            = m_pBxDFDiffuse->f( Out, In, Normal, RenderMat, LightColor, Shade );
  DiffuseMicroFacets = m_pDiffuseMicroFacets->f( Out, In, Normal, RenderMat, LightColor, Shade );
  Specular           = m_pBxDFSpecular->f( Out, In, Normal, RenderMat, LightColor, Shade );

  return ( Diffuse + fabs(DiffuseMicroFacets) ) * (1.0 - RenderMat.specular()) + Specular * std::max( RenderMat.specular(), 0.1 );
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CMatShinyMetal::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, double* pRandNr, CBxDFBase::BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0*/ )

// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
// continue recursively.
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Power );


  // choose between specular & diffuse reflection
  double Rho = 0;
  double TotalPDF = 0.0;

  if(pRandNr[0] < 0.5) {
    Rho = m_pBxDFSpecular->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade );
    TotalPDF += Pdf;
    TotalPDF += m_pBxDFDiffuse->GetPDF( -Out.dir(), In.dir(), Normal);
  }
  else {
    Rho = m_pBxDFDiffuse->Sample_f( Out, In, Pdf, Normal, RenderMat, pRandNr, Type, LightColor, Shade);
		TotalPDF += Pdf;
		TotalPDF += m_pBxDFSpecular->GetPDF( -Out.dir(), In.dir(), Normal);
  }

  // renormalize Pdf
  Pdf /= TotalPDF;

  return Rho;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
CMatShinyMetalEx::CMatShinyMetalEx( CBaseIntegrator* pInt, const float cRoughnessX, const float cRoughnessY, const float cAbsorption /* = 0.0 */ ) :
                  CMatShinyMetal( pInt, 0, cAbsorption)
{
	m_pFresnelReflection = new CFresnelPreDefConductor( m_Absorption, 1.0 );
	m_pFresnelSpecular = new CFresnelPreDefConductor( m_Absorption, 1.0 );
  m_pTrowbridgeReitz = new CTrowbridgeReitzDist( cRoughnessX, cRoughnessY, true );
	m_pBxDFDiffuse = new CBxDFPhong;

	m_pDiffuseMicroFacets = new CBxDFMicrofacet( m_pFresnelReflection, m_pTrowbridgeReitz );      // microfacet diffuse reflector
	m_pBxDFSpecular = new CBxDFSpecReflection( m_pFresnelSpecular );
}

CMatShinyMetalEx::~CMatShinyMetalEx()
{
	delete m_pFresnelReflection;
	delete m_pFresnelSpecular;
	delete m_pBxDFDiffuse;      // microfacet diffuse reflector
	delete m_pBxDFSpecular;     // specular reflection
	delete m_pTrowbridgeReitz;
	delete m_pDiffuseMicroFacets;
}

// vim:ts=2
