/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 *
 * The geometry module is split into the following classes
 *
 * Classes Prefix:
 *    Gxxx: geometry information (Geometry)
 *    Cxxx: caching information, (Engine)
 *    Vxxx: viewer related information (Viewer)
 *
 * Meaning:
 *    +---      owns the element
 *    +-->      pointer or reference to element
 *
 *    GBody
 *    =====
 *    +--- GBody::Type type
 *    +--- string   name
 *    +--- .... parameters
 *                Holder of the geometrical properties
 *                of each body. Provides setting
 *                intersection calculation, ...
 *                as well save/restore functions
 *
 *    GZone
 *    =====
 *    +--- Format   format     zone representation
 *    +--- GBody    expr[]
 *    +--- BBox     bbox
 *                Description of a single zone (normal)
 *                or an advanced with RPN representation.
 *                Calculation of intersection, inclusion
 *                etc..
 *
 *    GRegion
 *    +--- GZone    zones[]
 *    +--- Matrix4  matrix
 *    +--- Type     type
 *    +--- Material material
 *                Region as a union of a list of zones
 *
 *    VBody
 *    =====
 *    +--> GBody    _body
 *    +--- Conic    C[]     list of conics
 *    +--- Vertex2D V[]     list of arrays of intersecting vertices
 *                          Vertex2D
 *    +--- Location location
 *    +--- bool     visible
 *
 *    VZone
 *    =====
 *    +--> GZone    _zone
 *    +--> VRegion  _region
 *    +--- bool     location whether "in" information is valid
 *    +--- bool     in      inside window
 *
 *    VRegion
 *    =======
 *    +--> GRegion  _region
 *    +--> Kernel   _kernel
 *    +--- VZone    _zone   list of vzones
 *    +--- dword    _color  rgb color to plot
 *    +--- double   _value  region numerical value
 *    +--- int      _alpha  transparency
 *
 *    CBody
 *    +--- float    tmin,tmax
 *    +--> GBody    body
 *                 Cached body information, per viewport
 *                 (or per kernel) with the entry,exit
 *                 information
 *
 *    CZone, CRegion - no longer exist as it is faster
 *                 to recalculate than to cache the
 *                 information
 *
 *    Geometry
 *    ========
 *    +--- GBody    bodies
 *    +--- GRegion  regions
 *    +--- GVoxel   voxel
 *    +--- Material materials
 *    +--- Light    lights
 *                Owner class of geometry properties
 *                bodies, regions, lattices, lights, materials etc.
 *                Contains all methods for manipulating
 *                the geometry structure, common to all viewers
 *
 *    Engine (Geometry Low Level Single threaded Engine)
 *    ======
 *    +--- CBody    bodies[]
 *    +--- CZone    zonesSorted[]
 *    +--- CBody    clipbody
 *    +--- CBody    projectbody
 *    +--> Geometry geometry
 *    +--> Kernel   kernel
 *                An asynchronous low level class that
 *                owns C(ached) structures calculating and
 *                caching information on the current kernel/viewer/geometry
 *
 *    Kernel (Geometry Kernel)
 *    ======
 *    +--- Engine   engine (master), engines[] (threadpool)
 *    +--> Geometry geometry
 *    +--- VBody    bodies
 *    +--- VRegion  regions
 *    +--- VVoxel   voxel
 *    +--- Viewport view
 *    +--- ThreadPool pool
 *                Job scheduling class owner of the V(iewer)
 *                structures
 *
 *    Viewer (GeometryViewer)
 *    ======
 *    +--> Geometry
 *    +--> Kernel
 *    +--- Painter  painter
 *    +--- *Layer   *layer
 *                Holder class of viewer depended properties
 *                like layers, colors, etc Contains all methods
 *                for drawing either synchronous or asynchronous
 *                Each viewer is referring to the geometry
 *                as well the drawing kernels
 */

#ifndef __GEOMETRY_GEOMETRY_H
#define __GEOMETRY_GEOMETRY_H

#include <map>
#include <vector>

#include "graph/color.h"
#include "tool/bintree.h"

#include "geo.h"
#include "gzone.h"
#include "voxel.h"
#include "light.h"
#include "gregion.h"
#include "material.h"

class GBody;
class GZone;
class Token;
class Matrix4;

typedef void (*ErrorHandler)(std::string&);

/* ============================= Geometry ============================= */
/** Geometry class */
class Geometry {
public:
	enum class Mode {
		Fluka,
		Moira,
		Mcnp,
		Phits,
		Penelope,
	};
	using StringVector = std::vector<std::string>;

public:
	/* drawing flags */
	int	axisLen;		/** axis length			*/

	int	gridStep;		/** pixel step for grid		*/
	int	lighterLevel;		/** lighter non-selected regions*/

	/* colors */
	dword	_backgroundColor;	/** background Color		*/
	dword	errorColor;		/** error line color		*/
	dword	gridTextColor;		/** color of grid text		*/
	dword	paletteTextColor;	/** color of palette text	*/
	dword	latticeColor;		/** lattice color		*/
	dword	regionColor;		/** region boundary colors	*/
	dword	regionErrorColor;	/** region error color		*/
	dword	selectColor;		/** selected item color		*/
	dword	titleColor;		/** color for title		*/
	dword	vertexColor;		/** vertex color		*/
	dword	visibleColor;		/** visible item color		*/
	dword	voxelColor;		/** voxel color			*/
	dword	zoneColor;		/** zone line color		*/
	dword	messageColor;		/** message color		*/
	dword	labelColor;		/** labels color		*/

	Color3D select3DColor;		/** 3d selection color		*/
	Color3D	wireframeColor;		/** wireframe color		*/
	Color3D	bodyBBoxInColor;	/** body BB inside color	*/
	Color3D	bodyBBoxOutColor;	/** body BB outside color	*/
	Color3D	zoneBBoxColor;		/** zone BB color		*/
	Color3D	regionBBoxColor;	/** region BB color		*/

	// FIXME lights should be protected
	int	deflights;		/** number of default lights	*/
	int	nlights;		/** total number of lights	*/
	Light	light[MAXLIGHT];	/** # lights in the system	*/

	GVoxel	voxel;			/** voxel object		*/

	std::map<std::string, GBody*> bodiesMap;
	std::vector<GBody*> bodies;	/** list of bodies		*/

	std::map<std::string, GRegion*> regionsMap;
	std::vector<GRegion*> regions;	/** list of regions		*/

	std::map<std::string, Material*> materialsMap;
	std::vector<Material*> materials;/** list of materials		*/

	GBody*	editBody;		/** currently editing body	*/
	GRegion	editRegion;		/** editing region		*/
protected:
	Mode	_mode;			/** operation mode		*/
	dword	_transparentColor;	/** transparent/ignore Color	*/
	std::string	_msg;		/** msg to display in viewports	*/
	int	_generation;		/** calculation generation no	*/
	StringVector*	_errors;	/** error string vector		*/

#if _DEBUG>0
mutable	bool	_locked;
#endif
mutable pthread_rwlock_t rwlock;        /** Read/write lock,		*/
					/*  protects geometry definition*/
mutable	pthread_mutex_t	mutexEdit;	/** editRegion mutex		*/

public:
static	bool	developer;		/** developer flag		*/

public:
	Geometry();
	~Geometry();
	void	cleanup();

	/* mode */
	Mode	mode()		const	{ return _mode; }
	void	mode(Mode amode) {
				_mode = amode;
				assert(nBodies()==0);
				assert(nRegions()==0);
			}

	/* mutex */
#if _DEBUG>0
	// only for debugging. python for some reason when
	// input() or breakpoint() is used continues the execution of the Tk messages
	bool	isLocked()	const	{ return _locked; }
	int	lockWrite()	const	{ _locked = true;  return pthread_rwlock_wrlock(&rwlock); }
	int	unlockWrite()	const	{ _locked = false; return pthread_rwlock_unlock(&rwlock); }
#else
	int	lockWrite()	const	{ return pthread_rwlock_wrlock(&rwlock); }
	int	unlockWrite()	const	{ return pthread_rwlock_unlock(&rwlock); }
#endif
	int	lockRead()	const	{ return pthread_rwlock_rdlock(&rwlock); }
	int	unlockRead()	const	{ return pthread_rwlock_unlock(&rwlock); }
	int	lockEdit()	const	{ return pthread_mutex_lock(&mutexEdit); }
	int	unlockEdit()	const	{ return pthread_mutex_unlock(&mutexEdit); }

	/* colors */
	void	backgroundColor(const dword color) { _backgroundColor = color & FLAG_COLORMASK; }
	dword	backgroundColor()	const	{ return _backgroundColor; }
	void	makeLatticeColor();
	dword	transparentColor()	const	{ return _transparentColor; }

	/* generation */
	int	nextGeneration()		{ return ++_generation; }

	/* bodies */
	size_t	nBodies()		const	{ return bodies.size(); }
	GBody*	addBody(const std::string& name, const std::string& type);
	GBody*	addRegionBody(const std::string& name);
	//GBody*	newBody(const std::string& name, const std::string type, ...);
	GBody*	getBody(const std::string& name) {
				auto search = bodiesMap.find(name);
				if (search != bodiesMap.end())
					return search->second;
				else
					return nullptr;
			}
const	GBody*	getBody(const std::string& name) const	{	// same as above to keep compiler happy
				auto search = bodiesMap.find(name);
				if (search != bodiesMap.end())
					return search->second;
				else
					return nullptr;
			}
	GBody*	getBody(const size_t  id)	{ return (GBody*)bodies[id]; }	// WARNING check limits
const	GBody*	getBody(const size_t  id) const	{ return (GBody*)bodies[id]; }	// WARNING check limits
	void	delBodies();
	void	invalidateBody(GBody* body);

	/* regions */
	size_t	nRegions()		const	{ return regions.size(); }
	GRegion* addRegion(const std::string& name);
	GRegion* getRegion(const std::string& name) {
				auto search = regionsMap.find(name);
				if (search != regionsMap.end())
					return search->second;
				else
					return nullptr;
			}
const	GRegion* getRegion(const std::string& name) const {	// same as above to keep compiler happy
				auto search = regionsMap.find(name);
				if (search != regionsMap.end())
					return search->second;
				else
					return nullptr;
			}
	GRegion* getRegion(const size_t id)	{
				if (id < nRegions())
					return regions[id];
				else
					return NULL;
			}
const	GRegion* getRegion(const size_t id) const {
				if (id < nRegions())
					return regions[id];
				else
					return NULL;
			}
	void	delRegions(bool fast=false);

	/* zone setup */
	GZone*	addZone(GRegion* region, GZone::Format aformat);
	bool	add2exp(GRegion* reg, const char* token);

	/* materials */
	size_t	nMaterials()		const	{ return materials.size(); }
	Material* addMaterial(const std::string& name);
	Material* getMaterial(const std::string& name) {
				auto search = materialsMap.find(name);
				if (search != materialsMap.end())
					return search->second;
				else
					return nullptr;
			}
	Material* getMaterial(const size_t id)	{
				if (id < nMaterials())
					return materials[id];
				else
					return NULL;
			}
const	Material* getMaterial(const size_t id) const {
				if (id < nMaterials())
					return materials[id];
				else
					return NULL;
			}
	void	delMaterials();

	/* lights */
	bool	addLight(const Light& l);
	void	delLights()		{ nlights = deflights; }

	/** set message to display in all viewports */
	void	message(const std::string& a)	{ _msg = a; }
const	std::string&	message()	const	{ return _msg; }

	/** Derive */
	int	derive();

	/** error handler */
	void	errors(StringVector* vec)	{ _errors = vec;  }
	void	errorMsg(std::string msg);

	/* Info */
	size_t	bodiesMemory()	const;
	size_t	regionsMemory()	const;
	size_t	memory()	const;
	void	printInfo()	const;
	void	printMemory()	const;

	bool	parseRegion(GRegion& region, Token& token,
			bool convert=false, std::vector<int>* zonesSep=nullptr);
	bool	parseRegion(GRegion& region, const std::string& expstr, bool convert=false);
private:
	void	makeDefaultLights();

friend class GeometryViewer;
friend class DecorationLayer;
friend class D2Layer;
friend class D3Layer;
}; // Geometry

#endif
