/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	15-Mar-2012
 */

#include "gzone.h"

#include <climits>

#include "gregion.h"
#include "boxbody.h"
#include "geometry.h"
#include "regionbody.h"
#include "tool/stack.h"
#include "math/plane.h"

class Quad;

using std::cout;
using std::cerr;
using std::endl;

thread_local std::vector<GBodyOrderAccel> GZone::zoneAccel;

// =================================== GZone ===================================

/** zone constructor
 * @param reg	parent region of zone
 * @param ref	true if needed to establish references from bodies to the zone
 *		set to false for temporary zones
 */
GZone::GZone(GRegion* reg, const bool ref)
	: _id(-1),
	  _format(Format::PRODUCT),
	  _flags(0),
	  _hash(0),
	  _generation(0),
	 _bboxState(BBox::State::Invalid),
#ifdef USE_OBBOX
	 _obboxState(BBox::State::Invalid),
#endif
	region(reg)
{
	if (ref) _flags |= Flags::BodyRef;
} // GZone

/** @return regions name */
std::string GZone::name() const
{
	return region?region->name() : "<null>";
} // name

/** clear */
void GZone::clear(bool fast)
{
	if (!fast) {
		removeAllReferences();
		_format = Format::PRODUCT;
		_hash   = 0;
		_flags &= ~Flags::BodyRef;	// all except body ref
		resetBBox();
	}
	expr.clear();
} // clear

/** add body reference to this zone */
void GZone::addReference(GBody* body)
{
	if (body->isRegion()) _flags |= Flags::RegBody;
	if (not hasBodyRef()) return;
#ifdef ZONES_AS_SET
	if (body->isBody())
#else
	if (body->isBody() && !body->hasZone(this))
#endif
		body->addZone(this);
} // addReference

/** add all zones-bodies references to this zone */
void GZone::addAllReferences()
{
	if (not hasBodyRef()) return;
	for (int i=0; i<size(); i++) {
		GBody* body = expr[i];
#ifdef ZONES_AS_SET
		if (body->isBody())
#else
		if (body->isBody() && !body->hasZone(this))
#endif
			body->addZone(this);
	}
} // addAllReferences

/** remove all body references to this zone as parent */
void GZone::removeAllReferences()
{
	if (not hasBodyRef()) return;
	for (int i=0; i<size(); i++) {
		GBody* body = expr[i];
		if (body->isBody())
			body->delZone(this);
	}
} // removeAllReferences

/** add to the expression one body
 * The expression can either be in Reverse Polish Notation (RPN) mode or
 * STanDard mode with unions of intersection/subtractions
 * @param body		body pointer to add
 */
bool GZone::add(GBody* body)
{
	assert(body!=nullptr);
	_hash = 0;
	resetBBox();
	addReference(body);
	expr.push_back(body);
	//body->nextGeneration();
	return true;
} // add

/** replace the body at location i
 * @param i	location to insert the body (i<=0
 * @param body	new body to set
 */
void GZone::set(unsigned i, GBody* body)
{
	_hash = 0;
	//expr[i]->delZone(this);	// not correct if double defined in the zone
	expr[i] = body;
	addReference(body);
	resetBBox();
} // insert

/** insert a body at location i
 * @param i	location to insert the body (i<=0
 */
void GZone::insert(unsigned i, GBody* body)
{
	_hash = 0;
	expr.insert(i, body);
	addReference(body);
	resetBBox();
} // insert

/** add to expression one token
 * The expression can any format
 * @param token		token to be added
 * @return		false in case of error, true otherwise
 */
bool GZone::add(const char token)
{
	_hash = 0;
	resetBBox();
	switch (token) {
		case '$':
			assert(false);
			assert(std());
			expr.push_back(&GBody::tminus);
			break;
		case '|':
		case ':':
			assert(!std());
			expr.push_back(&GBody::tunion);
			break;
		case '+':
			assert(!std());
			// Ignore + on product
			if (!std())
				expr.push_back(&GBody::tplus);
			break;
		case '-':
			expr.push_back(&GBody::tminus);
			break;
		//case '#':
		//	expr.push_back(&GBody::tnot);
		//	break;
		case '*':
			assert(!std());
			expr.push_back(&GBody::tnil);
			break;
		case '@':
			assert(!std());
			expr.push_back(&GBody::tuniverse);
			break;
		case '(':
			assert(normal());
			expr.push_back(&GBody::tleft);
			break;
		case ')':
			assert(normal());
			expr.push_back(&GBody::tright);
			break;
		default:
			return false;
	}
	return true;
} // add

/** addPlus */
bool GZone::addPlus(GBody* body)
{
	if (!std()) return false;
	int negative = minusIndex();
	if (negative<size())
		expr.insert(negative, body);
	else
		expr.push_back(body);
	addReference(body);
	return true;
} // addPlus

/** addMinus */
bool GZone::addMinus(GBody* body)
{
	if (!std()) return false;
	if (find(&GBody::tminus)<0)
		expr.push_back(&GBody::tminus);
	expr.push_back(body);
	addReference(body);
	return true;
} // addMinus

/** copy zone */
void GZone::copy(const GZone& zone)
{
	_format     = zone._format;
	_flags      = zone._flags;
	_bboxState  = zone._bboxState;
	_bbox       = zone._bbox;
#ifdef USE_OBBOX
	_obboxState = zone._obboxState;
	_obbox      = zone._obbox;
#endif
	skip        = zone.skip;
	expr.resize(zone.expr.size());
	for (int i=0; i<zone.expr.size(); i++)
		expr.push_back(zone.expr[i]);
} // copy

/** compare if 'this' is a subset or equal to 'other' zone
 * provided they have the same format and only under some conditions
 * @param other		superset zone
 * @param ignore	ignore body in search
 * @return true if 'this' is a subset or equal to 'other' zone
 */
bool GZone::isSubsetOf(GZone& other, const GBody* ignore) const
{
	if (empty() or other.empty()) return false;
	if (format() != other.format()) return false;
	if (std()) {
		// WARNING: compare 'other' against 'this'
		// Accept as match
		//   this= +A -B +C   vs   other= +A -B
		// where 'this' is a subset of 'other'
		//
		// Check that for every body B of 'other' (superset) either:
		//  1. exists on 'this' with the same sign
		//                      with opposite sign return false
		//  2. there is a body A where the location is
		//             AinB with +A and +B
		//             BinA with -A and -B
		//          or OUTSIDE with +A and -B
		bool plusB = true;
		for (int i=0; i<other.size(); i++) {
			GBody* B = other[i];
			if (B==&GBody::tminus) {
				plusB = false;
				continue;
			}
			if (B==ignore) continue;
			bool found = false;
			bool plusA = true;
			for (int j=0; j<size(); j++) {
				GBody* A = expr[j];
				if (A == &GBody::tminus) {
					plusA = false;
					continue;
				}
				if (A==ignore) continue;
				if (A==B) {
					if (plusA != plusB) return false;
					found = true;
					break;
				} else {
					auto loc = A->locationWrt(B);
					if ((loc==GBody::Location::AinB and plusA and plusB) or
					    (loc==GBody::Location::BinA and !plusA and !plusB) or
					    (loc==GBody::Location::OUTSIDE and plusA and !plusB)) {
						found = true;
						break;
					}
				}
			}
			if (!found) return false;
		}
		return true;
	} else {
		if (size() != other.size()) return false;
		for (int i=0; i<size(); i++)
			if (expr[i]!=ignore and expr[i]!=other[i])
				return false;
	}
	return true;
} // isSubsetOf

/** check if this and zone can be merged or simplified.
 * zone should have a body that appears as + in one and -
 * in the other. The remaining bodies should be a subset
 * of one from the other.
 *
 * WARNING: works only on std().
 */
bool GZone::merge(GZone& zone)
{
	if (empty() or zone.empty()) return false;
	if (format() != zone.format()) return false;
	assert(std());
	assert(zone.std());

	// loop all bodies from 'this' to find a
	// common body with opposite operator in 'zone'
	int Anegative = minusIndex();
	int Bnegative = zone.minusIndex();

	for (int ia=0; ia<size(); ia++) {
		GBody* body = expr[ia];
		if (body==&GBody::tminus) continue;
		// Search for body in 'zone'
		int ib = zone.find(body);
		if (ib<0) continue;
		if ((ia<Anegative and ib>Bnegative) or
		    (ia>Anegative and ib<Bnegative)) {
			// body exists with opposite sign on zones
			// Check for potential sub-zone of 'this' against 'zone
			// or vice versa
			bool delA = isSubsetOf(zone,body);
			bool delB = zone.isSubsetOf(*this,body);
			if (delA) {
				//cout << "+++ A delete body: " << body->name() << endl;
				//cout << "<<< A: " << *this << endl;
				//cout << "<<< B: " <<  zone << endl;
				body->delZone(this);
				expr[ia] = nullptr;
				compact();
				//cout << ">>> A: " << *this << endl;
			}
			if (delB) {
				//cout << "+++ B delete body: " << body->name() << endl;
				//cout << "<<< A: " << *this << endl;
				//cout << "<<< B: " <<  zone << endl;
				body->delZone(&zone);
				zone.expr[ib] = nullptr;
				zone.compact();
				//cout << ">>> B: " <<  zone << endl;
			}
			if (delA or delB) return true;
		}
	}
	return false;
} // merge

/** @return hash value of expression
 * WARNING it has to be the same as in Viewer_zone("has",...) python binding
 */
dword GZone::hash() const
{
	if (_hash>0) return _hash;
	_hash = 0;
	for (int i=0; i<size(); i++) {
		GBody const* body = expr[i];
		_hash += (_hash<<5) + i;
		_hash += (_hash<<5) + body->hash();
	}
	return _hash;
} // hash

/** setup */
void GZone::setup(const BBox* bblimit)
{
	updateBBox(bblimit);
	if (rpn()) makeSkip();
} // setup

/** normal */
Vector GZone::normal(const double x, const double y, const double z,
		     const GBody* onbody) const
{
	if (onbody==nullptr) {
		// find the closest body
		onbody = nullptr;
		double dmin = INFINITE;
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (!body->isBody()) continue;
			double q = body->distance(x,y,z);
			if (q<dmin) {
				onbody = body;
				dmin = q;
			}
		}
	}
	assert(onbody != nullptr);	// FIXME do I need it
	Vector n = onbody->normal(x,y,z);
	if (std()) {
		// Search for the sign of the body
		int i=0;
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body==&GBody::tminus) break;
			if (body==onbody) return n;
		}
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body==onbody) return -n;
		}
		assert(false);
	} else
	if (rpn()) {
		Stack<bool, ZONE_STACK_SIZE> stack;
		bool invert = false;
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (body==&GBody::tplus || body==&GBody::tunion)
				if (stack.peek()) {
					stack.pop();
					stack.set(true);
				} else
					stack.pop();
			else
			if (body == &GBody::tminus) {
				if (stack.peek()) {
					invert = not invert;
					stack.pop();
					stack.set(true);
				} else
					stack.pop();
			} else
			//if (body == &GBody::tnot) {
			//	invert = not invert;
			//	stack.set(true);
			//} else
			if (body == &GBody::tuniverse)
				stack.push(false);
			else
				stack.push(body==onbody);
		}
		if (invert) return -n;
	}
	return n;
} // normal

/** test a point for being inside a zone.
 *  We are dealing with 3 values in/on-surface/out.
 *
 *           2 state in/out table
 *   #    A      B     A+B    A-B   A|B
 * ==== =====  =====  =====  ===== =====
 *   1   in     in     in     out   in
 *   3   in     out    out    in    in
 *
 *   7   out    in     out    out   in
 *   9   out    out    out    out   out
 *
 *    Bool equivalent  A&B   A&(!B) A|B
 *
 *
 *          3 state in/on/out table
 *   #    A      B     A+B    A-B   A|B
 * ==== =====  =====  =====  ===== =====             +-----8----+
 *   1   in     in     in     out   in               |B         |
 *   2   in     on     on     on    in        +--6---+----+     |
 *   3   in     out    out    in    in        |A     |  1 |     |
 *                                            | 3    2    4     |
 *   4   on     in     on     out   in        |A-B   |A+B |   7 |
 *   5   on     on     on     on    on        |A|B   |A|B |     |
 *   6   on     out    out    on    on        +------5----+     |
 *                                                   |A|B       |
 *   7   out    in     out    out   in        9      +----------+
 *   8   out    on     out    out   on
 *   9   out    out    out    out   out
 *
 *   if we combine io=in+on we have the following table
 *   #    A      B     A+B    A-B   A|B
 * ==== =====  =====  =====  ===== =====
 * 1245  io     io     io    on/out io
 *  36   io     out    out    io    io
 *
 *  78   out    io     out    out   io
 *   9   out    out    out    out   out
 *                    A&B          A|B
 *
 *  Logic:  (trying to sort it by number of combinations)
 *      A+B       if any is out    -> out
 *           else if any is on     -> on     (on behaves like restricted out)
 *           else -> in
 *
 *      A-B       if A=out | B=in  -> out
 *           else if A=in  & B=out -> in     (out of sort order)
 *           else -> on
 *
 *      A|B       if any  is  in   -> in
 *           else if any  is  on   -> on
 *           else -> out
 */
GBody::Side GZone::side(const double x, const double y, const double z,
			  const double dsurf, const GBody** onbody) const
{
	if (size()==0) return GBody::Side::Out;
	VERBOSE(1, cout << endl << "GZone: " << *this
			<< " pos= " << x << " " << y << " " << z
			<< " dsurf= " << dsurf << endl);

	*onbody = nullptr;
	if (std()) {
		// assume we are starting from the surface of the universe
		GBody::Side A = GBody::Side::In;

		// first A+B+C... plus terms (if any)
		int i=0;
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body==&GBody::tminus) break;
			GBody::Side B = body->side(x,y,z, dsurf);
			VERBOSE(1,cout << "   + " << body->name()
				     << " in=" << (int)B << endl);
			if (B==GBody::Side::Out) return GBody::Side::Out;
			if (B==GBody::Side::Surface) {
				*onbody = body;
				A = GBody::Side::Surface;
			}
			// for GBody::Side::In it remains either Inside or on Surface
		}
		// then -A-B-C... negative terms (if any)
		// here we come only for conditions 1,2,3,4,5,6 (anything in/on of A)
		// conditions 7,8,9 already excluded from the first passage
		// even in the case that there is no + term we do have the +@
		while (i<size()) {
			const GBody* body = expr[i++];
			GBody::Side B = body->side(x,y,z, dsurf);
			VERBOSE(1,cout << "   - " << body->name()
				   << " in=" << (int)B << endl);
			if (B==GBody::Side::In)  return GBody::Side::Out;
			if (B==GBody::Side::Surface) {
				*onbody = body;
				A = GBody::Side::Surface;
			}
		}
		VERBOSE(1,cout << "   | prod=" << (int)A << endl);
		return A;
	} else
	if (rpn()) {
		Stack<GBody::Side, ZONE_STACK_SIZE> stack;
		GBody::Side B;
		//GBody::Side stack[ZONE_STACK_SIZE], B;
		//int  sp = -1;		// stack pointer
		// FIXME Check VZone::inside() for a more optimized version
		// NULL if already pushed to stack, else needs evaluation of inside
		//CBody* bstack[ZONE_STACK_SIZE];
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (body == &GBody::tplus) {
				B = stack.pop();
				if (B==GBody::Side::Out || stack()==GBody::Side::Out)
					stack.set(GBody::Side::Out);
				else
				if (B==GBody::Side::Surface || stack()==GBody::Side::Surface)
					stack.set(GBody::Side::Surface);
				else
					stack.set(GBody::Side::In);
				VERBOSE(1,cout << "  +  " << (int)stack() << endl);
			} else
			if (body == &GBody::tminus) {
				B = stack.pop();
				if (B==GBody::Side::In || stack()==GBody::Side::Out)
					stack.set(GBody::Side::Out);
				else
				if (B==GBody::Side::Out && stack()==GBody::Side::In)
					stack.set(GBody::Side::In);
				else
					stack.set(GBody::Side::Surface);
				VERBOSE(1,cout << "  -  " << (int)stack() << endl);
			} else
			if (body == &GBody::tunion) {
				B = stack.pop();
				if (B==GBody::Side::In || stack()==GBody::Side::In)
					stack.set(GBody::Side::In);
				else
				if (B==GBody::Side::Surface || stack()==GBody::Side::Surface)
					stack.set(GBody::Side::Surface);
				else
					stack.set(GBody::Side::Out);
				VERBOSE(1,cout << "  |  " << (int)stack() << endl);
			} else
			if (body == &GBody::tuniverse) {
				stack.push(GBody::Side::In);
				VERBOSE(1,cout << "  @  " << (int)stack() << endl);
			} else {
				stack.push(body->side(x,y,z, dsurf));
				if (stack()==GBody::Side::Surface) *onbody = body;
				VERBOSE(1,cout << name() << "."<<body->name()<<"="<<(int)stack() << ":\t";
					for (int j=0; j<stack.size(); j++)
						cout << (int)stack[j] << ' ';
					cout << endl);
			}
		}
		VERBOSE(1,cout << name() << "=" << (int)stack() << endl << endl);
		assert(stack.one());
		return stack();
	} else {
		assert(0);
		return GBody::Side::Out;
	}
} // side

/** inside */
bool GZone::inside(const double x, const double y, const double z,
		   const double u, const double v, const double w) const
{
	if (size()==0) return false;
	VERBOSE(1, cout << endl << "GZone: " << *this
			<< "\n\tpos= " << x << " " << y << " " << z
			<< "\n\tdir= " << u << " " << v << " " << w << endl);
	if (std()) {
		int  i=0;
		// first plus terms (if any)
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body==&GBody::tminus) break;
			VERBOSE(1,cout << "   + " << body->name()
				     << " in=" << body->inside(x,y,z, u,v,w) << endl);
			if (!body->inside(x,y,z, u,v,w)) return false;
		}
		// then negative terms (if any)
		while (i<size()) {
			const GBody* body = expr[i++];
			VERBOSE(1,cout << "   - " << body->name()
				   << " in=" << body->inside(x,y,z, u,v,w) << endl);
			if (body->inside(x,y,z, u,v,w)) return false;
		}
		VERBOSE(1,cout << "   | prod=1" << endl);
		return true;
	} else { // RPN
		assert(rpn());
#if 1
		assert(skip.size()>0);
		bool result = true;
		int i = 0;
		while (i<expr.size()) {
			const GBody* body = expr[i];
			if (body->isBody())
				result = body->inside(x,y,z, u,v,w);
			else
			if (body == &GBody::tuniverse)
				result = true;
			if (skip[i]==0) {
				result = not result;
				i++;
			} else
			if (result == (skip[i]>0))
				i = abs(skip[i]);
			else
				i++;
		}
		return result;
#else
		Stack<bool, ZONE_STACK_SIZE> stack;
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (body == &GBody::tplus) {
				bool b = stack.pop();
				stack.set(stack() and b);
				VERBOSE(1,cout << "  +  " << stack() << endl);
			} else
			if (body == &GBody::tminus) {
				bool b = stack.pop();
				stack.set(stack() and !b);
				VERBOSE(1,cout << "  -  " << stack() << endl);
			} else
			if (body == &GBody::tnot) {
				assert(false);
				stack.set(not stack());
				VERBOSE(1,cout << "  #  " << stack() << endl);
			} else
			if (body == &GBody::tunion) {
				bool b = stack.pop();
				stack.set(stack() or b);
				VERBOSE(1,cout << "  |  " << stack() << endl);
			} else
			if (body == &GBody::tuniverse) {
				stack.push(true);
				VERBOSE(1,cout << "  @  " << stack() << endl);
			} else {
				stack.push(body->inside(x,y,z, u,v,w));
				VERBOSE(1,cout << name() << "."<<body->name()<<"="<<stack() << ":\t";
					for (int j=0; j<stack.size(); j++) cout << stack[j] << ' ';
					cout << endl);
			}
		}
		VERBOSE(1,cout << name() << "=" << stack() << endl << endl);
		assert(stack.one());
		assert(result == stack());
		return stack();
#endif
	}
} // inside

/** check if location p+v*t is inside the zone or not
 * @param x,y,z  initial position
 * @param u,v,w  direction vector
 * @param t      distance
 * @return true is point is inside the zone
 */
bool GZone::inside(const double x, const double y, const double z,
		   const double u, const double v, const double w,
		   const double t, Cache<CBody3D>& cache) const
{
	if (std()) {
		int  i=0;
		// first plus terms (if any)
		DEBUG(int oldprec = cout.precision(22));
		VERBOSE(1, cout << "*-* GZone::inside:RPN(pos=("
				<< x << ", " << y << ", " << z << "),\n"
				<< "\t\t, dir=("
				<< u << ", " << v << ", " << w << "),\n"
				<< "\t\t, t=" << t << ")" << endl);
		DEBUG(cout.precision(oldprec));
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body == &GBody::tminus) break;
			VERBOSE(1,
				oldprec = cout.precision(22);
				cout << "\t+ " << body->name()
					<< " in=" << body->inside(x,y,z, u,v,w, t, cache)
					<< " " << cache[body->id()].tmin
					<< " .. " << cache[body->id()].tmax
					<< endl;
				cout.precision(oldprec));
			if (!body->inside(x,y,z, u,v,w, t, cache)) return false;
		}
		// then negative terms (if any)
		while (i<size()) {
			const GBody* body = expr[i++];
			VERBOSE(1,
				oldprec = cout.precision(22);
				cout << "\t- " << body->name()
					<< " in=" << body->inside(x,y,z, u,v,w, t, cache)
					<< " " << cache[body->id()].tmin
					<< " .. " << cache[body->id()].tmax
					<< endl;
				cout.precision(oldprec));
			if (body->inside(x,y,z, u,v,w, t, cache)) return false;
		}
		return true;
	} else { // RPN
		assert(rpn());
#if 1
		//bool __debug=false;
		bool result = true;
		int i = 0;
		while (i<expr.size()) {
			const GBody* body = expr[i];
			//if (__debug) cout << i << ' ' << body->name();
			if (body->isBody())
				result = body->inside(x,y,z, u,v,w, t, cache);
			else
			if (body == &GBody::tuniverse)
				result = true;
			//if (__debug) cout << " result=" << result << " skip=" << skip[i];
			if (skip[i]==0) {
				result = not result;
				i++;
			} else
			if (result == (skip[i]>0))
				i = abs(skip[i]);
			else
				i++;
			//if (__debug) cout << " result_new=" << result << " inew=" << i << endl;
		}
		return result;
#else
		Stack<bool, ZONE_STACK_SIZE> stack;
		// e.g.   True  or  ? => True
		//        False and ? => False
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			// a+b = a && b
			if (body == &GBody::tplus) {
				DUMP(std::cout << "\tOp=" << (stack.peek(1) and stack.peek())
					<< "  " << stack.peek(1)
					<< " + " << stack.peek()
					<< std::endl);
				bool b = stack.pop();
				stack.set(stack() and b);
				//stack[sp-1] = stack[sp-1] and stack[sp];
				//sp--;
			} else
			// a-b = a && !b
			if (body == &GBody::tminus) {
				DUMP(std::cout << "\tOp=" << (stack.peek(1) and not stack.peek())
					<< "  " << stack.peek(1)
					<< " - " << stack.peek()
					<< std::endl);
				bool b = stack.pop();
				stack.set(stack() and not b);
				//stack[sp-1] = stack[sp-1] and not stack[sp];
				//sp--;
			} else
			// # a = !a
			//if (body == &GBody::tnot) {
			//	assert(false);
			//	stack.set(not stack());
			//} else
			// a | b
			if (body == &GBody::tunion) {
				DUMP(std::cout << "\tOp=" << (stack.peek(1) or stack.peek())
					<< "  " << stack.peek(1)
					<< " | " << stack.peek()
					<< std::endl);
				bool b = stack.pop();
				stack.set(stack() or b);
				//stack[sp-1] = stack[sp-1] or stack[sp];
				//sp--;
			} else
			if (body == &GBody::tuniverse) {
				DUMP(std::cout << "\tUniverse: true" << std::endl);
				//stack[++sp] = true;
				stack.push(true);
			} else {
				//stack[++sp]  = body->inside(x,y,z, u,v,w,t, cache);
				stack.push(body->inside(x,y,z, u,v,w,t, cache));
				DUMP(std::cout << "\tBody: " << body->name()
						<< " " << stack.peek() << std::endl);
			}
		}
		//if (sp!=0) return false;
		//return stack[0];
		if (stack.ptr()!=0) return false;
		return stack();
#endif
	}
} // inside

/** @return minimum distance in any direction to find a surface
 * in the zone
 */
double GZone::distance(const double x, const double y, const double z) const
{
	double dmin = INFINITE;
	for (int i=0; i<size(); i++) {
		const GBody* body = expr[i];
		if (!body->isBody()) continue;
		double q = body->distance(x,y,z);
		if (q<dmin) dmin = q;
	}
	return dmin;
} // distance

/** return minimum distance enter or exit the zone.
 * The smallest distance starting from tmin that enters/exits the zone.
 * @param x,y,z		location of ray
 * @param u,v,w		normalized direction of ray
 * @param tmin		I/O minimum intersection point
 * @param tmax		maximum to scan
 * @param toin		true when entering a zone, false to exit
 * @return body that an intersection in found value with the minimum (*tmin)
 */
GBody* GZone::distance(	const double x, const double y, const double z,
			const double u, const double v, const double w,
			double*   tmin, const double tmax,
			const bool toin, Cache<CBody3D>& cache) const
{
	DEBUG(int oldprec = cout.precision(22));
	VERBOSE(1, cout << "*-* GZone::distance(pos=(" << x << ", " << y << ", " << z << "),\n"
				 << "\t\t dir=(" << u << ", " << v << ", " << w << "),\n"
				 << "\t\t tmin=" << *tmin << ", tmax=" << tmax
				 << ", toin=" << toin
				 << ")" << endl);
#if 1
	// on first pass populate zoneAccel
	zoneAccel.clear();
	assert(zoneAccel.capacity()>0);
	for (int i=0; i<size(); i++) {
		GBody* body = expr[i];
		if (!body->isBody()) continue;
		double ta, tb;
		DEBUG(bool inv =)
		body->distance(x,y,z, u,v,w, &ta, &tb, cache);
		if (ta<tb) {
			VERBOSE(1, cout << "\t>>> " << body->name() << ' ');
#if _DEBUG>0
			if (GBody::verbose>1) {
				if (inv)
					cout << "-inf.." << ta << " or " << tb << "..+inf";
				else
					cout << ta << " .. " << tb;
				cout << endl;
			}
#endif
			if (InRangeOpen(*tmin, ta, tmax))
				zoneAccel.emplace_back(GBodyOrderAccel(body, ta));
			if (InRangeOpen(*tmin, tb, tmax))
				zoneAccel.emplace_back(GBodyOrderAccel(body, tb));
		}
	}
	// perform as many loops as zoneAccel elements
	for (unsigned iloop=0; iloop<zoneAccel.size(); iloop++) {
		GBody* tbody = nullptr;
		double t = Min(tmax, SINFINITE);
		DUMP(cout << "\t+++ range=" << *tmin << " .. " << tmax << endl);
		//unsigned it = 0;
		//for (unsigned i=iloop; i<zoneAccel.size(); i++) {
		for (unsigned i=0; i<zoneAccel.size(); i++) {
			if (InRangeOpen(*tmin, zoneAccel[i].t, t)) {
				//it = i;
				t = zoneAccel[i].t * (1.0+SMALL);
				tbody = zoneAccel[i].body;
				VERBOSE(1, cout << " --> " << t);
				VERBOSE(1, cout << " p=(" << x+u*t);
				VERBOSE(1, cout << ",  " << y+v*t);
				VERBOSE(1, cout << ",  " << z+w*t);
				VERBOSE(1, cout << ")" << endl);
			}
		}
#else
	// Loop for all solutions (2 assuming we have 2 per body)
	for (int j=size()*2; j--; ) {
		GBody* tbody = nullptr;
		double t = Min(tmax, SINFINITE);
		DUMP(cout << "\t+++ range=" << *tmin << " .. " << tmax << endl);
		for (int i=0; i<size(); i++) {
			GBody* body = expr[i];
			if (!body->isBody()) continue;
			double a, b;
			DEBUG(bool inv =)
			body->distance(x,y,z, u,v,w, &a, &b, cache);
			if (a<b) {
				VERBOSE(1, cout << "\t>>> " << body->name() << ' ');
#if _DEBUG>0
				if (GBody::verbose>1) {
					if (inv)
						cout << "-inf.." << a << " or " << b << "..+inf  t=" << t;
					else
						cout << a << " .. " << b << "  t=" << t;
				}
#endif
				if (*tmin<a and a<=t) {
					t = a * (1.0+SMALL);
					tbody = body;
				} else
				if (*tmin<b and b<=t) {
					t = b * (1.0+SMALL);
					tbody = body;
				}
				VERBOSE(1, cout << " --> " << t);
				VERBOSE(1, cout << " p=(" << x+u*t);
				VERBOSE(1, cout << ",  " << y+v*t);
				VERBOSE(1, cout << ",  " << z+w*t);
				VERBOSE(1, cout << ")" << endl);
			}
		}
#endif
		if (tbody) {
			*tmin = t;
			bool in = inside(x,y,z, u,v,w, t, cache);
			VERBOSE(1,cout << "\t>>> Test t=" << t << " inside=" << in << endl);
			if (toin==in) {
				VERBOSE(1,cout << "\t\tSuccess return body: " << tbody->name() << endl);
				// Improve precision by checking the position at the tmin location
				double acc = (Abs(x)+Abs(y)+Abs(z)+Abs(*tmin))*SMALL;
				bool more;
//				*tmin += acc;
//				*tmin *= (1.0+acc);
				double tinitial = *tmin;
				double tinc = 0.0;
#if _DEBUG>0
				static thread_local int maxloop = 5;
				int loop = 0;
#endif
//				for (int i=10; i-- and more; ) {
				do {
					DEBUG(loop++);
					more = false;
					double xn = *tmin*u+x;
					double yn = *tmin*v+y;
					double zn = *tmin*w+z;
					double a, b;
					tbody->distance(xn,yn,zn, u,v,w, &a, &b);
#if _DEBUG>0
					if (GBody::verbose>1) {
						cout	<< "\tloop=" << loop
							<< "  t=" << *tmin
							<< "  pn=" << xn
							<< "," << yn
							<< "," << zn
							<< endl;
					}
					VERBOSE(1, cout << "\ta=" << a << " b=" << b << endl);
#endif
					if (Abs(a) < Abs(b)) {
						if (0<a and a<8*acc) {
							tinc += 2*a + acc;
							*tmin = tinitial + tinc;
							more = true;
						}
					} else {
						if (0<b and b<8*acc) {
							tinc += 2*b + acc;
							*tmin = tinitial + tinc;
							more = true;
						}
					}
					acc *= 8;
#if _DEBUG>0
					if (loop > maxloop) {
						maxloop = loop;
							cerr << ">>>>>>> GZone::distance() maxloop= " << maxloop << endl;
							assert(loop<10);
						}
#endif
				} while (more);
//				assert(inside(fma(u,*tmin,x), fma(v,*tmin,y), fma(w,*tmin,z),
//						u,v,w, 0., nullptr) == in);
				DEBUG(cout.precision(oldprec));
				*tmin = Min(*tmin, tmax);
				return tbody;
			}
		} else
			break;
	}
	VERBOSE(1,cout << "\t>>> Failed" << endl);
	*tmin = tmax;
	DEBUG(cout.precision(oldprec));
	return nullptr;
} // distance

/** optimize the zone's expression only in Product format
 *
 * @return true if any optimization took place
 */
//#define _DUMP
bool GZone::optimize()
{
	if (!std()) return false;

	// Remove all references
	removeAllReferences();

	// 1st find negative terms starting index
	int negative = minusIndex()+1;

#ifdef _DUMP
	cout << "Zone::optimize\nIn= ";
	for (int i=0; i<size(); i++)
		if (expr[i]->isBody())
			cout << (i<negative?" +":" -") << expr[i]->name() << '[' << expr[i]->typeStr() << ']';
		else
			cout << " [" << expr[i]->name() << ']';
	cout << endl;
#endif
	// Check bounding box and use for better location estimation
	updateBBox(nullptr,true);
	if (!bbox().isValid()) {
		expr.clear();
		return true;
	}

	// 2nd remove all duplicate terms:
	//    +A +A -> +A
	//    +A -A ->  0
	//    -A -A -> -A
	//    +* = null
	//    -* = ignore=NULL ptr
	//    +@ = ignore=NULL ptr
	//    -@ = 0
	for (int i=0; i<size(); i++) {
		GBody const* body = expr[i];
		if (body == &GBody::tminus) continue;		// skip
		else
		if (body == &GBody::tnil) {
			if (i<negative) {
				expr.clear();			// +* = 0
				return true;
			} else
				expr[i] = NULL;			// -* = ignore
		} else
		if (body == &GBody::tuniverse) {
			if (i<negative) {
				expr[i] = NULL;			// +@ = ignore
			} else
			if (i>=negative) {
				expr.clear();			// -@ -> 0
				return true;
			}
		}
		for (int j=i+1; j<size(); j++) {
			if (expr[j]==NULL) continue;
			if (body != expr[j]) continue;
			if (i<negative && j<negative)
				expr[j] = NULL;			// +A +A -> +A
			else
			if (i>=negative && j>=negative)
				expr[j] = NULL;			// -A -A -> -A
			else {
				expr.clear();			// +A -A -> 0
				return true;
			}
		}
	}

	// Geometrical optimization
	for (int i=0; i<size(); i++) {
		GBody const* A = expr[i];
		if (A==NULL or !A->isBody()) continue;
		bool plusA = i<negative;
		for (int j=i+1; j<size(); j++) {
			GBody const* B = expr[j];
			if (B==NULL or !B->isBody()) continue;
			bool plusB = j<negative;

			// Find location of A vs B
			auto loc = A->bbLocationWrt(B);			// Check first bounding boxes
			if (loc != GBody::Location::OUTSIDE)		// if overlap then check better
				loc = A->locationWrt(B);
#ifdef _DUMP
			cout <<"A="<<A->name()<<" B="<<B->name() << " loc=" << (int)loc << endl;
			cout << "\tAbb="<<A->bbox() << endl;
			cout << "\tBbb="<<B->bbox() << endl;
#endif
			switch (loc) {
				case GBody::Location::OUTSIDE:
					// +A +B = 0
					// +A -B = +A
					// -A +B = +B
					// -A -B = same
					if (plusA) {
						if (plusB) {	// +A +B -> 0
							expr.clear();
							DUMP(cout << "Out= -none-" << endl);
							return true;
						} else		// +A -B -> +A
							expr[j] = NULL;
					} else
						if (plusB)	// -A +B -> +B
							expr[i] = NULL;
					break;

				case GBody::Location::AinB:
					// +A +B = +A
					// +A -B = 0
					// -A +B = same
					// -A -B = -B
					if (plusA) {
						if (plusB)	// +A +B -> +A
							expr[j] = NULL;
						else {		// +A -B -> 0
							expr.clear();
							DUMP(cout << "Out= -none-" << endl);
							return true;
						}
					} else
						if (!plusB)	// -A -B -> -B
							expr[i] = NULL;
					break;

				case GBody::Location::BinA:
					// +A +B = +B
					// +A -B = same
					// -A +B = 0
					// -A -B = -A
					if (plusA) {
						if (plusB)	// +A +B -> +B
							expr[i] = NULL;
					} else {
						if (plusB) {	// -A +B -> 0
							expr.clear();
							DUMP(cout << "Out= -none-" << endl);
							return true;
						} else		// -A -B -> -A
							expr[j] = NULL;
					}
					break;

				default: /* do nothing */;
			}
			// If A==NULL go to next
			if (expr[i]==NULL) break;
		}
	}

#ifdef _DUMP
	cout << "Mid=";
	for (int i=0; i<size(); i++)
		if (expr[i]) {
			if (expr[i]->isBody())
				cout << (i<negative?" +":" -") << expr[i]->name() << '[' << expr[i]->typeStr() << ']';
			else
				cout << " [" << expr[i]->name() << ']';
		}
	cout << endl;
#endif

	// Finally: fix by removing nulls
	bool changed = compact();
	addAllReferences();		// restore references

#ifdef _DUMP
	cout << "Out=";
	char sign='+';
	for (int i=0; i<size(); i++)
		if (expr[i]->isBody())
			cout << ' ' << sign << expr[i]->name() << '[' << expr[i]->typeStr() << ']';
		else {
			sign = expr[i]->name()[0];
			cout << " [" << expr[i]->name() << ']';
		}
	cout << endl;
#endif
	return changed;
} // optimize

/* compact zone by removing null pointers */
bool GZone::compact()
{
	// removing nulls pointers from expression
	bool changed = false;
	int pos = 0;
	int negative = minusIndex();
	for (int i=0; i<negative; i++) {
		GBody* body = expr[i];
		if (body and body != &GBody::tuniverse)
			expr[pos++] = body;
		else
			changed = true;
	}
	if (negative<size()-1) {
		expr[pos++] = &GBody::tminus;
		for (int i=negative+1; i<size(); i++) {
			GBody* body = expr[i];
			if (body and body != &GBody::tnil)
				expr[pos++] = body;
			else
				changed = true;
		}
	}
	expr.erase(pos, size());	// Fix size
	return changed;
} // compact

/* optimize possible products inside an RPN expression in the range [from,to) */
bool GZone::optimizeProducts(int from, int to)
{
	assert(rpn());
	assert(not isCompressed());
	DUMP(cout << "optimizeProducts: " << from << " .. " << to << endl);
	DUMP(cout << "zone: " << *this << endl);

	// Product target should be in the form
	// ... t1 t2 [+-] [ti [+-]]* ...
	//     i               iend
	bool optimized = false;
	int istart = from;
	if (to<0) to = size();
	while (istart < to-1) {
		if (expr[istart]->isOperator()) {
			istart++;
			continue;
		}
		assert(expr[istart]->isAnyBody());
		// istart is pointing to the first body t1
		int iend = istart+1;
		// Find as many repetition of [ti [#] [+-]]...
		while (iend < to-1)
			if (expr[iend]->isAnyBody()) {
				// term [+-]
				if (expr[iend+1]->isIntersection())
					iend += 2;
				//else
				//// term # [+-]
				//if (iend<to-2 and expr[iend+1] == &GBody::tnot and
				//    expr[iend+2]->isIntersection())
				//	iend += 3;
				else
					break;
			} else
				break;

		// Product found
		if (iend != istart+1) {
#ifdef _DUMP
			cout << "Product\n\tRange= [" << istart << ".." << iend << "]/to="
				<< to << ":" << size() << endl
				<< "\tExpr:";
			for (int j=istart; j<iend; j++)
				cout << ' ' << expr[j]->name();
			cout << endl;
#endif
			GZone product(nullptr,false);
			product.format(Format::PRODUCT);
			// add first + term (Universe? @)
			if (expr[istart] != &GBody::tuniverse)
				product.add(expr[istart]);
			for (int j=istart+1; j<iend-1; j++)
				if (expr[j+1]==&GBody::tplus)
					product.add(expr[j]);
			product.add(&GBody::tminus);
			for (int j=istart+1; j<iend-1; j++)
				if (expr[j+1]==&GBody::tminus)
					product.add(expr[j]);
			DUMP(cout << "\tProduct:" << product << endl);

			if (product.optimize()) {
				// Replace expression with optimized one!
				DUMP(cout << "\tOptimized:" << product << endl);
				bool plus=true;
				bool first=true;
				int ii = istart;
				if (product.size()==0)
					expr[ii++] = &GBody::tnil;
				else
					for (int j=0; j<product.size(); j++) {
						GBody* body = product[j];
						if (body==&GBody::tminus)
							plus = false;
						else
						if (first) {
							// (plus)  A $  ->  A
							// (!plus) $ A  ->  @ A -
							if (!plus) expr[ii++] = &GBody::tuniverse;
							expr[ii++] = body;
							if (!plus) expr[ii++] = &GBody::tminus;
							first = false;
						} else {
							expr[ii++] = body;
							expr[ii++] = plus? &GBody::tplus : &GBody::tminus;
						}
					}
				// delete remaining terms from ii-iend
				optimized = true;
				assert(ii<=iend);
				expr.erase(ii,iend);
				DUMP(cout << "Expr:" << *this << endl);
				if (product.size()==0 and ii<size()) {
					int dummy;
					int oldSize = size();
					if (rpnrule(ii, &dummy, false, true))
						to += size() - oldSize;
					DUMP(cout << "Exp2:" << *this << endl);
				}
				to -= iend-ii;
				istart = ii;	// istart at the end of the new product
#if _DEBUG>1
				assert(verify());
#endif
				continue;
			}
		}
		istart = iend;
	}
	if (size()==1 and expr[0]==&GBody::tnil)
		expr.clear();

#if _DEBUG>1
	assert(verify());
#endif
	return optimized;
} // optimizeProducts
//#undef _DUMP

/** @return end pointer for zone starting from 'from' else -1 */
int GZone::findZone(int from)
{
	assert(normal());
	int nest = 0;
	if (from>=size()) return -1;
	if (expr[from] == &GBody::tunion) from++;
	for (int i=from; i<size(); i++) {
		const GBody* body = expr[i];
		if (body == &GBody::tunion and nest==0)
			return i;
		if (body == &GBody::tleft)
			++nest;
		else
		if (body == &GBody::tright) {
			if (--nest < 0)
				throw std::string("Too many closing parenthesis");
		}
	}
	return size();
} // findZone

/** subZone */
void GZone::subZone(GZone* subzone, int from, int to)
{
	assert(normal());

	// check the possible format of the sub-zone
	Format zf = Format::PRODUCT;
	for (int i=from; i<to; i++)
		if (expr[i]==&GBody::tleft  or
		    expr[i]==&GBody::tright or
		    expr[i]->isRegion()) {
			zf = Format::NORMAL;
			break;
		}

	subzone->format(zf);
	if (zf == Format::PRODUCT) {
		// For standard product zone format:
		// 1st push the plus terms
		bool plus = true;
		for (int i=from; i<to; i++) {
			GBody* body = expr[i];
			assert(body != &GBody::tunion);
			//if (body == &GBody::tunion)
			//	continue;
			//else
			if (body == &GBody::tplus)
				plus = true;
			else
			if (body == &GBody::tminus)
				plus = false;
			else
			if (plus) {
				assert(body->isBody());
				*subzone << body;
			}
		}

		// 2nd push the minus terms if any
		plus = true;
		bool tminus = false;	// is tminus pushed?
		for (int i=from; i<to; i++) {
			GBody* body = expr[i];
			if (body == &GBody::tunion)
				continue;
			else
			if (body == &GBody::tplus)
				plus = true;
			else
			if (body == &GBody::tminus)
				plus = false;
			else
			if (!plus) {
				assert(not body->isOperator());
				if (!tminus) {
					*subzone << &GBody::tminus;
					tminus = true;
				}
				*subzone << body;
			}
		}
	} else {
		// For a normal expression format
		// 1st copy the expression as such
		for (int i=from; i<to; i++)
			*subzone << expr[i];

		// 2nd convert to RPN
		subzone->normal2rpn();		// FIXME Do I need here?
	}
} // subZone

/** return token input and output
 * @param body  body to find priority
 * @param ip(O) input priority
 * @return output priority
 */
//               Operator In Out
//priorities = {"(" : [ 99,  0],
//              "|" : [  1,  2],
//              ":" : [  1,  2],
//              "+" : [  3,  4],
//              "-" : [  3,  4],
//              ")" : [  0, 99],
//              " " : [101,100]}
inline int GZone::priority(const GBody* body, int* ip)
{
	// Find priorities
	if (body==&GBody::tleft) {
		if (ip) *ip = 99;
		return 0;
	} else
	if (body==&GBody::tunion) {
		if (ip) *ip =  1;
		return 2;
	} else
	if (body==&GBody::tplus || body==&GBody::tminus) {
		if (ip) *ip =  3;
		return 4;
	} else
	if (body==&GBody::tright) {
		if (ip) *ip =  0;
		return 99;
	} else {
		if (ip) *ip = 101;
		return 100;
	}
} // priority

/** @return depth of the rpn expression */
bool GZone::depth()
{
	assert(rpn());
	int d = 0;
	for (int i=0; i<size(); i++) {
		if (expr[i]->isOperator())
			d--;
		else
			d++;
	}
	return d==1;
} // depth

/* verify the expression if it is syntactically correct */
bool GZone::verify() const
{
	if (std()) {
		int i=0;
		while (i<size()) {
			const GBody* body = expr[i++];
			if (body == &GBody::tminus) break;
			if (!body->isBody()) return false;
		}
		while (i<size()) {
			const GBody* body = expr[i++];
			if (!body->isBody()) return false;
		}
		return true;
	} else
	if (rpn()) {
		int sp = 0;	// stack pointer
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			//if (body == &GBody::tnot) {
			//	if (sp<1) return false;
			//} else
			if (body->isOperator()) {
				if (sp<2) return false;
				sp--;
			} else
			if (body->isAnyBody()) {
				if (++sp >= ZONE_STACK_SIZE)
					return false;
			} else
				return false;
		}
		return sp==1;
	}
	if (normal()) {
		const GBody* prev = nullptr;
		int nest = 1;
		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (body == &GBody::tleft) {
				if (prev and prev->isBox()) return false;
				nest++;
			} else
			if (body == &GBody::tright) {
				if (prev and prev->isOperator()) return false;
				nest--;
				if (nest<1) return false;
			} else
			if (body == &GBody::tunion) {
				if (prev and prev == &GBody::tunion) return false;
			} else
			if (body->isIntersection()) {
				if (prev and
				    prev->isIntersection()) return false;
			} else {
				assert(body->isBody());
			}
		}
		return nest==1;
	}
	return false;
} // verify

/* ---------------------------------------------------------------------- *
 * normal2rpn
 *
 * Convert the FLUKA Boolean expression to Reverse Polish Notation (RPN)
 * Since the normalization routine does not accept the +/- as signs to
 * objects, the routine is converting the leading - to @- (Universe minus)
 * where the special symbol @ is treated as the universe.
 * Furthermore the leading + is ignored as well as the leading | which is
 * accepted by fluka.
 *
 * WARNING: No check is done for the correctness of the expression apart
 *          from the parenthesis nesting
 *
 * ie.
 *        A+B         -> A B +
 *       (A+B)|C      -> A B + C |
 *       (A|B)|C+D|E  -> A B | C D + | E |
 *       -A           -> @ A -
 *       -(-A)        -> @ @ A - -
 *
 * The routine is using the same array for returning the Reverse Polish
 * expression, since the format is more compact.
 * This is generally true apart one case -A -> @ A -
 *
 * Priorities are treated as:
 *     Operator  Priority         In     Out
 *     --------  --------        ---     ---
 *       |       lower             1       2
 *       +       high              3       4
 *       -       high              3       4
 *       (       higher           99       0
 *       )       higher            0      99
 *       object  highest         101     100
 *
 * Algorithm
 *  Consider the expression as a train moving on a railroad with a
 *  T-shape, where each token is one wagon
 *
 *                      <-  (A|B)|C+D
 *      ------------.   .------------
 *      RPN-End      \ /      Exp-End
 *                    |
 *                   S|
 *                   t|
 *                   a|
 *                   c|
 *                   k|
 *
 *  Each wagon to move from the Exp-End to the RPN-End it has to make
 *  a stop first in the Stack-End. Before entering in the stack, the
 *  priority-IN will be checked against the objects in the stack.
 *  All top-most objects currently present in the stack with
 *  higher priority-OUT will be transferred from the stack to the
 *  RPN-End. Apart from the opening parenthesis ( which is discarded.
 *
 *  Example:
 *  (1)                         A+B|C (1)
 *  (2) A                        +B|C (2)
 *  (3) A                         B|C (3)
 *  (4) A                          |C (4)
 *  (5) A B +                       C (5)
 *  (6) A B + C
 *  (7) A B + C |
 *      ------------.   .------------
 *      RPN-End      \ /      Exp-End
 *                    |
 *                   S|
 *                   t|
 *                   a|
 *                   c|     B   C
 *                   k| A + + | | |
 *                      1 2 3 4 5 6 7
 */
void GZone::normal2rpn()
{
	assert(normal());
#if _DEBUG>1
	assert(verify());
#endif

	bool newproduct = true;
	int m = 0;	// new expression pointer (using the same expr array)
	std::vector<GBody*> stack;

	for (int i=0; i<size(); i++) {
		GBody* body = expr[i];

		// Check for special leading chars
		if (newproduct and (body==&GBody::tunion or body==&GBody::tplus)) {
			newproduct = (body==&GBody::tunion);
			continue;
		}

		if (newproduct and body==&GBody::tminus) {
			// insert universe in i'th position
			expr.insert(i, &GBody::tuniverse);
			body = &GBody::tuniverse;
		}

		newproduct = (body==&GBody::tleft or body==&GBody::tunion);

		// input priority
		int ip;
		priority(body,&ip);

		// Remove from the stack everything with higher priority
		while (!stack.empty() and ip<priority(stack.back())) {
			if (stack.back() != &GBody::tleft) {
				assert(m<i);
				expr[m] = stack.back();
				m++;
			}
			stack.pop_back();
		}

		// Push it into the stack
		if (body != &GBody::tright)
			stack.push_back(body);
		else {
			// Should be an opening parenthesis
			if (stack.empty())
				throw std::string("Unbalanced parenthesis");
			stack.pop_back();
		}
	}

	// Empty Stack
	while (!stack.empty()) {
		if (stack.back() == &GBody::tleft)
			throw std::string("Unbalanced parenthesis");
		if (m >= size())
			expr.push_back(stack.back());
		else
			expr[m] = stack.back();
		stack.pop_back();
		m++;
	}

	// Delete unwanted items
	expr.erase(m, size());

	format(Format::RPN);
	if (not verify())
		throw std::string("Invalid expression. Conversion to rpn failed");
} // normal2rpn

/** convertTo */
void GZone::convertTo(const Format fmt)
{
	if (format()==fmt) return;
	switch (format()) {
		case Format::STD:
			switch (fmt) {
				case Format::RPN:
					assert(false);
					break;
				case Format::NORMAL:
					assert(false);
					break;
				default: ;	// never reaches here but keeps compiler happy
			}
			break;

		case Format::RPN:
		//case Format::POSTFIX:
			switch (fmt) {
				case Format::STD:
					assert(false);
					break;
				case Format::NORMAL:
					assert(false);
					break;
				default: ;	// never reaches here but keeps compiler happy
			}
			break;

		case Format::NORMAL:
		//case Format::INFIX:
			switch (fmt) {
				case Format::STD:
					normal2std();
					break;
				case Format::RPN:
					normal2rpn();
					break;
				default: ;	// never reaches here but keeps compiler happy
			}
			break;

		//case Format::PREFIX:
		//	assert(false);
		//	break;

		default:
			assert(false);
	}
} // convertTo

/** convert a normal expression to product
 * WARNING for the moment it expects an expanded expression as a single product
 */
void GZone::normal2std()
{
	assert(normal());
	GBodyArray plus;
	GBodyArray minus;

	bool opplus = true;	// previous operator is plus

	for (int i=0; i<size(); i++) {
		GBody* body = expr[i];
		if (body == &GBody::tplus)
			opplus = true;
		else
		if (body == &GBody::tminus)
			opplus = false;
		else
		if (body->isBody()) {
			if (opplus)
				plus.push_back(body);
			else
				minus.push_back(body);
		} else
		// ignore NIL and parenthesis only on opening and closing
		if ((body == &GBody::tnil) or
		    (body == &GBody::tleft and i==0) or
		    (body == &GBody::tright and i==size()-1)) {
			NOP;
		} else {
			throw std::string("Invalid body type in expression");
		}
	}
	expr.clear();

	format(Format::STD);
	for (int i=0; i<plus.size(); i++)
		expr.push_back(plus[i]);

	if (!minus.empty()) {
		expr.push_back(&GBody::tminus);
		for (int i=0; i<minus.size(); i++)
			expr.push_back(minus[i]);
	}
} // normal2std

/** ---------------------------------------------------------------------- *
 * rpnorm2normal
 *
 * Convert a NORMALIZED Reverse Polish notation to a standard expression
 * WARNING: The routine expects an expression where for each UNION
 * operator the right-sub-expression is a product while the left can be
 * UNION or a product
 * ---------------------------------------------------------------------- */
void GZone::rpnorm2normal()
{
	assert(rpn());

	GBodyArray newexpr(size());
	GBodyArray plus;
	GBodyArray minus;

	int i = 0;
	int nstack = 0;
	bool endprod = false;
	while (i<size()) {
		GBody* body = expr[i];
		bool lastPlus = true;

		if (body->isOperator())
			nstack--;
		else {
			// First term is always a plus
			// .. peek then next operator to check for sign
			if (plus.empty() || expr[i+1] == &GBody::tplus) {
				assert(expr[i]->isBody());
				plus.push_back(expr[i]);
				lastPlus = true;
			} else
			if (expr[i+1] == &GBody::tminus) {
				assert(expr[i]->isBody());
				minus.push_back(expr[i]);
				lastPlus = false;
			}
			nstack++;
		}

		if (nstack==0)
			endprod = true;
		else
		if (nstack==3) {
			if (lastPlus)
				plus.pop();
			else
				minus.pop();
			i -= 2;
			endprod = true;
		} else
		//if (body == &GBody::tnot) {
		//	assert(false);
		//} else
		if (body==&GBody::tunion) {
			i -= 2;
			endprod = true;
		} else
		if (i==size()-1)
			endprod = true;

		if (endprod) {
			plus.compare(GBody::compare);
			minus.compare(GBody::compare);
			plus.sort();
			minus.sort();
			plus.compare(nullptr);
			minus.compare(nullptr);
			bool something_added = false;
			for (int ii=0; ii<plus.size(); ii++)
				if (plus[ii]->name()[0] != ' ') {
					newexpr.push_back(&GBody::tplus);
					newexpr.push_back(plus[ii]);
					something_added = true;
				}
			for (int ii=0; ii<minus.size(); ii++)
				if (minus[ii]->name()[0] != ' ') {
					newexpr.push_back(&GBody::tminus);
					newexpr.push_back(minus[ii]);
					something_added = true;
				}
			if (something_added)
				newexpr.push_back(&GBody::tunion);
			plus.clear();
			minus.clear();
			nstack  = 0;
			endprod = false;
		}
		i++;
	}

	// Remove last |
	if (!newexpr.empty()) newexpr.pop();

	// Copy back expression
	expr.clear();
	expr.allocate(newexpr.size());
	memcpy(&expr[0], &newexpr[0], sizeof(GBody*)*newexpr.size());

	//# Remove duplicates of products
//	rmDoubles(zones);

#if 0
	// Reconstruct expression
	expr = []
	for plus,minus in zones:
		if len(expr)>0 and expr[-1]!="|": expr.push_back("|")
		//# Fill the new array
		for j in plus:
			expr.push_back("+")
			expr.push_back(j)
		for j in minus:
			expr.push_back("-")
			expr.push_back(j)
#endif
	format(Format::RPN);
//	return expr
} // rpnorm2normal

/** set first time then insert on subsequent calls */
void GZone::_set_or_insert(bool& ins, int pos, GBody* body)
{
	if (ins)
		insert(pos, body);
	else {
		set(pos, body);
		ins = true;
	}
} // _set_or_insert

/** expand #regions from expression */
bool GZone::expandRegions(Geometry* geometry)
{
	//std::cout << "*-* GZone::expandRegions" << std::endl;
	//std::cout << "<<< Zone= " << *this << std::endl;
	assert(format() == Format::RPN);
	if (not hasRegionBody() or isCompressed()) return false;

	// expand all #regions from zone
	for (int ib=0; ib<size(); ib++) {
		GBody* body = expr[ib];
		if (!body->isRegion()) continue;
		bool ins = false;	// use insert or set first body

		// Find region from body
		GRegionBody* regbody = static_cast<GRegionBody*>(body);
		GRegion* reg = regbody->region();
		if (reg == nullptr) {
			reg = geometry->getRegion(regbody->name().substr(1));
			if (reg==nullptr) {
				// FIXME show an error and continue
				// nothing added delete the "#region -"
				expr.erase(ib,ib+2);
				ib--;
				continue;
			}
			regbody->region(reg);
		}
		//std::cout << "inserting region=" << *reg << endl;

		// remember last #region position in the expanded region
		int iregion = -1;
		bool added = false;

		// insert region expression
		for (int iz=0; iz<reg->size(); iz++) {
			GZone* zn = (*reg)[iz];
			if (zn->size()==0)
				continue;
			added = true;		// something will be added
			if (zn->std()) {
				// convert to rpn on the fly
				GBody* op = &GBody::tplus;
				for (int j=0; j<zn->size(); j++) {
					GBody* b = (*zn)[j];
					if (b->isRegion() and iregion<0) iregion = ib;
					if (b==&GBody::tminus) {
						op = b;
						if (j==0)	// only - terms => we need an infinite @
							_set_or_insert(ins, ib++, &GBody::tuniverse);
						continue;
					}
					_set_or_insert(ins, ib++, b);
					if (j>0) insert(ib++, op);
				}
			} else
			if (zn->rpn() or zn->format()==Format::RPN) {
				for (int j=0; j<zn->size(); j++) {
					GBody* b = (*zn)[j];
					if (b->isRegion() and iregion<0) iregion = ib;
					_set_or_insert(ins, ib++, b);
				}
			}
			if (iz>0) insert(ib++, &GBody::tunion);
			//std::cout << "+++ zzz: " << *this << std::endl;
			assert(verify());
		}
		if (!added) {	// if nothing added delete the "#region -"
			expr.erase(ib,ib+2);
			ib--;
		}
		// expand recursively if another #region is present during expansion
		if (iregion>=0) ib = iregion-1;
	}
	format(GZone::Format::RPN);
	_flags &= ~Flags::RegBody;
	//std::cout << ">>> Zone: " << *this << std::endl;
	//for (int iii=0; iii<size(); iii++)
	//	assert(!expr[iii]->isRegion());
	assert(verify());
	return true;
} // expandRegions

/** --------------------------------------------------------------------- *
 * expand
 *
 * Normalize a CG expression given in Reverse Polish Notation.
 * Normalized CG expression is an expression given as sum (Boolean OR) of
 * products (Boolean intersection or subtraction).
 * The normalization (expansion of parenthesis and operator priorities)
 * should be performed by recursively calling the RPNRULE subroutine.
 * Since Fortran-77 doesn't have recursion, call the RPNRULE for every
 * operator starting from the right-most one, until no rule is found.
 *
 * @param limit  expand up to "limit" terms
 * @return true on success
 *
 * WARNING: expression is not using the complement at the end
 * ----------------------------------------------------------------------
 */
bool GZone::expand(unsigned limit/*, const int timelimit*/)
{
	DUMP(cout << "*-* GZone::expand :" << *this << endl);
	assert(rpn());
	if (isCompressed()) return false;

	// Loop until there is no any extra change needed
	// Scan to find the first operators
	bool changed;

	if (limit==0) limit = UINT_MAX;

	optimizeProducts();
	DUMP(int lastreport = size() + 1000);
	do {
		changed = false;
		DUMP(cout << "\nExpr=" << *this << endl);
		for (int i=2; i<size() and i>0; i++) {
			if (expr[i]->isOperator()) {
				//assert(verify());
				int rule;
				int start;
				int oldSize = size();
				if ((rule=rpnrule(i, &start)) != 0) {
					DUMP(cout << ">> ExpR=" << *this << endl);
					changed = true;
#ifdef _DUMP
					cout << "i=" << i
						<< " rule=" << rule
						<< " old=" << oldSize
						<< " size=" << size();
#endif
					i += size()-oldSize;
					DUMP(cout << " newi=" << i << " start=" << start << endl);
					oldSize = size();
					optimizeProducts(start, i+1);
					DUMP(cout << " opt-size=" << size());
					i += size()-oldSize;
					DUMP(cout << " opti=" << i << endl);
					DUMP(cout << ">> ExpO=" << *this << endl << endl);
				}
#ifdef _DUMP
				if (size() > lastreport) {
					cout << i << " size = " << size() << endl;
					lastreport = size() + 1000;
				}
#endif
				//cout << "size=" << size() << " limit=" << limit << endl;
				if ((unsigned)size()>limit)
					return false;
			}
		}
		// Find possible products and optimize them
		if (changed)
			optimizeProducts();
	} while (changed);
	optimizeProducts();
	return true;
} // expand

/** recursive function to prepare the skip record
 * skip:  >0   jump to skip-location when current result is true
 *        =0   save complement of current result
 *        <0   jump to |skip|-location when current result is false
 *
 *  location:           m           n-1 n
 *       rpn:  [left-expr] [right-expr] +
 *      skip:          -n           -n  ? (coming from prev op)
 *
 *       rpn:  [left-expr] [right-expr] -
 *      skip:          -n            0  ? (coming from prev op)
 *
 *       rpn:  [left-expr] [right-expr] |
 *      skip:          +n           +n  ? (coming from prev op)
 *
 */
int GZone::_findSkip(int ptr, int when, int to)
{
	skip[ptr] = when*to;
	const GBody* body = expr[ptr];
	int cptr = ptr;
	ptr--;
	if (body == &GBody::tunion) {
		ptr = _findSkip(ptr,  1, cptr);		// when either branch is true
		ptr = _findSkip(ptr,  1, cptr);
	} else
	if (body == &GBody::tplus) {
		ptr = _findSkip(ptr, -1, cptr);		// when either branch is false
		ptr = _findSkip(ptr, -1, cptr);
	} else
	if (body == &GBody::tminus) {
		ptr = _findSkip(ptr,  0, cptr);		// when left is true
		ptr = _findSkip(ptr, -1, cptr);		// or right is false
	}
	return ptr;
} // _findSkip

/** prepare the skip record for RPN expressions */
void GZone::makeSkip()
{
	assert(rpn());
	//assert(skip.size()==0);

	// prepare the skip array
	skip.resize(size());
	std::fill(skip.begin(), skip.end(), 0);
	_findSkip(size()-1, -1, size());

	// merge chained skip jumps
	//   .   i   .   .   j   .   .   .  k   k+1
	//      -j  ------> -k  --------> -k-1 -->
	//     -k-1 ----------------------------->
	//                 -k-1 ----------------->
	//                                -k-1 -->
	for (int i=0; i<(int)skip.size(); i++) {
		int s = skip[i];
		int ss = Sign(s);
		while (abs(s)<size() and Sign(skip[abs(s)]) == ss)
			s = skip[abs(s)];
		skip[i] = s;
	}
} // makeSkip

/** delete a skipped item during compressing */
void GZone::_deleteSkip(int i)
{
	expr.erase(i);
	skip.erase(skip.begin()+i);
	for (int j=0; j<(int)skip.size(); j++)
		if (abs(skip[j]) > i) {
			if (skip[j]>0)
				skip[j]--;
			else
			if (skip[j]<0)
				skip[j]++;
		}
} // _deleteSkip

/* compress a skipped expression
 * WARNING: the process is not reversible
 */
void GZone::compress()
{
	assert(rpn());
	assert(skip.size()>0);
	int i = 0;
	int prev = 0;

	while (i<size()) {
		const GBody* body = expr[i];
		if (body->isOperator() and prev != 0 and prev == skip[i])
			_deleteSkip(i);
		//else
		//if (body == &GBody::tuniverse and skip[i] < 0) {
		//	// it doesn't work. One has to push 'true' in the result
		//	// for nonsense expressions like:
		//	//     | @ |   or @-@
		//	if (i+1<len(skip)) prev = skip[i+1];
		//	_deleteSkip(i);
		//} else
		//if (body == &GBody::tuniverse and skip[i] > 0) {
		//	// it should delete until skip e.g.   | @
		//}
		else {
			prev = skip[i];
			i++;
		}
	}

	_flags |= Flags::Compressed;
} // compress

/** --------------------------------------------------------------------- *
 * rpnrule
 *
 * Find a matching rule and apply it on the sub-expression starting from
 * the N position in the Reverse Polish Notation
 *
 * An expression is in normal form when all the parenthesis are expanded
 * and the expression is described as a sum (UNIONS) of products
 * (INTERSECTIONS and/or SUBTRACTIONS)
 *
 * An expression can be converted to normal form by repeatedly applying
 * the following set of production rules to the expression and then to its
 * sub-expressions:
 *
 *      Normal Form                        Reverse Polish Notation
 *   1. X-(Y|Z) -> (X-Y)-Z                X Y Z | -  ->  X Y - Z -
 *   2. X+(Y|Z) -> (X+Y)|(X+Z)            X Y Z | +  ->  X Y + X Z + |
 *   3. X-(Y+Z) -> (X-Y)|(X-Z)            X Y Z + -  ->  X Y - X Z - |
 *   4. X+(Y+Z) -> (X+Y)+Z                X Y Z + +  ->  X Y + Z +
 *   5. X-(Y-Z) -> (X-Y)|(X+Z)            X Y Z - -  ->  X Y - X Z + |
 *   6. X+(Y-Z) -> (X+Y)-Z                X Y Z - +  ->  X Y + Z -
 *   7. X|(Y|Z) -> (X|Y)|Z                X Y Z | |  ->  X Y | Z |
 *   8. (X-Y)+Z -> (X+Z)-Y                X Y - Z +  ->  X Z + Y -
 *   9. (X|Y)-Z -> (X-Z)|(Y-Z)            X Y | Z -  ->  X Z - Y Z - |
 *  10. (X|Y)+Z -> (X+Z)|(Y+Z)            X Y | Z +  ->  X Z + Y Z + |
 * X,Y, and Z here match both primitives or sub-expressions.
 *
 * Additionally check operations with special bodies
 *  * : null
 *  @ : universe
 *
 *      Normal Form              Reverse Polish Notation
 *  -1. @ + X -> X               @ X + -> X
 *  -2. @ - X -> @ - X           @ X - -> @ X -
 *  -3. @ | X -> @               @ X | -> @
 *
 *  -4. * + X -> *               * X + -> *
 *  -5. * - X -> *               * X - -> *
 *  -6. * | X -> X               * X | -> X
 *
 *  -7. X + @ -> X               X @ + -> X
 *  -8. X - @ -> *               X @ - -> *
 *  -9. X | @ -> @               X @ | -> @
 *
 * -10. X + * -> *               X * + -> *
 * -11. X - * -> X               X * - -> X
 * -12. X | * -> X               X * | -> X
 *
 *  @param n		index of the top operator (right-most)
 *  @param start [O]	return index of the left most term
 *  @param expand	perform expansion rules if applicable
 *  @param special	perform special body rules @,* if applicable
 *  @return		rule applied
 * ---------------------------------------------------------------------- */
int GZone::rpnrule(int n, int* start, bool expand, bool special)
{
	// Top-most operator
	GBody* op = expr[n];
	if (!op->isOperator()) return 0;

	// Find left and right sub-trees
	//    ll            lr            n
	//    <---left--->  <---right---> op
	int ll, lr;
	subTerms(n, &ll, &lr);
	*start = ll;
	if (special) {
		if (lr-ll == 1)	{	 // left = one body
			if (expr[ll] == &GBody::tuniverse) {
				// @ + X -> X          @ X + -> X
				if (op == &GBody::tplus) {
					expr.erase(n);		// operator
					expr.erase(ll);		// @
					return -1;
				} else
				// @ | X -> @          @ X | -> @
				if (op == &GBody::tunion) {
					//assert(false);	// should never arrive to this condition
					expr.erase(lr,n+1);	// right + operator
					return -3;
				}
				// @ - X -> @ - X       @ X - -> @ X -
				// do nothing
			} else
			if (expr[ll] == &GBody::tnil) {
				// * + X -> *           * X + -> *
				// * - X -> *           * X - -> *
				if (op == &GBody::tplus || op == &GBody::tminus) {
					expr.erase(lr, n+1);	// right + operator
					return -4;
				} else
				// * | X -> X           * X | -> X
				if (op == &GBody::tunion) {
					expr.erase(n);		// operator
					expr.erase(ll);		// left
					return -6;
				}
			}
		}
		if (n-lr == 1) {	// right = one body
			if (expr[lr] == &GBody::tuniverse) {
				// X + @ -> X           X @ + -> X
				if (op == &GBody::tplus) {
					expr.erase(lr,n+1);	// right + operator
					return -7;
				} else
				// X - @ -> *           X @ - -> *
				if (op == &GBody::tminus) {
					expr.erase(ll+1,n+1);	// left+1 + right + operator
					expr[ll] = &GBody::tnil;	// replace with null
					return -8;
				} else
				// X | @ -> @           X @ | -> @
				if (op == &GBody::tplus || op == &GBody::tunion) {
					//assert(false);	// should never arrive to this condition
					expr.erase(ll+1,n+1);	// left+1 + right + operator
					expr[ll] = &GBody::tuniverse;	// replace with universe
					return -9;
				}
			} else
			if (expr[lr] == &GBody::tnil) {
				// X + * -> *           X * + -> *
				if (op == &GBody::tplus) {
					expr.erase(ll+1,n+1);	// left+1 + right + operator
					expr[ll] = &GBody::tnil;	// replace with null
					return -10;
				} else
				// X - * -> X           X * - -> X
				if (op == &GBody::tminus) {
					expr.erase(lr,n+1);	// right + operator
					return -11;
				} else
				// X | * -> X           X * | -> X
				if (op == &GBody::tplus || op == &GBody::tunion) {
					expr.erase(lr,n+1);	// right + operator
					return -12;
				}
			}
		}
	}

	if (!expand) return 0;

	// Left operator
	GBody* lop = nullptr;
	if (lr>0) lop = expr[lr-1];

	int rule;
	// Right operator
	GBody* rop = expr[n-1];
	// Find Rule
	if (op==&GBody::tminus && rop==&GBody::tunion)
		rule = 1;
	else
	if (op==&GBody::tplus  && rop==&GBody::tunion)
		rule = 2;
	else
	if (op==&GBody::tminus && rop==&GBody::tplus)
		rule = 3;
	else
	if (op==&GBody::tplus  && rop==&GBody::tplus)
		rule = 4;
	else
	if (op==&GBody::tminus && rop==&GBody::tminus)
		rule = 5;
	else
	if (op==&GBody::tplus  && rop==&GBody::tminus)
		rule = 6;
	else
	if (op==&GBody::tunion && rop==&GBody::tunion)
		rule = 7;
	else
	if (op==&GBody::tplus  && lop==&GBody::tminus)
		rule = 8;
	else
	if (op==&GBody::tminus && lop==&GBody::tunion)
		rule = 9;
	else
	if (op==&GBody::tplus  && lop==&GBody::tunion)
		rule = 10;
	else
		return 0;

	// limits of sub-expressions X,Y,Z
	int Xu, Xl, /*Yu,*/ Yl, Zu, Zl;
	int L;		// aux variables

	// Find sub expressions X Y Z
	if (rule<=7) {    // X op (Y rop Z)
		Xu = lr-1;
		Xl = ll;
		subTerms(n-1, &ll, &lr);
		//Yu = lr-1;
		Yl = ll;
		Zu = n-2;
		Zl = lr;
	} else {	// (X lop Y) op Z
		Zu = n-1;
		Zl = lr;
		L  = lr-1;
		subTerms(L, &ll, &lr);
		Xu = lr-1;
		Xl = ll;
		//Yu = L-1;
		Yl = lr;
	}

	//printf("n=%d size=%d Rule: %d ",n,size(),rule);
	//printf(" X[%d:%d %d] ",Xl,Xu,Xu-Xl+1); //,"]=",expr[Xl:Xu+1]
	//printf(" Y[%d:%d %d] ",Yl,Yu,Yu-Yl+1); //,"]=",expr[Yl:Yu+1]
	//printf(" Z[%d:%d %d]\n",Zl,Zu,Zu-Zl+1); //,"]=",expr[Zl:Zu+1]

	// Expand the rule
	switch (rule) {
		//# 1. X-(Y|Z) -> (X-Y)-Z        X Y Z | -  ->  X Y - Z -
		case 1:
			// Leave X Y
			// Insert a - operator after Y
			expr.insert(Zl, &GBody::tminus);
			// Chop length by 1
			expr.erase(Zu+2);
			// Change the last operator to -
			expr[Zu+2] = &GBody::tminus;
			break;

		// 2. X+(Y|Z) -> (X+Y)|(X+Z)     X Y Z + |  ->  X Y + X Z + |
		case 2:
			// Leave X Y
			// Insert a + operator after Y
			expr.insert(Zl, &GBody::tplus);
			// Copy X after the + operator
			copy(Zl+1, Xl, Xu-Xl+1);
			Zu += Xu-Xl+2;
			// Change last 2 operators to + |
			expr[Zu+1] = &GBody::tplus;
			expr[Zu+2] = &GBody::tunion;
			break;

		// 3. X-(Y+Z) -> (X-Y)|(X-Z)     X Y Z + -  ->  X Y - X Z - |
		case 3:
			// Leave X Y
			// Insert a - operator after Y
			expr.insert(Zl, &GBody::tminus);
			// Copy X after the - operator
			copy(Zl+1, Xl, Xu-Xl+1);
			Zu += Xu-Xl+2;
			// Change last 2 operators to - |
			expr[Zu+1] = &GBody::tminus;
			expr[Zu+2] = &GBody::tunion;
			break;

		// 4. X+(Y+Z) -> (X+Y)+          X Y Z + +  ->  X Y + Z +
		case 4:
			// Leave X Y
			// Insert a + operator after Y
			expr.insert(Zl, &GBody::tplus);
			// Chop length by 1
			expr.erase(Zu+2);
			// Change the last operator to +
			expr[Zu+2] = &GBody::tplus;
			break;

		// 5. X-(Y-Z) -> (X-Y)|(X+Z)     X Y Z - -  ->  X Y - X Z + |
		case 5:
			// Leave X Y
			// Insert a - operator after Y
			expr.insert(Zl, &GBody::tminus);
			// Copy X after the - operator
			copy(Zl+1, Xl, Xu-Xl+1);
			Zu += Xu-Xl+2;
			// Change last 2 operators to + |
			expr[Zu+1] = &GBody::tplus;
			expr[Zu+2] = &GBody::tunion;
			break;

		// 6. X+(Y-Z) -> (X+Y)-Z	 X Y Z - +  ->  X Y + Z -
		case 6:
			// Leave X Y
			// Insert a + operator after Y
			expr.insert(Zl, &GBody::tplus);
			// Chop length by 1
			expr.erase(Zu+2);
			// Change the last operator to -
			expr[Zu+2] = &GBody::tminus;
			break;

		// 7. X|(Y|Z) -> (X|Y)|Z	 X Y Z | |  ->  X Y | Z |
		case 7:
			// Leave X Y
			// Insert a | operator after Y
			expr.insert(Zl, &GBody::tunion);
			// Chop length by 1
			expr.erase(Zu+2);
			// Change the last operator to |
			expr[Zu+2] = &GBody::tunion;
			break;

		// 8. (X-Y)+Z -> (X+Z)-Y	 X Y - Z +  ->  X Z + Y -
		case 8:
			// Leave X
			// Copy "Z +" after X
			L = Zu-Zl+2;
			copy(Yl, Zl, L);
			// Delete old "Z +"
			expr.erase(Zl+L,Zl+L+L);	// XXX to be verified!
			break;

		// 9. (X|Y)-Z -> (X-Z)|(Y-Z)     X Y | Z -  ->  X Z - Y Z - |
		//10. (X|Y)+Z -> (X+Z)|(Y+Z)     X Y | Z +  ->  X Z + Y Z + |
		case 9:
		case 10:
			// Leave X
			// Copy "Z -" or "Z +" after X
			L = Zu-Zl+2;
			// Correct Z position
			copy(Yl, Zl, L);
			Zl += L;
			Zu += L;
			// Delete the | in front of Z
			expr.erase(Zl-1);
			// Add | at the end
			expr.insert(Zu+1, &GBody::tunion);
			break;
	}

	return rule;
} // rpnrule

/** --------------------------------------------------------------------- *
 * subTerms
 *
 * This routine returns the pointers in the RPN terms array of the
 * starting point of the left sub-expression LOWLEFT and right
 * sub-expression LOWRIGHT given the pointer of the expr-operator NTX.
 * The searching is performed by scanning from right to left the number
 * of operators and objects pushed into the stack
 *
 * EXP          (expr-left) op (expr-right)
 *
 * RPN ...... | expr-left | expr-right | op | .......
 *           Lowleft      LowRight  op-1 ntx
 * ---------------------------------------------------------------------- */
void GZone::subTerms(int n, int* lowLeft, int* lowRight)
{
	int nop = 0;
	*lowRight = 0;
	*lowLeft  = 0;

	while (n>=0) {
		if (expr[n]->isOperator())
			nop++;
		else
			nop--;

		n--;
		if (nop==0) {
			if (*lowRight==0) {
				*lowRight = n+1;
				nop++;
				continue;
			} else {
				*lowLeft = n+1;
				return;
			}
		}
	}
} // subTerms

/** copy a sub-expression
 *  @param dst		insert position
 *  @param src		starting position
 *  @param length	length of sub-expression
 */
void GZone::copy(int dst, int src, int length)
{
	// Create space by shifting everything by length
	expr.shiftby(dst, length);

	// Correct starting position if needed
	int nf;
	if (src > dst)
		nf = src + length;
	else
		nf = src;

	// copy terms
	memcpy(&expr[dst], &expr[nf], sizeof(GBody*)*length);
} // copy

/** inside_threshold
 * @param p	location to search
 * @param ignore_a, ignore_b, ignore_c quads to ignore in the computation
 * @return:
 *	true if the point is inside
 *	false if outside
 */
bool GZone::_insideThreshold(const Vector& p, const Quad* ignore_a,
			const Quad* ignore_b, const Quad* ignore_c) const
{
	assert(std());

	// Positive terms
	int i = 0;
	for (; i<size(); i++) {
		GBody const* body = expr[i];
		if (body==&GBody::tminus) break;
		if (!body->inside(p.x, p.y, p.z, ignore_a, ignore_b, ignore_c))
			return false;
	}

	i++; // Skip the minus term

	// Negative terms
	for (; i<size(); i++) {
		GBody const* body = expr[i];
		if (body==&GBody::tminus) break;
		if (!body->outside(p.x, p.y, p.z, ignore_a, ignore_b, ignore_c))
			return false;
	}

	return true;
} // _insideThreshold

/** Create a bounding box from a body that is only described by planes
 * (including BOX,WED,RHP,ARB only with a plus sign
 * @param bb	limiting bounding box
 */
void GZone::_bboxFromPlanes(const BBox& bb)
{
	// Temporary RPP body out of the limiting bounding box
	GBOXBody rpp("<bb>", GBody::Type::RPP);
	rpp.set(bb.low().x, bb.high().x,
		bb.low().y, bb.high().y,
		bb.low().z, bb.high().z);
	rpp.create();

	_bbox.reset();
	// Produce every possible combination of 3 planes
	// assume -1 body is the new bounding box rpp
	for (int i=-1; i<size(); i++) {
		GBody const* body = (i>=0)? expr[i] : &rpp;
		if (body==&GBody::tminus) continue;
		for (int ii=0; ii<body->nQ(); ii++) {
			const Quad& pa = body->Q(ii);
			int j, jj;
			if (ii<body->nQ()-1) {
				j  = i;		// quads of the same body
				jj = ii+1;
			} else {
				j  = i+1;	// quads of the next body
				jj = 0;
			}

			for (; j<size(); j++, jj=0) {
				GBody const* bbody = (j>=0)? expr[j] : &rpp;
				if (bbody==&GBody::tminus) continue;
				for (; jj<bbody->nQ(); jj++) {
					const Quad& pb = bbody->Q(jj);
					int k,kk;
					if (jj<bbody->nQ()-1) {
						k  = j;		// quads of the same body
						kk = jj+1;
					} else {
						k  = j+1;	// quads of the next body
						kk = 0;
					}

					for (; k<size(); k++, kk=0) {
						GBody const* cbody = (k>=0)? expr[k] : &rpp;
						if (cbody==&GBody::tminus) continue;
						for (; kk<cbody->nQ(); kk++) {
							const Quad& pc = cbody->Q(kk);
							// Solve system and find vertices,
							// add them to the bbox
							// Cax*x + Cay*y + Caz*z + Ca = 0
							// Cbx*x + Cby*y + Cbz*z + Cb = 0
							// Ccx*x + Ccy*y + Ccz*z + Cc = 0
							// A x X = B
							Matrix3 M;
							M(0,0) = pa.Cx; M(0,1) = pa.Cy; M(0,2) = pa.Cz;
							M(1,0) = pb.Cx; M(1,1) = pb.Cy; M(1,2) = pb.Cz;
							M(2,0) = pc.Cx; M(2,1) = pc.Cy; M(2,2) = pc.Cz;
							if (not M.inverse()) continue;
							Vector v = M.mult(-pa.C, -pb.C, -pc.C);
							if (_insideThreshold(v, &pa, &pb, &pc))
								_bbox.add(v);
						}
					}
				}
			}
		}
	}
} // _bboxFromPlanes

/** guess of 3D bounding box
 * @return bounding box of zone
 */
void GZone::_updateBBox(const BBox* bblimit)
{
	assert(bblimit==nullptr or bblimit->isValid());
	if (std()) {
		// Check if expression contains only planes!
		bool plus = true;
		bool onlyPlanes = true;
		for (int i=0; i<size(); i++) {
			GBody const* body = expr[i];
			if (body==&GBody::tminus) {
				plus = false;
				continue;
			}
			if (body->isPlane()) continue;
			if (plus and (body->isBox() or	// only on plus side RPP, BOX, WED, ARB, RHP
				body->type()==GBody::Type::WED or
				body->type()==GBody::Type::ARB or
				body->type()==GBody::Type::RHP)) continue;

			onlyPlanes = false;
			break;
		}

		// Start with an infinite bounding box or bblimit
		if (bblimit==nullptr or !bblimit->isValid())
			_bbox.infinite();
		else
			_bbox = *bblimit;

		if (onlyPlanes) {
			_bboxFromPlanes(_bbox);
			return;
		}

		// FIRST PASS
		// Compute a rough bounding box using the bounding boxes of the bodies
		bool negative = false;
		for (int i=0; i<size(); i++) {
			GBody* body = expr[i];

			// Detect negative term
			if (body==&GBody::tminus) {
				negative = true;
				continue;
			}
			body->updateBBox();

			if (!negative)
				_bbox += body->bbox();
			else {
				// Only subtract bodies for which
				// inside bbox = outside bbox
				// FIXME Difference should ONLY work if there is no
				//       arbitrary rotation (different than 0,90,180,270deg)
				if (!body->hasMatrix() and
				    (body->type() <= GBody::Type::XYP or
				     InRange(GBody::Type::PX, body->type(), GBody::Type::PZ) or
				     body->type() == GBody::Type::RPP))
					_bbox.differenceEqual(body->bbox());
			}
			if (!_bbox.isValid()) return;
		}

		// SECOND PASS
		// Find all planes, intersect them with the bounding box and
		// create a new bounding box containing these vertices +
		// bounding box vertices if they are inside the zone
		bool changed=true;
		for (int ii=100; ii-- && changed && _bbox.isValid();) {
			// Find all planes in bodies
			negative = false;
			changed  = false;
			double vol = _bbox.volume();
			for (int i=0; i<size(); i++) {
				GBody * body = expr[i];
				// Detect negative term
				if (body==&GBody::tminus) {
					negative = true;
					continue;
				}
				if (body->isPlane() and
				   _bbox.intersectPlane(body->Q(0), negative)) {
					if (not _bbox.isValid()) break;
					if (!Eq(vol, _bbox.volume(), vol*SMALL3D))
						changed = true;
					if (Eq0(_bbox.volume(),SMALL))
						break;
				}
			}
		}
		//cout << name() << "." << id() << ": Iterations:" <<  ii
		//	<< "  vol=" << _bbox.volume() << " bb=" << _bbox << endl;
	} else
	if (rpn()) {
		// FIXME Stack<pair<BBox,GBody*> >
		// special care for PLANES and
		// for difference with RPP without transformation
		Stack<std::pair<BBox,const GBody*>, ZONE_STACK_SIZE> stack;
		if (bblimit==nullptr or !bblimit->isValid())
			_bbox.reset();
		else
			_bbox = *bblimit;
		for (int i=0; i<size(); i++) {
			GBody* body = expr[i];
			if (body == &GBody::tplus) {
				BBox b = stack.pop().first;
				stack().first  += b;
				stack().second = nullptr;
			} else
			if (body == &GBody::tminus) {
				// FIXME planes could be used to limit bounding box!!!
				auto B = stack.pop();
				if (B.second and B.second->isPlane())
					stack().first.intersectPlane(B.second->Q(0), true);
				stack().second = nullptr;
			} else
			if (body == &GBody::tunion) {
				BBox b = stack.pop().first;
				stack().first |= b;
				stack().second = nullptr;
			} else
			//if (body == &GBody::tnot) {
			//	// FIXME should be treated as tminus for planes
			//	stack.pop();
			//} else
			if (body == &GBody::tuniverse)
				stack.push(std::make_pair(BBox(true),nullptr));
			else {
				body->updateBBox();
				stack.push(std::make_pair(body->bbox(),body));
			}
		}
		assert(stack.one());

		_bbox = stack().first;
	}
} // _updateBBox

#ifdef USE_OBBOX
/** update the oriented bounding box */
void GZone::_updateOBBox()
{
	_obbox.reset();
	if (std()) {
		// Iterate body number times so each body has a change to
		// improve the BB
		for (int pass=0; pass<size()-1; pass++) {
			bool negative = false;
			//double vol = _obbox.volume();
			for (int i=0; i<size(); i++) {
				GBody const* body = expr[i];

				// Detect negative term
				if (body==&GBody::tminus) {
					if (!_obbox.isValid()) return;
					negative = true;
					continue;
				}

				if (!negative) {
					if (_obbox.isValid())
						_obbox.Intersect(body->obbox(false));
					else
						_obbox = body->obbox(false);
				} else {
					// Subtract the inner bounding box
					_obbox.Difference(body->obbox(true));
				}
				if (!_obbox.isValid()) return;
			}
			// If the volume didn't change from last time, assume it
			// is stable
			//if (Eq0(_obbox.volume() - vol, SMALL2))
			//break;
		}
		// All bodies had a chance to improve the BB
		// if (pass == size()-1) cout << "Exhausted bb stabilization passes, must be optimal" << endl;
	} else
	if (rpn()) { // RPN
#if 00
		// FIXME: review!
		OBBox aux;
		const int stack_size = 100;
		OBBox* stack[stack_size], *b;

		int  sp = -1;   /* stack pointer */

		for (int i=0; i<size(); i++) {
			const GBody* body = expr[i];
			if (body == GZone::tplus) {
				if (sp<1) break;
				b = stack[sp];
				sp--;
				aux = OBBox(*stack[sp]);
				stack[sp] = &aux;
				stack[sp]->Intersect(*b);
			} else
			if (body == GZone::tminus) {
				if (sp<1) break;
				b = stack[sp];
				sp--;
				aux = OBBox(*stack[sp]);
				stack[sp] = &aux;
				stack[sp]->Difference(*b);
			} else
			if (body == &GBody::tnot) {
				assert(false);
			} else
			if (body == GZone::tunion) {
				if (sp<1) break;
				b = stack[sp];
				sp--;
				aux = OBBox(*stack[sp]);
				stack[sp] = &aux;
				stack[sp]->Union(*b);
			} else
			if (body == GZone::tuniverse) {
				++sp;
				aux.infinite();
				stack[sp] = &aux;
			} else {
				++sp;
				stack[sp] = body->obbox(false);
				if (sp>=100) break;
				assert(sp<100);
			}
		}
		if (sp==0)
			*_obbox = *stack[0];
#endif
	}
} // _updateOBBox
#endif

/** memory */
size_t GZone::memory() const
{
	return sizeof(GZone); // + _maxSize * sizeof(GBody*);
} // memory

/** operator << */
//#define SHOW_INDEX
std::ostream& operator << (std::ostream& s, const GZone& zone)
{
	if (zone.region!=nullptr)
		s << zone.name() << "-" << zone.id() << ": ";

	switch (zone.format()) {
		case GZone::Format::STD:
			s << "[STD]";
			break;
		case GZone::Format::RPN:
			s << "[RPN]";
			break;
		case GZone::Format::NORMAL:
			s << "[NORMAL]";
			break;
	}
	s << ' ';

	bool space = false;
	if (zone.std()) {
		bool plus = true;
		for (int i=0; i<zone.size(); i++) {
			GBody const* body = zone[i];
			if (body==&GBody::tminus)
				plus = false;
			else {
				if (space) s << ' ';
				s << (plus?'+':'-');
				s << body->name();
				space = true;
			}
		}
	} else
	for (int i=0; i<zone.size(); i++) {
		GBody const* body = zone[i];
		if (zone.rpn()) {
			if (space) s << ' ';
			space = true;
		} else
		if (body->isOperator()) {
			if (space) s << ' ';
			space = (body != &GBody::tleft);
		}
		s << body->name();
#ifdef SHOW_INDEX
		s << "/" << i;
#endif
	}
	return s;
} /* operator << */
