/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2011
 */

#ifndef __GEOMETRY_TETRA_H
#define __GEOMETRY_TETRA_H

#include <assert.h>
#include <vector>

#include "vertex.h"

#include "math/plane.h"

// =================================== Tetra ===================================
/**
 * Class implementing the basic tetrahedra
 *
 *                      v
 *                    .
 *                  ,/
 *                 /
 *              2
 *            ,/|`\
 *          ,/  |  `\
 *        ,/    '.   `\
 *      ,/       |     `\
 *    ,/         |       `\
 *   0-----------'.--------1 --> u
 *    `\.         |      ,/
 *       `\.      |    ,/
 *          `\.   '. ,/
 *             `\. |/
 *                `3
 *                   `\.
 *                      ` w
 */
class Tetra {
protected:
	// FIXME what do we really need?
	Vertex*	_vertex[4];		/** pointers to vertices		*/
	Plane*	_plane[4];		/** pointers to planes			*/
	bool	_side[4];		/** positive/negative side of plane	*/
	Tetra*	_neighbor[4];		/** neighbor tetrahedra			*/
	int	_processed;		/** processed flag			*/

private:
const	static int faceOrder[4];
const	static int faceVertex[4][3];

public:
	Tetra()		{ clear(); }
	Tetra(Vertex* a, Vertex* b, Vertex* c, Vertex* d);
//	Tetra(const Tetra&);
	~Tetra() {};

	void	set(Plane* ABC, Plane* ABD, Plane* ACD, Plane* BCD);
	Point	barycenter()	const { return 0.25*(*_vertex[0] + *_vertex[1] + *_vertex[2] + *_vertex[3]); }

const	Vertex&	A()		const	{ return *_vertex[0]; }
const	Vertex&	B()		const	{ return *_vertex[1]; }
const	Vertex&	C()		const	{ return *_vertex[2]; }
const	Vertex&	D()		const	{ return *_vertex[3]; }

	bool	checkNeighbor(Tetra* t);

	/** @return surface only of borders */
	double	surface(bool border)	const;
	double	volume() const;

#if 0
	/** @return normal */
//const	Vector&	normal()		const	{ return _normal; }

	/** @return point face-plane distance */
	// the plane equation is   n*(r-A) = n*r - n*A = 0
	// distance is given as n*(p-A)/|n| = (n*p-n*A)/|n|
	// normal n is already normalized |n| = 1
	double	distance(const Point& p) const { return normal()*(p - A()); }

//	bool	show(const int i)	const	{ return _show[i]; }
	bool	interectRay(const Vertex& O, const Vector& D, double* out, double eps=1e-7);
#endif

	int	nneighbors()		const {
			int n=0;
			for (int i=0; i<4; i++)
				if (_neighbor[i]) n++;
			return n;
		}

	/** set/get processed flag */
	int	processed()		 const	{ return _processed; }
	void	processed(int p)		{ _processed = p; }

protected:
	void	clear();
	int	has(const Vertex* v) {
			for (int i=0; i<4; i++)
				if (v==_vertex[i]) return i;
			return -1;
		}
	int	faceNumber(int a, int b, int c);

	/** Print face. */
friend	std::ostream& operator << (std::ostream&, const Tetra& tetra);
//friend class TetraMesh;
}; // Tetra

typedef std::vector<Plane*>	PlaneArray;
typedef std::vector<Tetra*>	TetraArray;
//typedef int	(*TetraFunc) (Tetra*, void*);

#endif
