/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	14-Jan-2014
 */

#ifndef __GRAPH_COLOR_H
#define __GRAPH_COLOR_H

#include <ostream>
#include <cassert>

#include "os.h"

#define COLOR_TRANSPARENT	0xFFFFFFFF
#define COLOR_ALPHA		0xFF000000

// ================================== Color32 ==================================
/* int breakup into color bytes for little endian*/
union Color32 {
	struct {
//#if CPU_ENDIAN == LITTLE_ENDIAN
		byte b;
		byte g;
		byte r;
		byte a;
//#else
//		byte alpha;
//		byte red;
//		byte green;
//		byte blue;
//#endif
	} rgb;
	dword value;

public:
	Color32() {}
	Color32(dword v) : value(v) {}
//#if CPU_ENDIAN == LITTLE_ENDIAN
	Color32(byte ar, byte ag, byte ab, byte aa=0) : rgb{ab,ag,ar,aa} {}
//#else
//	Color32(byte ar, byte ag, byte ab, byte aa=0) : rgb.a(aa), rgb.r(ar), rgb.g(ag), rgb.b(ab) {}
//#endif

	Color32& operator = (dword v)	{ value=v; return *this; }
	void	operator() (dword v)	{ value=v; }
	void	operator() (byte ar, byte ag, byte ab, byte aa=0) {
				rgb.r = ar;
				rgb.g = ag;
				rgb.b = ab;
				rgb.a = aa;
			}
	operator dword () const	{ return value;	}
	operator dword& ()	{ return value;	}
	byte&	red()		{ return rgb.r;	}
	byte&	green()		{ return rgb.g;	}
	byte&	blue()		{ return rgb.b;	}
	byte&	alpha()		{ return rgb.a;	}
	byte	red()	const	{ return rgb.r;	}
	byte	green()	const	{ return rgb.g;	}
	byte	blue()	const	{ return rgb.b;	}
	byte	alpha()	const	{ return rgb.a;	}
	void	red(byte x)	{ rgb.r = x;	}
	void	green(byte x)	{ rgb.g = x;	}
	void	blue(byte x)	{ rgb.b = x;	}
	void	alpha(byte x)	{ rgb.a = x;	}
	byte	R()	const	{ return rgb.r;	}
	byte	G()	const	{ return rgb.g;	}
	byte	B()	const	{ return rgb.b;	}
	byte	A()	const	{ return rgb.a;	}
	int	iR()	const	{ return (int)rgb.r;	}
	int	iG()	const	{ return (int)rgb.g;	}
	int	iB()	const	{ return (int)rgb.b;	}
	int	iA()	const	{ return (int)rgb.a;	}

	void	R(byte x)	{ rgb.r = x;	}
	void	G(byte x)	{ rgb.g = x;	}
	void	B(byte x)	{ rgb.b = x;	}
	void	A(byte x)	{ rgb.a = x;	}

	// Get components by index
	byte get(int i) const {
			switch (i) {
				case 0: return rgb.r;
				case 1: return rgb.g;
				case 2: return rgb.b;
				case 3: return rgb.a;
				default:
					assert(false);
					return 0;   // satisfy compiler, but we should never get here
			}
		}
	byte operator[](int i) const { return get(i); }

	/** make RGB color */
static	dword	RGB(unsigned red, unsigned green, unsigned blue)
			{ return red<<16 | green<<8 | blue; }

	/** darken by color in the range 0..level */
static	Color32	darker(const dword col, const int level) {
			Color32 color(col);
			color.R((byte)((color.iR() * level)>>8));
			color.G((byte)((color.iG() * level)>>8));
			color.B((byte)((color.iB() * level)>>8));
			return color;
		}

	/** darken by color in the range 0..level */
	Color32	darker(const int level) const { return Color32::darker(value, level); }

	/** lighten by color in the range level..255*/
static	Color32	lighter(const dword col, const int level) {
			Color32 color(col);
			color.R((byte)(((color.iR() * (256-level)) >> 8) + level));
			color.G((byte)(((color.iG() * (256-level)) >> 8) + level));
			color.B((byte)(((color.iB() * (256-level)) >> 8) + level));
			return color;
		}

	/** lighten by color in the range level..255*/
	Color32	lighter(const int level) const	{ return Color32::lighter(value, level); }

	/** intensity */
	int	intensity()	const { return ((77*iR() + 151*iG() + 28*iB())>>8); }

static	int	intensity(const dword col) { return Color32(col).intensity(); }

	float	gray()	const	{ return 0.299f*(float)R() +
					 0.587f*(float)G() +
					 0.114f*(float)B(); }

	/** adjust color between lw .. hg */
	void	remap(const Color32& lw, const Color32& hg);

	/** blend - transparent addition of colors */
	void	blend(const Color32& color);

	// blend this with color and alpha
	void	blend(const Color32& color, int a) {
			// 255-a = a^0xff
			red()   = (byte)((iR()*a + color.iR()*(a^0xff)) >> 8);
			green() = (byte)((iG()*a + color.iG()*(a^0xff)) >> 8);
			blue()  = (byte)((iB()*a + color.iB()*(a^0xff)) >> 8);
		}

	/** convert to HSV */
	void	getHSV(double* h, double* s, double* v) const;

	/** set from HSV */
	void	setHSV(const double h, const double s, const double v);

	/** incremental average by adding one more color */
	void	average(const Color32& color, int n) {
			int delta_red   = (color.iR() - iR()) / n;
			int delta_green = (color.iG() - iG()) / n;
			int delta_blue  = (color.iB() - iB()) / n;

			alpha()  = color.alpha();
			red()   += delta_red;
			green() += delta_green;
			blue()  += delta_blue;
		}

	/* stream inserter*/
friend	std::ostream& operator<<(std::ostream& os, const Color32& rhs) {
			return os << "(" << rhs.iR()
				  << ", " << rhs.iG()
				  << ", " << rhs.iB()
				  << ")";
		}
}; // union Color32

// =================================== Color ===================================
// Colors are using SINGLE precision numbers "float"
// No need for the extra digits
// WARNING: Colors can have values from 0.0 to +inf (more than 1.0 when saturated)
#define _CEPSILON 1e-7
class Color {
private:
	// The components.
	float	_red;		/** visible red	ranging from 0..1	*/
	float	_green;		/** green				*/
	float	_blue;		/** blue				*/

public:	// Some basic colors
static const Color	Black;
static const Color	Red;
static const Color	Green;
static const Color	Blue;
static const Color	Magenta;
static const Color	Turquoise;
static const Color	Yellow;
static const Color	White;

public:
	Color(const float R=0.0, const float G=0.0, const float B=0.0) { set(R,G,B); }
	Color(const double R, const double G, const double B) { set(R,G,B); }
	Color(const int R, const int G, const int B) { set(R,G,B); }
	Color(const dword color) { set(color); }
	Color(const Color32& color) { set(color); }

	Color(const Color&) = default;
	Color&	operator = (const Color&) = default;

	// The components in Cartesian coordinate system.
	float red()	const { return _red; }
	float green()	const { return _green; }
	float blue()	const { return _blue; }

	void red(const float R)		{ _red   = Max(0.0f, R); }
	void green(const float G)	{ _green = Max(0.0f, G); }
	void blue(const float B)	{ _blue  = Max(0.0f, B); }
	void red(const int R)		{ _red   = Max(0.0f, (float)R/255.0f); }
	void green(const int G)		{ _green = Max(0.0f, (float)G/255.0f); }
	void blue(const int B)		{ _blue  = Max(0.0f, (float)B/255.0f); }

	// Get components by index
	float operator () (int i) const {
			if (i == 0) return _red;
			else
			if (i == 1) return _green;
			return _blue;
		}
	float& operator () (int i) {
			if (i == 0) return _red;
			else
			if (i == 1) return _green;
			return _blue;
		}

	// return color as dword
	dword	color32() const {
			Color32 color;
			color.red()   = (byte)Min(255, (int)(255.0*red()));
			color.green() = (byte)Min(255, (int)(255.0*green()));
			color.blue()  = (byte)Min(255, (int)(255.0*blue()));
			return color;
		}

	// Set the components in Cartesian coordinate system.
	void set(const Color& color) {
			_red   = color._red;
			_green = color._green;
			_blue  = color._blue;
		}

	void set(const float R, const float G, const float B) {
			_red   = Max(0.0f, R);
			_green = Max(0.0f, G);
			_blue  = Max(0.0f, B);
		}

	void set(const double R, const double G, const double B) {
			_red   = Max(0.0f, (float)R);
			_green = Max(0.0f, (float)G);
			_blue  = Max(0.0f, (float)B);
		}

	void set(const int R, const int G, const int B) {
			_red   = Max(0.0f, (float)R/255.0f);
			_green = Max(0.0f, (float)G/255.0f);
			_blue  = Max(0.0f, (float)B/255.0f);
		}

	void set(const Color32& color) {
			_red   = (float)color.R()/255.0f;
			_green = (float)color.G()/255.0f;
			_blue  = (float)color.B()/255.0f;
		}

	void set(const dword color) { set(Color32(color)); }

	// Comparisons
	bool operator == (const Color& c) const {
			return (Eq(_red,   c._red,   _CEPSILON) &&
				Eq(_green, c._green, _CEPSILON) &&
				Eq(_blue,  c._blue,  _CEPSILON)) ? true : false;
		}

	bool operator != (const Color& c) const {
			return (Eq(_red,   c._red,   _CEPSILON) &&
				Eq(_green, c._green, _CEPSILON) &&
				Eq(_blue,  c._blue,  _CEPSILON)) ? false : true;
		}

	bool operator < (const Color& c) const {
			return  _red   < c._red   &&
				_green < c._green &&
				_blue  < c._blue;
		}

	bool operator <= (const Color& c) const {
			return  _red   <= c._red   &&
				_green <= c._green &&
				_blue  <= c._blue;
		}

	bool operator > (const Color& c) const	{ return !(*this <= c); }
	bool operator >= (const Color& c) const	{ return !(*this < c); }

	// Addition.
	// No range limitation since colors can have any positive value!
	Color& operator += (const Color& p) {
			_red   += p._red;
			_green += p._green;
			_blue  += p._blue;
			return *this;
		}

	// Subtraction.
	Color& operator -= (const Color& p) {
			_red   = Max(0.0f, _red   - p._red);
			_green = Max(0.0f, _green - p._green);
			_blue  = Max(0.0f, _blue  - p._blue);
			return *this;
		}

	// Scaling with real numbers.
	Color& operator *= (float a) {
			_red   *= a;
			_green *= a;
			_blue  *= a;
			return *this;
		}

	// Scaling with real numbers.
	Color& operator *= (double a) {
			_red   *= (float)a;
			_green *= (float)a;
			_blue  *= (float)a;
			return *this;
		}

	// Scaling with real numbers.
	Color& operator /= (float a) {
			float inva = 1.0f/a;
			_red   *= inva;
			_green *= inva;
			_blue  *= inva;
			return *this;
		}

	// Scaling with another color.
	Color& operator *= (const Color& color) {
			_red   *= color._red;
			_green *= color._green;
			_blue  *= color._blue;
			return *this;
		}

	// blend this with color
	void	blend(Color& color, float alpha) {
			_red   = red()  *alpha + color.red()  *(1.0f-alpha);
			_green = green()*alpha + color.green()*(1.0f-alpha);
			_blue  = blue() *alpha + color.blue() *(1.0f-alpha);
		}

	// blend this with color, using an alpha RGB triplet
	void	blend(Color& color, const Color& alpha) {
			_red   = red()   * alpha.red()   + color.red()   * (1.0f - alpha.red());
			_green = green() * alpha.green() + color.green() * (1.0f - alpha.green());
			_blue  = blue()  * alpha.blue()  + color.blue()  * (1.0f - alpha.blue());
		}

	// Color transformations
	void	setHSV(float, float, float);
	void	getHSV(float&, float&, float&);
	float	gray()	const { return	0.299f*(float)_red   +
					0.587f*(float)_green +
					0.114f*(float)_blue; };

	// return luminance
	double luminance() const {
			const double YWeight[3] = { 0.212671, 0.715160, 0.072169 };
			double v = 0.;
			for (int i = 0; i < 3; ++i)
				v += YWeight[i] * operator()(i);
			return v;
		}

	// convert from/to XYZ color space
	Color XYZ() const {
			const float XWeight[3] = { 0.412453f, 0.357580f, 0.180423f };
			const float YWeight[3] = { 0.212671f, 0.715160f, 0.072169f };
			const float ZWeight[3] = { 0.019334f, 0.119193f, 0.950227f };

			Color retCol;
			for (int i = 0; i < 3; ++i) {
				retCol._red   += XWeight[i] * operator()(i);
				retCol._green += YWeight[i] * operator()(i);
				retCol._blue  += ZWeight[i] * operator()(i);
			}
			return retCol;
		}

	Color XYZ(double x, double y, double z) {
			double c[3];
			c[0] =  3.240479 * x -  1.537150 * y -  0.498535 * z;
			c[1] = -0.969256 * x +  1.875991 * y +  0.041556 * z;
			c[2] =  0.055648 * x -  0.204043 * y +  1.057311 * z;
			return Color(c[0], c[1], c[2]);
		}

	// convert spectral to RGB
	Color spectralToRGB() const {
			Color XYZCol, retCol;
			const float redWeight[3]   = {  3.240479f, -1.537150f, -0.498535f };
			const float greenWeight[3] = { -0.969256f,  1.875991f,  0.041556f };
			const float blueWeight[3]  = {  0.055648f, -0.204043f,  1.057311f };

			XYZCol = XYZ();

			retCol._red   = redWeight[0]   * XYZCol(0) +
			                redWeight[1]   * XYZCol(1) +
			                redWeight[2]   * XYZCol(2);
			retCol._green = greenWeight[0] * XYZCol(0) +
			                greenWeight[1] * XYZCol(1) +
			                greenWeight[2] * XYZCol(2);
			retCol._blue  = blueWeight[0]  * XYZCol(0) +
			                blueWeight[1]  * XYZCol(1) +
			                blueWeight[2]  * XYZCol(2);
			return retCol;
		}

	// clamp range
	void clamp() {
			_red   = Range(0.f, _red,   1.f);
			_green = Range(0.f, _green, 1.f);
			_blue  = Range(0.f, _blue,  1.f);
		}

	// calculate approximate (!) squared distance in RGB color space [0,~1]
	// https://en.wikipedia.org/wiki/Color_difference
	// this allows for comparing two colors based on a metric, 0 = equal, 1 = very different.
	// To be perceptually correct one would need to go to HSL color space but that's costly.
	double calcDiff2( const Color& rhs ) {
			double rmean = (_red - rhs._red) * 0.5;
			return ((2 + rmean / 256.) * (_red - rhs._red) * (_red - rhs._red) +
				4 * (_green - rhs._green) * (_green - rhs._green) +
				(2 + (255 - rmean) / 256) * (_blue - rhs._blue) * (_blue - rhs._blue)) / 1e6;
		}
}; // Color

// Output to a stream.
inline std::ostream& operator << (std::ostream& os, const Color& col) {
		os << "[" << col.red() << ", " << col.green() << ", " << col.blue() << "]";
		return os;
	}

// Addition of 3-colors
inline Color operator + (const Color& a, const Color& b) {
		return Color(a.red()+b.red(), a.green()+b.green(), a.blue()+b.blue());
	}

// Subtraction of 3-colors.
inline Color operator - (const Color& a, const Color& b) {
		return Color(a.red()-b.red(), a.green()-b.green(), a.blue()-b.blue());
	}

// Scaling of a color with a real number
inline Color operator * (const Color& p, float a) {
		return Color(a*p.red(), a*p.green(), a*p.blue());
	}

inline Color operator * (float a, const Color& p) {
		return Color(a*p.red(), a*p.green(), a*p.blue());
	}

inline Color operator * (const Color& p, double a) {
		return Color(a*p.red(), a*p.green(), a*p.blue());
	}

inline Color operator * (double a, const Color& p) {
		return Color(a*p.red(), a*p.green(), a*p.blue());
	}

// Scaling of a color with a real number
inline Color operator / (const Color& p, float a) {
		return Color(p.red()/a, p.green()/a, p.blue()/a);
	}

inline Color operator / (float a, const Color& p) {
		return Color(a/p.red(), a/p.green(), a/p.blue());
	}

// Scaling with another color
inline Color operator * (const Color& a, const Color& b) {
		return Color(a.red()*b.red(), a.green()*b.green(), a.blue()*b.blue());
	}

/* ============================== Color3D ============================= */
/** Color3D that holds 3 variations of the same color for
 * pseudo 3D drawing
 */
class Color3D {
public:
	Color32	color;			/** default color		*/
	Color32	dark;			/** darker color		*/
	Color32	bright;			/** brighter color		*/
public:
	Color3D() { color=0; dark=0; bright=0; }
	Color3D(dword c)		{ (*this)(c); }
	void	operator ()(dword c);
	dword	operator ()()	const	{ return (dword)color; }
}; // class Color3D

#endif
