/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "matrix4.h"
#include <iomanip>

using std::endl;

constexpr double Matrix4::_identity[16];

bool Matrix4::isIdentity() const {
        return !memcmp(data, _identity, sizeof(data));
}

void Matrix4::identity() {
        assert(_identity[0]==1); memcpy(data, _identity, sizeof(data));
}

/** transpose */
void Matrix4::transpose()
{
	for (int i=0; i<rows(); i++)
		for (int j=i+1; j<cols(); j++) {
			double t = M(i,j);
			M(i,j) = M(j,i);
			M(j,i) = t;
		}
} // transpose

/** operator * *
Matrix4& Matrix4::operator *= (const Matrix4& mat)
{
	// FIXME Optimize
	return (*this) = (*this)*mat;
} * operator *= */

/** negate matrix */
void Matrix4::negate()
{
	for (int i=0; i<rows(); i++)
		for (int j=0; j<cols(); j++)
			M(i,j) = -M(i,j);
} // negate

/** return trace of matrix (sum of diagonal elements) */
double Matrix4::trace() const
{
	double t = 0.0;
	for (int i=0; i<4; i++)
		t += M(i,i);
	return t;
} // trace

/** inverse the matrix. */
void Matrix4::inverse()
{
	Matrix4 inv;
	invertMatrix(data, inv.data);
	copy(inv);
} // inverse

/** rotate matrix around an axis
 * @param angle
 * @param axis X=0, Y=1, Z=2
 *
 * WARNING possibly a call to fix should be made after
 */
void Matrix4::rotate(const double angle, const int axis)
{
	double s, c;
	identity();

	bsincos(angle, &s, &c);

	int m1 = ((axis+1)%3)+1;
	int m2 = m1%3;
	m1 -= 1;

	M(m1,m1) =  c;
	M(m2,m2) =  c;
	M(m1,m2) = -s;
	M(m2,m1) =  s;
} // rotate

/** rotate matrix around a vector */
void Matrix4::rotate(const double angle, double x, double y, double z)
{
	double s, c;
	identity();

	bsincos(angle, &s, &c);

	double invl = 1.0/sqrt(x*x + y*y + z*z);
	x *= invl;
	y *= invl;
	z *= invl;

	double c1 = 1.0 - c;
	M(0,0) = x*x + (1-x*x)*c;
	M(0,1) = x*y*c1 - z*s;
	M(0,2) = x*z*c1 + y*s;

	M(1,0) = x*y*c1 + z*s;
	M(1,1) = y*y + (1-y*y)*c;
	M(1,2) = y*z*c1 - x*s;

	M(2,0) = x*z*c1 - y*s;
	M(2,1) = y*z*c1 + x*s;
	M(2,2) = z*z + (1-z*z)*c;
} // rotate

/** @return determinate of matrix */
double Matrix4::det() const
{
	double d12 = ab_cd(M(2,0),M(3,1), M(3,0),M(2,1));
	double d13 = ab_cd(M(2,0),M(3,2), M(3,0),M(2,2));
	double d23 = ab_cd(M(2,1),M(3,2), M(3,1),M(2,2));
	double d24 = ab_cd(M(2,1),M(3,3), M(3,1),M(2,3));
	double d34 = ab_cd(M(2,2),M(3,3), M(3,2),M(2,3));
	double d41 = ab_cd(M(2,3),M(3,0), M(3,3),M(2,0));

	double t0 =  (M(1,1) * d34 - M(1,2) * d24 + M(1,3) * d23);
	double t1 = -(M(1,0) * d34 + M(1,2) * d41 + M(1,3) * d13);
	double t2 =  (M(1,0) * d24 + M(1,1) * d41 + M(1,3) * d12);
	double t3 = -(M(1,0) * d23 - M(1,1) * d13 + M(1,2) * d12);

	/* Compute determinant as early as possible using these cofactors. */
	return M(0,0) * t0 + M(0,1) * t1 + M(0,2) * t2 + M(0,3) * t3;
} // det

#define MAT(m,r,c) (m)[(c)*4+(r)]
/* Here's some shorthand converting standard (row,column) to index. */
#define m11 MAT(m,0,0)
#define m12 MAT(m,0,1)
#define m13 MAT(m,0,2)
#define m14 MAT(m,0,3)
#define m21 MAT(m,1,0)
#define m22 MAT(m,1,1)
#define m23 MAT(m,1,2)
#define m24 MAT(m,1,3)
#define m31 MAT(m,2,0)
#define m32 MAT(m,2,1)
#define m33 MAT(m,2,2)
#define m34 MAT(m,2,3)
#define m41 MAT(m,3,0)
#define m42 MAT(m,3,1)
#define m43 MAT(m,3,2)
#define m44 MAT(m,3,3)

/* From Mesa-2.2\src\glu\project.c
 * Compute the inverse of a 4x4 matrix.  Contributed by scotter@lafn.org
 */
void Matrix4::invertMatrixGeneral(const double* m, double* out)
{
	double tmp[16]; /* Allow out == in. */

	// Inverse = adjoint / det. (See linear algebra texts.)
	// pre-compute 2x2 dets for last two rows when computing
	// cofactors of first two rows.
	double d12 = ab_cd(m31,m42,m41,m32);
	double d13 = ab_cd(m31,m43,m41,m33);
	double d23 = ab_cd(m32,m43,m42,m33);
	double d24 = ab_cd(m32,m44,m42,m34);
	double d34 = ab_cd(m33,m44,m43,m34);
	double d41 = ab_cd(m34,m41,m44,m31);

	tmp[0] =  (m22 * d34 - m23 * d24 + m24 * d23);
	tmp[1] = -(m21 * d34 + m23 * d41 + m24 * d13);
	tmp[2] =  (m21 * d24 + m22 * d41 + m24 * d12);
	tmp[3] = -(m21 * d23 - m22 * d13 + m23 * d12);

	/* Compute determinant as early as possible using these cofactors. */
	double det = m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2] + m14 * tmp[3];

	/* Run singularity test. */
	if (det == 0.0) {
		/* printf("invert_matrix: Warning: Singular matrix.\n"); */
		memcpy(out, _identity, sizeof(_identity));
	} else {
		double invDet = 1.0 / det;

		/* Compute rest of inverse. */
		tmp[0] *= invDet;
		tmp[1] *= invDet;
		tmp[2] *= invDet;
		tmp[3] *= invDet;

		tmp[4] = -(m12 * d34 - m13 * d24 + m14 * d23) * invDet;
		tmp[5] =  (m11 * d34 + m13 * d41 + m14 * d13) * invDet;
		tmp[6] = -(m11 * d24 + m12 * d41 + m14 * d12) * invDet;
		tmp[7] =  (m11 * d23 - m12 * d13 + m13 * d12) * invDet;

		/* Pre-compute 2x2 dets for first two rows when computing */
		/* cofactors of last two rows. */
		d12 = ab_cd(m11,m22, m21,m12);
		d13 = ab_cd(m11,m23, m21,m13);
		d23 = ab_cd(m12,m23, m22,m13);
		d24 = ab_cd(m12,m24, m22,m14);
		d34 = ab_cd(m13,m24, m23,m14);
		d41 = ab_cd(m14,m21, m24,m11);

		tmp[8]  =  (m42 * d34 - m43 * d24 + m44 * d23) * invDet;
		tmp[9]  = -(m41 * d34 + m43 * d41 + m44 * d13) * invDet;
		tmp[10] =  (m41 * d24 + m42 * d41 + m44 * d12) * invDet;
		tmp[11] = -(m41 * d23 - m42 * d13 + m43 * d12) * invDet;
		tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23) * invDet;
		tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13) * invDet;
		tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12) * invDet;
		tmp[15] =  (m31 * d23 - m32 * d13 + m33 * d12) * invDet;

		memcpy(out, tmp, sizeof(tmp));
	}
} // invertMatrixGeneral

//
// Invert matrix m.  This algorithm contributed by Stephane Rehel
// <rehel@worldnet.fr>
//
void Matrix4::invertMatrix(const double* m, double* out)
{
	double det;
	double tmp[16]; /* Allow out == in. */

	if (m41 != 0.0 || m42 != 0.0 || m43 != 0.0 || m44 != 1.0) {
		invertMatrixGeneral(m, out);
		return;
	}

	/* Inverse = adjoint / det. (See linear algebra texts.)*/
	tmp[0]= ab_cd(m22, m33, m23, m32);
	tmp[1]= ab_cd(m23, m31, m21, m33);
	tmp[2]= ab_cd(m21, m32, m22, m31);

	/* Compute determinant as early as possible using these cofactors. */
	det= m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2];

	/* Run singularity test. */
	if (det == 0.0) {
		/* printf("invert_matrix: Warning: Singular matrix.\n"); */
		memcpy(out, _identity, sizeof(_identity));
	} else {
		double d12, d13, d23, d24, d34, d41;
		double im11, im12, im13, im14;

		det= 1. / det;

		/* Compute rest of inverse. */
		tmp[0] *= det;
		tmp[1] *= det;
		tmp[2] *= det;
		tmp[3]  = 0.;

		im11= m11 * det;
		im12= m12 * det;
		im13= m13 * det;
		im14= m14 * det;
		tmp[4] = ab_cd(im13, m32, im12, m33);
		tmp[5] = ab_cd(im11, m33, im13, m31);
		tmp[6] = ab_cd(im12, m31, im11, m32);
		tmp[7] = 0.;

		/* Pre-compute 2x2 dets for first two rows when computing */
		/* cofactors of last two rows. */
		d12 = im11*m22 - m21*im12;
		d13 = im11*m23 - m21*im13;
		d23 = im12*m23 - m22*im13;
		d24 = im12*m24 - m22*im14;
		d34 = im13*m24 - m23*im14;
		d41 = im14*m21 - m24*im11;

		tmp[8]  =  d23;
		tmp[9]  = -d13;
		tmp[10] = d12;
		tmp[11] = 0.0;

		tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23);
		tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13);
		tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12);
		tmp[15] =  1.0;

		memcpy(out, tmp, 16*sizeof(double));
	}
} // invertMatrix
#undef m11
#undef m12
#undef m13
#undef m14
#undef m21
#undef m22
#undef m23
#undef m24
#undef m31
#undef m32
#undef m33
#undef m34
#undef m41
#undef m42
#undef m43
#undef m44
#undef MAT

/** @return rank  of matrix
 * Warning: destroys the matrix as it applies a rref (reduced row echelon form*/
int Matrix4::rank(const double eps)
{
	bool row_selected[4] = {false, false, false, false};

	int rank_ = 0;
	for (int i=0; i<cols(); ++i) {
		int j;
		for (j=0; j<rows(); ++j)
			if (!row_selected[j] && !Eq0(M(j,i),eps))
				break;

		if (j != rows()) {
			++rank_;
			row_selected[j] = true;
			for (int p=i+1; p<cols(); ++p)
				M(j,p) /= M(j,i);
			for (int k=0; k<rows(); ++k)
				if (k!=j && !Eq0(M(k,i),eps))
					for (int p=i+1; p<cols(); ++p)
						M(k,p) -= M(j,p) * M(k,i);
		}
	}
	return rank_;
} // rank

/** fix matrix to be a valid rotation matrix */
// FIXME look the graphics GEMS-I for orthogonalizing the matrix
void Matrix4::fix01()
{
	// round to 0.0, 1.0 and -1.0 if needed with precision
	// of 1e-10 since 1e-11**2 = limit of precision
	for (int i=0; i<3; i++) {
		for (int j=0; j<3; j++) {
			// Something that should never be done for floating point
			// however up to phi=1.06e-8, cos(phi)==1.0 exactly
			// and sin(phi)>0.0, so cos(phi)**2 + sin(phi)**2 > 1.0
			if (M(i,j) == 1.0) {
				// Set the other two to zero
				M(i,(j+1)%3) = 0.0;
				M(i,(j+2)%3) = 0.0;
				break;
			} else
			if (M(i,j) == -1.0) {
				// Set the other two to zero
				M(i,(j+1)%3) = 0.0;
				M(i,(j+2)%3) = 0.0;
				break;
			}
		}
	}
} // fix01

/** fix matrix to be a valid transformation matrix
 * @param rotation	if true normalize all coefficients to 1, else
 *			assume it is an expansion/skew matrix so leave them as is
 */
void Matrix4::fix(bool rotation)
{
	fix01();

	// XXX FIXME Error normalizing to 1 since we can have expansion or skew matrices

	// Try to correct (*this) for possible numerical precision problems
	// normalize vectors
	double len[3];
	for (int i=0; i<3; i++) {
		len[i] = sqrt(Sq(M(i,0)) + Sq(M(i,1)) + Sq(M(i,2)));
		assert(len[i] > 1e-10);
		M(i,0) /= len[i];
		M(i,1) /= len[i];
		M(i,2) /= len[i];
	}

	// Perform a Gram Schmidt orthogonalization of vectors
	Vector X = basisX();
	Vector Y = basisY();
	basisZ().gramSchmidt(X,Y);
	X.renormalize();
	Y.renormalize();
	basisX(X);
	basisY(Y);

	assert(sqrt(Sq(M(2,0)) + Sq(M(2,1)) + Sq(M(2,2)))>1e-10);
	len[2] /= sqrt(Sq(M(2,0)) + Sq(M(2,1)) + Sq(M(2,2)));

	if (not rotation)
		for (int i=0; i<3; i++) {
			M(i,0) *= len[i];
			M(i,1) *= len[i];
			M(i,2) *= len[i];
		}

	// force last row
	M(3,0) = 0.0;
	M(3,1) = 0.0;
	M(3,2) = 0.0;
	M(3,3) = 1.0;

	fix01();
} // fix

/** matrix print out */
std::ostream& operator << (std::ostream& s, const Matrix4& matrix)
{
	for (int r=0; r<matrix.rows(); r++) {
		if (r==0)
			s << "/";
		else
		if (r+1==matrix.rows())
			s << "\\";
		else
			s << "|";
		for (int c=0; c<matrix.columns(); c++) {
			s << " " << std::setw(22) << std::setprecision(16) << matrix(r,c);
		}
		if (r==0)
			s << " \\" << endl;
		else
		if (r+1==matrix.rows())
			s << " /" << endl;
		else
			s << " |" << endl;
	}
	return s;
} /* operator << */
