/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 *					Christian.Theis@cern.ch (blue noise distributions and LD samples)
 * Date:	26-Jul-2022
 */

#include "eps.h"
#include "random.h"
#include "niederreiter.h"

/**
 * return a random angle sin and cos from 0..2PI
 * @param s	random sin
 * @param c	random cosine
 */
void Random::sincos(double* s, double* c)
{
	double x,y,x2,y2,d;
	do {
		x  = 2.0*real() - 1.0;
		y  = 2.0*real() - 1.0;
		x2 = Sq(x);
		y2 = Sq(y);
		d  = x2 + y2;
	} while (d > 1.0);

	*s = 2.0*x*y / d;
	*c = (x2 - y2) / d;
} // sincos

/**
 * @return a normal distributed random number generator
 * based on particle data book
 */
double Random::normal()
{
	if (_normalSave) {
		_normalSave = false;
		return _normal;
	}

	double z1, z2;
	normal(&z1, &z2);

	_normal = z2;
	_normalSave = true;

	return z1;
} // normal

/**
 * return 2 normal distributed random numbers
 * @param z1,z2
 */
void Random::normal(double* z1, double* z2)
{
#if 0
	// Slower variant
	double r1 = PI2*real();
	double r2 = sqrt(-2.0*log(real()));
	*z1 = sin(r1) * r2;
	*z2 = cos(r2) * r2;
#endif
	double u1, u2;
	double r2;

	do {	// calculate sin/cos like above method sincos()
		u1 = 2.0*real() - 1.0;
		u2 = 2.0*real() - 1.0;
		r2 = u1*u1 + u2*u2;
	} while (r2 > 1.0);

	r2 = sqrt(-2.0*log(r2)/r2);
	*z1 = u1 * r2;
	*z2 = u2 * r2;
} // normal

/** sample uniform x,y coordinates inside a disc or radius r
 * @param r	radius of disc
 * @param x,y	return position
 */
void Random::disc(const double r, double* x, double* y)
{
	double R = r * sqrt(real());
	double s,c;
	sincos(&s, &c);
	*x = R * c;
	*y = R * s;
} // disc

/** sample uniform x,y,z coordinates inside a sphere of radius r
 * @param r	radius of sphere
 * @param x,y,z	return position
 */
void Random::sphere(const double r, double* x, double* y, double* z)
{
	double R = r * pow(real(), 1.0/3.0);

	double sphi, cphi;
	sincos(&sphi, &cphi);

	double cth = real(-1.0, 1.0);
	double sth = sqrt(1.0 - cth*cth);

	*x = R * sth * cphi;
	*y = R * sth * sphi;
	*z = R * cth;
} // sphere

/**
 * create an isotropic random vector of unit length using a
 * rejection technique
 * @return *this
 */
Vector Random::vector()
{
	double x, y, z;
	double len;

	do {
		x = 2.0*real() - 1.0;
		y = 2.0*real() - 1.0;
		z = 2.0*real() - 1.0;
		len = x*x + y*y + z*z;
	} while (len > 1.0);

	len = sqrt(1.0/len);
	x *= len;
	y *= len;
	z *= len;
	return Vector(x,y,z);
} // vector

/** sample random position and direction on a unit sphere to generate
 * a uniform and isotropic fluence
 * @param pos	position to return
 * @param dir	direction to return
 * @param rnd1 & rnd2
 *		uniform random nr. used for sampling.
 *		by default the internal RND is used but this can
 *		be substituted by QMC generators for better & faster convergence
 */
void Random::sflood(Point* pos, Vector* dir, double rnd1, double rnd2)
{
	//assert(rnd1 != rnd2);
	// Random position direction
	double costh0 = 2.0 * rnd1 - 1.0;
	double sinth0 = sqrt((1.0-costh0) * (1.0+costh0));
	double sinph0, cosph0;
	this->sincos(&sinph0, &cosph0);

	// These are the position vector components
	pos->set(cosph0*sinth0, sinph0*sinth0, costh0);

	// Now sample the direction vector
	double snth   = sqrt(rnd2);
	double csth   = sqrt(1.0 - rnd2 );
	double snph, csph;
	this->sincos(&snph, &csph);

	double tr = -csth;
	double tp = -snth * snph;
	double tt = -snth * csph;

	if (sinth0 > TOOSMALL) {
		dir->x = pos->x * tr + (pos->x * pos->z * tt - pos->y * tp) / sinth0;
		dir->y = pos->y * tr + (pos->y * pos->z * tt + pos->x * tp) / sinth0;
		dir->z = pos->z * tr - sinth0 * tt;
	} else
		dir->set(tt, tp, pos->z * tr);
} // sflood

/** @return a real stratified random number [0,1) */
double Random::realqmc()
{
	return Niederreiter2( static_cast<long int*>(&_qmc_seed) );
} // realqmc

/** @return an N-dimensional tuple of real stratified random numbers [0,1)
 * based on Niederreiter2 sequences
 */
void Random::realqmcND( int dim, double randVec[] )
{
	Niederreiter2ND( dim, static_cast<long int*>(&_qmc_seed), randVec );
} // realqmcND

/** @return a vector of pairs that contain samples with blue noise that can be used for dithering.
 *  We generate a LUT of blue noise using Mitchell's Best candidate algorithm, which practically
 *  samples a Poisson disk by dart throwing and retaining those samples where the toroidally rotated
 *  distance is maximized.
 *
 *  This is relatively simple and gives satisfying results.
 *  Another (better) possibility would be to start from white noise which is tempered according to
 *  an energy function as proposed by Georgiev & Fajardo, see:
 *  https://www.arnoldrenderer.com/research/dither_abstract.pdf
 */
std::vector<std::array<double, 2>> Random::generateBlueNoise( size_t nrOfSamples, const int cSampleMultiplier /*= 5*/ )
{
	std::vector<std::array<double, 2>> samplesPos;

	for(size_t i = 0; i < nrOfSamples; ++i) {

		// keep the candidate that is farthest from it's closest point
		size_t numCandidates = samplesPos.size() * cSampleMultiplier + 1; // for better quality of the final distribution we increase the sample # by a multiplier
		double bestDistance = 0.0;                                        // in order to keep the quality of the distribution consistent we increase the number of candidates
		double bestCandidateX = 0;                                        // proportionally to the # of samples already successfully stored and used as reference.
		double bestCandidateY = 0;

		for(size_t j = 0; j < numCandidates; ++j) {
			double x = real();
			double y = real();

			// calculate the closest distance from this point to an existing sample
			double minDist = INFINITY;
			for(const auto& curSample : samplesPos) {
				double curDist = ToroidalDistSquared( x, y, curSample[0], curSample[1]);
				minDist = std::min( minDist, curDist );
			}

			// retain the best candidate with the maximized distance
			if(minDist > bestDistance) {
				bestDistance = minDist;
				bestCandidateX = x;
				bestCandidateY = y;
			}
		}

		samplesPos.push_back( { bestCandidateX, bestCandidateY } );
	}

	return samplesPos;
} // generateBlueNoise
