/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author: Vasilis.Vlachoudis@cern.ch
 * Date:   2 Mar 2014
 */

#ifndef __TOOL_MEMORY_H
#define __TOOL_MEMORY_H

#include <pthread.h>
#include <new>

#define _MAGIC	0xDECAFFEE
#define _MAGIC3	0xDE
#define _MAGIC2	0xCA
#define _MAGIC1	0xFF
#define _MAGIC0	0xEE

//#define _MEMORY_STAMP

//namespace MemoryCheck {
/*
 * This file provides some debugging functions for memory allocation
 */
struct Memory {
private:
	const char*	name;
	Memory*		next;
	Memory*		prev;
	size_t		size;
	size_t		line;
	unsigned	magic;
	unsigned char	data[sizeof(unsigned)];

public:
static	void	init();
static	bool	stop();

static	void*	malloc(size_t size, const char* filename, const size_t lineno);
static	void*	realloc(void* ptr, size_t size);
static	void	free(void* ptr);

static	void	dump();
static	void	coredump();
static	void	stamp(void* ptr, const char* filename, const size_t lineno /*const char* typename*/);

static	void	check();
static	size_t	allocated()		{ return total; }
static	size_t	items()			{ return n; }
static	double	resident();

#ifdef _MEMORY_STAMP
	class Stamp {
		public:
			const char* filename;
			int  line;
			void* ptr;
		public:
			Stamp(const char* fn, const int n) : filename(fn), line(n) {}
			~Stamp() {}
			operator bool() const {return false; }
	}; // class Stamp
#endif

private:
	void	setTail() {
			data[size]   = _MAGIC0;
			data[size+1] = _MAGIC1;
			data[size+2] = _MAGIC2;
			data[size+3] = _MAGIC3;
		}
	bool	checkTail() {
			return	data[size]   == _MAGIC0 &&
				data[size+1] == _MAGIC1 &&
				data[size+2] == _MAGIC2 &&
				data[size+3] == _MAGIC3;
		}
	void	print(int count);

private:
static	void	lock()		{ pthread_mutex_lock(&mutex);   }
static	void	unlock()	{ pthread_mutex_unlock(&mutex); }
static	pthread_mutex_t	mutex;	// locking mutex when running through threadpool

static	bool	_init;
static	Memory*	head;
static	size_t	total;
static	size_t	n;
//public:
//static	const char*	__file__;
//static	size_t		__line__;
}; // Memory

#ifdef _MEMORY_STAMP
template <typename T>
//inline T* operator*(const Memory::Stamp& stamp, T* p)
inline const Memory::Stamp& operator*(const Memory::Stamp& stamp, T* p)
{
	stamp.ptr = p;
	Memory::stamp(p, stamp.filename, stamp.line); //, typeid(T).name());
	return stamp;
	//return p;
}

template <typename T>
inline T* operator!(const Memory::Stamp& stamp)
{
	return (T*)stamp.ptr;
}
#endif

inline void  operator delete(void* ptr)   noexcept { Memory::free(ptr); }
inline void  operator delete[](void* ptr) noexcept { Memory::free(ptr); }

/* C++ -Wsized-deallocation requests to implement this */
inline void  operator delete(void* ptr, std::size_t)   noexcept { Memory::free(ptr); }
inline void  operator delete[](void* ptr, std::size_t) noexcept { Memory::free(ptr); }

inline void* operator new(size_t size)
		{ return Memory::malloc(size, __builtin_FILE(), __builtin_LINE()); }
inline void* operator new[](size_t size)
		{ return Memory::malloc(size, __builtin_FILE(), __builtin_LINE()); }

#if 0
inline void* operator new(size_t size, const char* filename, const int line)
		{ return Memory::malloc(size, filename, line); }
inline void* operator new[](size_t size, const char* filename, const int line)
		{ return Memory::malloc(size, filename, line); }
#define new new(__FILE__,__LINE__)
#endif

// All of them fail since in STL there are calls like
// ::operator new(size) or ::new
// and the expansion is generating something which is not valid any more
// ::operator Memory::Stamp() * new
#ifdef _MEMORY_STAMP
//#define new (Memory::__file__=__FILE__, Memory::__line__=__LINE__) && 0 ? NULL : new
//#define new Memory::Stamp(Memory::__file__=__FILE__, Memory::__line__=__LINE__) && 0 ? NULL : new
//#define new Memory::Stamp(__FILE__, __LINE__) ? NULL : new
//#define new Memory::Stamp(__builtin_FILE(), __builtin_LINE()) * new
//#define new !Memory::Stamp(__FILE__,__LINE__)*new
#endif

#endif
