/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#include "token.h"

#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <fstream>

#include "tool/bstring.h"

using std::cerr;
using std::endl;

const static char* tokenType[] = {
			"EOF",
			";",
			"ident",
			"string",
			"file",
			"function",
			"number",
			",",
			":",
			"=",
			"+",
			"-",
			"*",
			"/",
			"^",
			"|",
			"&",
			"{",
			"}",
			"[",
			"]",
			"(",
			")",
			"$",
			"#"
		};

/* init */
void Token::init(bool inl, bool icase)
{
	newline        = true;
	linenum        = 0;
	colpos         = 0;
	type           = Type::UNKNOWN;
	num            = 0.0;
	comment        = '\0';          // comment char = nothing
	str[0]         = 0;
	allow_sign     = true;
	ignore_case    = icase;
	ignore_newline = inl;
	newline_token  = Type::NEWLINE; // default
	token_pushed   = false;

	// FIXME add isalnum() as SUFFIX and isalpha() as ANY
	memset(extended_alpha, 0, sizeof(extended_alpha));
	setAlpha('_', ANY);
} // init

/** push token if not needed */
void Token::push()
{
	token_pushed = true;
} // push

/** parseNumber */
bool Token::parseNumber()
{
	int ch;
	bool dadded = isdigit(str[0]);	// digit added?

	ch = nextchar();

	if (str[0] != '.') {
		// accept many digits before possible dot
		while (isdigit(ch)) {
			ch = addnext(ch);
			dadded = true;
		}

		// accept one dot
		if (ch=='.')
			ch = addnext(ch);
	}

	// accept many digits after dot
	while (isdigit(ch)) {
		ch = addnext(ch);
		dadded = true;
	}

	// accept exponent e or E
	if (dadded and (ch=='e' or ch=='E')) {
		ch = addnext(ch);

		// accept sign
		if (ch=='+' || ch=='-')
			ch = addnext(ch);

		// must have many digits
		while (isdigit(ch))
			ch = addnext(ch);
	}
	putback(ch);	// put back last char

	if (str[1]==0 and not isdigit(str[0]))	// nothing was added
		return false;

	char* endptr;
	num = strtod(str, &endptr);
	if (*endptr!=0) return false;

	type = Type::NUMBER;
	return true;
} // parseNumber

/** next
 * Gets the next token from the input stream and
 * returns it's value.  If the previous token has been
 * pushed back, then it is returned again.
 * @return next token
 */
Token::Type Token::next()
{
	int ch;

	if (token_pushed) {
		token_pushed = false;
		return type;
	}

	reset();

	while (newline) {
		newline = false;
		ch = nextchar();
		if (ch==comment) {	// skip lines starting with comment char
			while ((ch=nextchar())!='\n')
				if (ch==EOF) break;

			linenum++;
			newline = true;
			colpos  = 0;
		} else
			putback(ch);
	}

	/* ignore white space */
	while (isspace(ch=nextchar())) {
		if (ch == '\n') {
			linenum++;
			newline = true;
			colpos  = 0;
			if (!ignore_newline)
				return type=newline_token;
		}
	}

	// FIXME check for end of file
	add(ch);	// store in string

	// decide what next based on following char
	if (isalpha(ch) || extended_alpha[(byte)ch]==ANY) {         /* must be a keyword */
		ch = nextchar();
		while (isalnum(ch) || extended_alpha[(byte)ch])
			ch = addnext(ch);
		if (ch=='(')
			type = Type::FUNCTION;
		else {
			if (ignore_case) upper();
			type = Type::IDENT;
			// push back the character that doesn't belong
			putback(ch);
		}
		return type;
	} else
	switch (ch) {
		case EOF:
			return type = Type::ENDOFILE;

		case '{':
			return type = Type::LEFT_BRACE;

		case '}':
			return type = Type::RIGHT_BRACE;

		case '[':
			return type = Type::LEFT_BRACKET;

		case ']':
			return type = Type::RIGHT_BRACKET;

		case '(':
			return type = Type::LEFT_PAREN;

		case ')':
			return type = Type::RIGHT_PAREN;

		case ';':
			return type = Type::SEMICOLON;

		case ':':
			return type = Type::COLON;

		case '+':
			if (allow_sign and parseNumber())
				return type;
			ch = nextchar();
			if (ch=='+')
				return type = Type::PLUS_PLUS;
			putback(ch);
			return type = Type::PLUS;

		case '-':
			if (allow_sign and parseNumber())
				return type;
			ch = nextchar();
			if (ch=='-')
				return type = Type::MINUS_MINUS;
			putback(ch);
			return type = Type::MINUS;

		case '0': case '1':
		case '2': case '3':
		case '4': case '5':
		case '6': case '7':
		case '8': case '9':
			if (parseNumber())
				return type;
			return type = Type::UNKNOWN;

		case '.':
			if (parseNumber())
				return type;
			return type = Type::DOT;

		case '*':
			return type = Type::MULT;

		case '/':
			return type = Type::DIV;

		case '^':
			return type = Type::NOT;

		case '|':
			return type = Type::OR;

		case '=':
			return type = Type::EQUAL;

		case ',':
			return type = Type::COMMA;

		case '$':
			return type = Type::DOLLAR;

		case '#':
			return type = Type::HASH;

		case '\"':
			while ((ch = nextchar()) != '\"')
				add(ch);
			return type = Type::STRING;

		case '<':
			while ((ch = nextchar()) != '>') {
				add(ch);
			}
			return type = Type::FILE_STRING;

		default:
			return type = Type::UNKNOWN;
	}

	// if we get down here something is really wrong
	error(sformat("Invalid character '%s' found", (char)ch));
	return type;	// never gets here but keeps compiler happy
} // next

/** printErrorLine
 * if an error is found print current line with a pointer to the wrong token
 */
void Token::printErrorLine(std::ostream& os)
{
	// Move file pointer "colpos" back and display current line
	stream.seekg(-(long)colpos-1, std::ios::cur);
	os.width(5);
	os << linenum << " *-* ";
	os.width();
	int col = 0;
	int pos = 0;
	int colerr = 0;
	while (1) {
		int ch = stream.get(); pos++;
		if (pos==colpos) colerr = col;
		if (ch=='\n' || ch==EOF) break;
		if (ch=='\t') {
			ch = 8-col%8;
			for (int i=0; i<ch; i++) os << ' ';
			col += ch;
		} else {
			os << (char)ch;
			col++;
		}
	}
	colerr += 10;
	os << endl;

	for (int i=0; i<colerr; i++) os << ' ';
	os << '^' << endl;
} // printErrorLine

/** error handler
 * @param errmsg error message to display
 */
void Token::error(const std::string& errmsg)
{
	printErrorLine(cerr);
	cerr << "Error: line # " << linenum << ": " << errmsg << endl;
	//assert(0);
} // error

/** setAlpha */
void Token::setAlpha(const char* st, AlphaNumeric f)
{
	for (const char* c=st; *c; c++)
		extended_alpha[(byte)(*c)] = f;
} // setAlpha

/** skipSemicolon */
void Token::skipSemicolon()
{
	while (type==Type::SEMICOLON) next();
} // skipSemicolon

/** skipLine */
void Token::skipLine()
{
	int ch;
	do {
		ch = stream.get();
	} while (ch!='\n' && ch!=-1);
	next();
} // skipLine

/** nextSkipEqual */
void Token::nextSkipEqual()
{
	next();
	if (type == Type::COLON || type==Type::EQUAL)
		next();
} // nextSkipEqual

/** mustbe
 * @param tok next token must be tok otherwise display error
 * @param errmsg error message to display
 */
bool Token::mustbe(const Type tok, const std::string& errmsg)
{
	if (type != tok)
		error(errmsg);
	else {
		next();
		return true;
	}
	return false;
} // mustbe

/** mustbe
 * @param tok next token must be tok otherwise display error
 * @param errmsg error message to display
 */
bool Token::mustbe(const char* tokstr, const std::string& errmsg)
{
	if (type != Type::IDENT)
		error(errmsg);
	else
	if (strcmp(ident(), tokstr))
		error(errmsg);
	else {
		next();
		return true;
	}
	return false;
} // mustbe

/** mustbe
 * @param tok next token must be tok otherwise display error
 * @param errmsg error message to display
 */
bool Token::mustbe(const char tokchar, const std::string& errmsg)
{
	if (character() != tokchar)
		error(errmsg);
	else {
		next();
		return true;
	}
	return false;
} // mustbe

/**  upper - convert to uppercase
 */
void Token::upper()
{
	for (char* ch=str; *ch; ch++)
		*ch = toupper(*ch);
} // upper

/** cmp
 * Perform an abbreviated comparison
 * a star '*' marks the start of abbreviation
 * @param teststr	abbreviation to check token
 * @return true/false if it matches the abbreviation
 */
bool Token::cmp(const char* teststr)
{
	char* ch = str;

	// until the star must be exactly equal
	while (*ch && *teststr && *ch == *teststr) {
		ch++;
		teststr++;
	}
	// skip the star
	if (*teststr=='*')
		teststr++;
	else
	if (*teststr || *ch)
		return false;
	else
		return true;

	// compare abbreviated
	while (*ch && *teststr && *ch == *teststr) {
		ch++;
		teststr++;
	}
	if (*ch == 0)
		return true;
	else
		return false;
} // cmp

/** @return a string or identifier */
std::string Token::getIdent()
{
	if (type == Type::IDENT) {
		std::string s(str);
		next();
		return s;
	} else {
		error("Identifier expected");
		return std::string();
	}
} // getString

/** getString, get string from file
 * @param getstr	string to return
 * @param maxsize	maximum size of getstr
 */
void Token::getString(char* getstr, const int maxsize)
{
	if (In(type, Type::STRING, Type::IDENT)) {
		strncpy(getstr, str, maxsize);
		next();
	} else
		error("String expected");
} // getString

/** @return a string or identifier */
std::string Token::getString()
{
	if (type == Type::STRING || type == Type::IDENT) {
		std::string s(str);
		next();
		return s;
	} else {
		error("String expected");
		return std::string();
	}
} // getString

/** getInteger
 * @return next integer token
 */
int Token::getInteger()
{
	if (type == Type::NUMBER) {
		int n = (int)num;
		if (!EqRel((double)n, num, 1.0e-15))
			error("Integer expected");
		next();
		return n;
	} else
		error("Integer expected");
	return 0;	// Never gets here but keeps compiler happy
} // getInteger

/** getBoolean
 * @return next boolean token
 */
bool Token::getBoolean()
{
	if (type == Type::NUMBER) {
		int n = (int)num;
		if ((!EqRel((double)n, num, 1.0e-15)) || (n!=0 && n!=1))
			error("Invalid boolean");
		next();
		return n;
	} else
	if (type == Type::STRING || type == Type::IDENT) {
		upper();
		if (cmp("T*RUE") || cmp("ON") || cmp("YES")) {
			next();
			return true;
		}
		if (cmp("F*ALSE") || cmp("OFF") || cmp("NO")) {
			next();
			return false;
		}
	}
	error("Invalid boolean");
	return 0;	// Never gets here but keeps compiler happy
} // getBoolean

/** getNumber
 * @return next floating point number
 */
double Token::getNumber()
{
	if (type == Type::NUMBER) {
		double	n = num;
		next();
		return n;
	} else
		error("Number expected");
	return 0.0;	// Never gets here but keeps compiler happy
} // getNumber

/** getUpperIdent
 * @param errmsg error message to display
 * @return next identifier as uppercase
 */
bool Token::getUpperIdent(const std::string& errmsg)
{
	skipSemicolon();
	if (current() == Type::RIGHT_BRACE) {
		next();
		return false;
	}

	if (current() != Type::IDENT) error(errmsg);
	upper();
	return true;
} // getUpperIdent

/** print token */
std::ostream& operator << (std::ostream& st, const Token& token)
{
	st << "Token ";
	st << " Type=" << tokenType[(int)token()];
	st << " Value=\"" << token.string() << '"';
	return st;
} /* operator << */
