/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __TOOL_TOKEN_H
#define __TOOL_TOKEN_H

#include <ostream>
#include <iostream>

#include "os.h"


/*-------------------------------------------- */
/*                Token                        */
/*-------------------------------------------- */
class Token {
public:
	enum class Type {
		ENDOFILE,	//  0
		NEWLINE,	//  1
		SEMICOLON,	//  2 ;
		IDENT,		//  3 identifier
		STRING,		//  4 "string"
		FILE_STRING,	//  5 <string> file string, to be searched
		FUNCTION,	//  6 ident(
		NUMBER,		//  7 0.12345789E-5
		COMMA,		//  8 ,
		DOT,		//  9 .
		COLON,		// 10 :
		EQUAL,		// 11 =
		PLUS,		// 12 +
		PLUS_PLUS,	// 13 ++
		MINUS,		// 14 -
		MINUS_MINUS,	// 15 --
		MULT,		// 16 *
		DIV,		// 17 /
		NOT,		// 18 ^
		OR,		// 19 |
		AND,		// 20 &
		LEFT_BRACE,	// 21 {
		RIGHT_BRACE,	// 22 }
		LEFT_BRACKET,	// 23 [
		RIGHT_BRACKET,	// 24 ]
		LEFT_PAREN,	// 25 (
		RIGHT_PAREN,	// 26 )
		DOLLAR,		// 27 $
		HASH,		// 28 #
		UNKNOWN		// 29 anything else
	};

	enum AlphaNumeric {	// alphanumeric symbols
		NONE   = 0,	// not accepted
		SUFFIX = 1,	// except first character
		ANY    = 2,	// any position even as first
	};

protected:
	std::istream&	stream;		/* input stream				*/
	bool	newline;		/* newline character found		*/
	int	linenum;		/* current line position		*/
	int	colpos;			/* current line column			*/
	Type	type;			/* current token			*/
	bool	token_pushed;		/* saved token				*/

	double	num;			/* current floating point value		*/
	char	str[256];		/* current string value			*/
	AlphaNumeric extended_alpha[256]; /* alphanumeric chars in identifiers	*/

private:
	int	ic;			/* character pointer position		*/

public:
	bool	allow_sign;		/* allow numbers to have a sign +/-	*/
	bool	ignore_case;		/* ignore character case		*/
	bool	ignore_newline;		/* ignore new lines			*/
	char	comment;		/* comment character at start of line	*/
	Type	newline_token;		/* what token to return on new line	*/
					/* if it is not ignored			*/

public:
	Token(std::istream& st) : stream(st) { init(); }
virtual	~Token() {}

	void	printErrorLine(std::ostream& os);
virtual	void	error(const std::string& errormsg);
	/** extended alphanumeric characters to be considered in identifiers */
	void	setAlpha(byte ch, AlphaNumeric f=SUFFIX)	{ extended_alpha[ch] = f; }
	void	setAlpha(const char* st, AlphaNumeric f=SUFFIX);

	void	push();
	Type	next();
		operator bool()	const	{ return type != Type::ENDOFILE; }
	Type	operator ()() const	{ return type; }
	Type	current() const		{ return type; }
	bool	is(const Type t)	{ return type == t; }

	double	number() const		{ return num; }
	int	integer() const		{ return (int)num; }
const	char*	string() const		{ return str; }
const	char*	ident()	 const		{ return str; }
	char	character() const	{ return str[0]; }
	void	upper();		// Translate str to uppercase
	bool	cmp(const char *);	// abbreviated comparison

	void	skipSemicolon();
	void	skipLine();
	void	nextSkipEqual();

	bool	mustbe(const Type, const std::string&);
	bool	mustbe(const char*, const std::string&);
	bool	mustbe(const char, const std::string&);
	bool	isInteger() const {
				if (type != Type::NUMBER) return false;
				int n = (int)num;
				return EqRel((double)n, num, 1.0e-15);
			}

	std::string	getIdent();
	void	getString(char*, const int);
	std::string	getString();
	int	getInteger();
	bool	getBoolean();
	double	getNumber();
	bool	getUpperIdent(const std::string& errmsg);

	std::istream&	getStream() const { return stream; }

private:
	/** initialize token
	 * @param inl	ignore new lines
	 * @param ic	ignore case
	 */
	void	init(bool inl=false, bool icase=false);

	/** reset string buffer */
	void	reset()	 {
			ic = 0;
			str[ic] = 0;
		}

	/** add one character */
	void	add(int ch) {
			str[ic++] = (char)ch;
			str[ic] = 0;
		}

	/** add one character and fetch new one from stream */
	int	addnext(int ch) {
			add(ch);
			return nextchar();
		}

	/** fetch new one from stream */
	int	nextchar() {
			int ch = stream.get();
			colpos++;
			return ch;
		}

	/** peek the next character in stream */
	int	peek() {
			int ch = stream.get();
			stream.putback(ch);
			return ch;
		}

	/** putback a character in stream */
	void	putback(int ch) {
			stream.putback(ch);
			colpos--;
		}

	/** parse a number */
	bool	parseNumber();
}; // Token

std::ostream& operator << (std::ostream& st, const Token& token);

#define MUSTBE_SEMICOLON	token.mustbe(Token::Type::SEMICOLON,ERR_EXTRA_DATA)
#endif
