/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 *
 */

#include "d2layer.h"
#include "viewer.h"

#include "tool/timer.h"

using std::cout;
using std::endl;

// maximum line length^2 on curved conic
const double MAXLEN2 = Sq(10.0);

#ifdef THREAD
/* =============================== BodyFeeder =============================== */
/** allocate */
void BodyFeeder::allocate()
{
	if (nworkers != pool.nthreads()) {
		if (workers) delete [] workers;
		nworkers = pool.nthreads();
		workers = new BodyWorker[nworkers];
		for (int i=0; i<nworkers; i++)
			workers[i].feeder(this);
	}
} // allocate

/** reset */
void BodyFeeder::reset(D2Method method, int pfrom, int pto)
{
	pool.reset();
	allocate();
	idx = 0;
	perc_from = pfrom;
	perc_span = pto - pfrom;

	// initialize workers
	for (int i=0; i<nworkers; i++) {
		workers[i].d2     = d2;
		workers[i].method = method;
		workers[i].engine = d2->kernel.engine(i);
		workers[i].body   = NULL;
	}
} // reset

/** feed */
ThreadPoolWorker* BodyFeeder::feed(int threadId)
{
	// Check for ending condition
	if (idx >= d2->kernel.nBodies() || d2->viewer.stop()) return NULL;

	// update worker
	workers[threadId].body = d2->kernel.getBody(idx++);

	// Update information done up to now
	d2->viewer.percent(perc_from + (perc_span*idx)/d2->kernel.nBodies());

	return (ThreadPoolWorker*)&workers[threadId];
} // feed
#endif

/* ============================== D2Layer ============================= */
/** D2Layer */
D2Layer::D2Layer(const Geometry& g, GeometryKernel& k, GeometryViewer& v) :
		Layer(g, k, v),
		labels(32)
#ifdef THREAD
		, feeder(k.threadpool, this)
#endif
{
	_projectAll   = true;
	showVertex    = true;
	showLabels    = true;
	fillRegions   = true;
	_projectionValid = false;

	for (int i=0; i<percMax; i++)
		percIntersect[i] = 20;
} // D2Layer

/** _intersectBody threaded version of intersect body, called from the threadpool */
void D2Layer::_intersectBody(VBody* a, GeometryEngine*)
{
	kernel.intersectBody(a, _projectAll);
} // _intersectBody

/** project - create the projection on the viewport
 * two main cases can happen depending on the value of _projectAll:
 * - if _projectAll is true: all the bodies have to be processed
 * - if _projectAll is false: process only invalid bodies and changed bodies
 */
void D2Layer::project()
{
	if (!show) {
		viewer.state(GeometryViewer::Projection::FINISHED, 100);
		_projectionValid = false;
		return;
	} else
		_projectionValid = true;
	kernel.lock();

	// keep timer information
	Timer timer(true);
	int avgPercIntersect = 0;
	for (int i=0; i<percMax; i++)
		avgPercIntersect += percIntersect[i];
	avgPercIntersect /= percMax;

	// XXX FIXME XXX
	// I could use some adaptive technique from the previous runs to learn
	// the time needed for the plots to make a better estimate
	// e.g. keep the 2 or 3 last runs information
	// and predict using the average for the next ones
	viewer.state(GeometryViewer::Projection::START, 0);
	kernel.clearErrors();

	IFSTAT(kernel.resetStatistics());
	IFSTAT(kernel.stat.start("01 make_body_conics"));
	DUMP(cout << endl << "*-* D2Layer::project("<< viewer.title() <<")" << endl);

	/* generate and transform bodies */
	if (viewer.state(GeometryViewer::Projection::CONICS, 0)) {
		kernel.unlock();
		return;
	}

	// Project conics to viewer plane
	kernel.makeBodyConics(_projectAll);

	// update region location
	kernel.updateRegionLocation(_projectAll);

	// assume a fixed percentage of 1% for Location searching
	// typically is less than that
	if (viewer.state(GeometryViewer::Projection::LOCATION, 1)) {
		kernel.unlock();
		return;
	}

	IFSTAT(kernel.stat.stop("01 make_body_conics"));
	IFSTAT(kernel.stat.start("02 update_region_location"));

	IFSTAT(kernel.stat.stop("02 update_region_location"));
	IFSTAT(kernel.stat.start("03 body_intersections"));

	// Compute all intersections
#ifdef THREAD
	if (kernel.nthreads()) {
		viewer.pState = GeometryViewer::Projection::INTERSECT;
		feeder.reset(&D2Layer::_intersectBody, 1, avgPercIntersect);
		kernel.threadpool.execute(&feeder);
	} else
#endif
	{
		for (size_t i=0; i<kernel.nBodies(); i++) {
			VBody* a = kernel.getBody(i);
			if (a->ignore()) continue;

			if (viewer.state(GeometryViewer::Projection::INTERSECT,
					1 + ((avgPercIntersect-1)*i)/kernel.nBodies())) {
				kernel.unlock();
				return;
			}

			// Generate all body combinations and intersect them
			for (size_t j=i+1; j<kernel.nBodies(); j++) {
				VBody* b = kernel.getBody(j);
				if (b->ignore()) continue;
				if (_projectAll || a->invalid() || b->invalid())
					a->intersectBody(b, view());
			}
		}
	}

	double timeIntersect = timer.elapsedUser();
	IFSTAT(kernel.stat.stop("03 body_intersections"));
	IFSTAT(kernel.stat.start("04 body_cleanup"));

	// clean-up invalid bodies
	for (size_t i=0; i<kernel.nBodies(); i++) {
		VBody* body = kernel.getBody(i);
		body->sortVertices();
		body->removeWrongVertices();

		if (_projectAll || body->invalid()) {
			// FIXME review: maybe calcAcc should be done for all bodies?
			body->calculateAccuracy(view());
			// FIXME maybe this calculation can be moved to the scan segments
			// and remove the notref variable from the VBody!
			if (!body->body()->zones.empty()) {
				bool notused = true;
				CBody* cbody = engine()->getBody(body);

				// Check if all referring zones are not displayed
				for (int iz = 0; iz < (int)cbody->zones.size(); iz++) {
					if (!(cbody->zones[iz])->ignore()) {
						notused = false;
						break;
					}
				}
				if (notused)
					body->notReferenced(true);
			}
		}
	}
	IFSTAT(kernel.stat.stop("04 body_cleanup"));
	IFSTAT(kernel.stat.start("05 process_segments"));

	// Scan segments: loop through all the segments in bodies and assign them an in and out region
	viewer.pState = GeometryViewer::Projection::SCAN;
	kernel.clearErrorHash();
#ifdef THREAD
	if (kernel.nthreads()) {
		if (kernel.nBodies()) {
			feeder.reset(&D2Layer::_scanBodySegments, avgPercIntersect,100);
			kernel.threadpool.execute(&feeder);
		}
	} else
#endif
	{
		for (size_t ib=0; ib<kernel.nBodies(); ib++) {
			kernel.scanBodySegments(kernel.getBody(ib), engine(), _projectAll);
			// Update information done up to now
			viewer.pState = GeometryViewer::Projection::SCAN;
			if (kernel.nBodies())
				viewer.percent(avgPercIntersect
						+ ((100-avgPercIntersect)*ib)/kernel.nBodies());
			else
				viewer.percent(99);
			if (stop()) break;
		}
	}

	// finished: do book keeping
	timer.stop();
	double timeTotal = timer.userTime();
	viewer.state(GeometryViewer::Projection::FINISHED, 100);

	if (timeTotal>0.0) {
		for (int i=percMax-1; i>0; i--)
			percIntersect[i] = percIntersect[i-1];
		percIntersect[0] = (int)(100*timeIntersect / timeTotal);
	}

	IFSTAT(kernel.stat.stop("05 process_segments"));
#ifdef STAT
	cout << "project: all= " << _projectAll << endl;
	cout << "\t# Bodies:  " << kernel.nBodies()  << endl;
	cout << "\t# Regions: " << kernel.nRegions()  << endl;
	cout << "Kernel Statistics\n" << kernel.stat << endl;
	kernel.mergeStatistics();
	cout << "Engine Statistics\n" << engine()->stat << endl;
#endif
	DUMP(cout << endl << "*-* D2Layer::endproject('"<< viewer.title() <<"')" << endl);
	kernel.unlock();
} // project

/** scanBodySegments called from the threadpool */
void D2Layer::_scanBodySegments(VBody* body, GeometryEngine* eng)
{
	kernel.scanBodySegments(body, eng, _projectAll);
} // _scanBodySegments

/** find closest vertex
 * @param u,v	location to search
 * @param d	minimum distance
 * @param vu, vv	return vertex u,v location
 * @return true	if vertex is found
 * */
bool D2Layer::closestVertex(const double u, const double v, const double d,
				double* vu, double* vv)
{
	// Can only check if the projection exists
	if (viewer.state()!=GeometryViewer::Projection::FINISHED and
	    viewer.state()!=GeometryViewer::Projection::DRAW)
		return false;

	double d2 = d*d;
	kernel.lock();	// <--- Locking Geometry?
	for (size_t ib=0; ib<kernel.nBodies(); ib++) {
		VBody* body = kernel.getBody(ib);
		if (!body->show() || body->show()&BIT_MOVE) continue;

		// Check body center on UVW system
		Point p = view().xyz2uvw(body->body()->position());
		if (Eq0(p.z,SMALL)) {
			if (Sq(p.x-u) + Sq(p.y-v) <= d2) {
				*vu = p.x;
				*vv = p.y;
				kernel.unlock();
				return true;
			}
		}

		// Check conic vertices
		for (int i=0; i < body->nC; i++) {
			if (body->V[i].size()<2) continue;

			auto& V = body->V[i];

			int kprev = -1;
			for (int k=0; k<(int)V.size(); k++) {
				Segment& vertex = V[k];
				if (vertex.body && vertex.body->show()) {
					if (Sq(vertex.x-u) + Sq(vertex.y-v) <= d2) {
						*vu = vertex.x;
						*vv = vertex.y;
						kernel.unlock();
						return true;
					}
					if (body == vertex.body) {
						if (kprev<0)
							kprev = k;
						else {
							double tm = 0.5 * (vertex.t + V[kprev].t);
							double xm, ym;
							body->C[i].getXY(tm,&xm,&ym);
							if (Sq(xm - u) + Sq(ym - v) <= d2) {
								*vu = xm;
								*vv = ym;
								kernel.unlock();
								return true;
							}
						}
					}
				}
			}
		}
	}
	kernel.unlock();
	return false;
} // closestVertex

/** project vertices along an axis an populate the array
 * it is used for scanning possible locations of errors in the geometry
 */
bool D2Layer::projectVertices(const char axis, std::vector<double>& vertices)
{
	// Can only check if the projection exists
	if (viewer.state()!=GeometryViewer::Projection::FINISHED and
	    viewer.state()!=GeometryViewer::Projection::DRAW)
		return false;

	kernel.lock();	// <--- Locking Geometry?
	for (size_t ib=0; ib<kernel.nBodies(); ib++) {
		VBody* body = kernel.getBody(ib);
		// Check conic vertices
		for (int i=0; i < body->nC; i++) {
			if (body->V[i].size()<2) continue;
			auto& V = body->V[i];
			for (int k=0; k<(int)V.size(); k++) {
				Segment& vertex = V[k];
				switch (axis) {
					case 'x':
					case 'X':
						vertices.push_back(view().uv2x(vertex.x,vertex.y));
						break;
					case 'y':
					case 'Y':
						vertices.push_back(view().uv2y(vertex.x,vertex.y));
						break;
					case 'z':
					case 'Z':
						vertices.push_back(view().uv2z(vertex.x,vertex.y));
						break;
					case 'u':
					case 'U':
						vertices.push_back(vertex.x);
						break;
					case 'v':
					case 'V':
						vertices.push_back(vertex.y);
						break;
				}
			}
		}
	}
	kernel.unlock();
	std::sort(vertices.begin(), vertices.end(), [](double a, double b) { return a<b; });
	return true;
} // projectVertices

/** drawSegment
 * @param body	body of which to draw the segment
 * @param ic	index of conic to draw
 * @param from	starting index in the vertex array
 * @param to	ending index (+1) in the vertex array
 * @param color	color to use for drawing
 * @param step	I/O step used in drawing (<0.0 to reset)
 * @return true is segment is visible (inside clipping area)
 */
bool D2Layer::drawSegment(Painter& painter, VBody* body, const int ic,
			const int from,    const int to,
			const dword color, const int width,
			double* step)
{
	double x, y;
	int x1,y1, x2,y2;
	bool visible=false;
	const Conic& conic = body->C[ic];
	const auto& V = body->V[ic];
	const Segment* v = &V[from];
	const double ts = v->t;
	const double xs = v->x;
	const double ys = v->y;
	v = &V[to];
	const double te = v->t;
	const double xe = v->x;
	const double ye = v->y;

	DUMP(bool debug = body->name() == "256");
	if (te-ts < SMALL2) return false;
	double maxstep = (conic.type()==Conic::Type::PARABOLA)? (te-ts)/10.0 : MAXSTEP;

#if _DEBUG>2
	// Draw each segment
	for (int i=from+1; i<to-1; i++) {
		v = &V[i];
		painter.setPixel(view().u2i(v->x), view().v2j(v->y), 0xFFFFFF);
	}
	// Draw mid point
	double tm = 0.5*(ts+te);
	double xm,ym;
	conic.getXY(tm,&xm,&ym);
	x1 = view().u2i(xm);
	y1 = view().v2j(ym);
	painter.setPixel(x1,  y1,  0xFFFFFF);
	painter.setPixel(x1-1,y1-1,0xFFFFFF);
	painter.setPixel(x1+1,y1-1,0xFFFFFF);
	painter.setPixel(x1-1,y1+1,0xFFFFFF);
	painter.setPixel(x1+1,y1+1,0xFFFFFF);
#endif

	if (ISZERO(Sq(xe-xs) + Sq(ye-ys)) && ISEQ(ts,te))
		return false;

	// draw lines
	x1 = view().u2i(xs);
	y1 = view().v2j(ys);
	if (conic.type() == Conic::Type::LINE) {
		x2 = view().u2i(xe);
		y2 = view().v2j(ye);
		if (!width)
			visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
		else
			visible |= painter.lineThick(x1,y1,x2,y2,width,color);
		DUMP(if (debug)
			printf("drawSegment %s [1] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
	} else {
		// initial guess
		if (*step<=0.0) *step = maxstep;
		if (*step > te-ts) *step = te-ts;
		double t1 = ts;
		do {
			double t2 = t1 + *step;
			if (t2>te) {
				t2 = te;
				*step = te - ts;
			}
			conic.getXY(t2, &x, &y);
			x2 = view().u2i(x);
			y2 = view().v2j(y);

			// find at half way what is the distance from a line
			while (1) {
				if (Abs(x2-x1)<=2 && Abs(y2-y1)<=2) {
					if (!width)
						visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
					else
						visible |= painter.lineThick(x1,y1,x2,y2,width,color);
					DUMP(if (debug)
						printf("drawSegment %s [2] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
#if _DEBUG>1
					painter.setPixel(x1,y1,0xFFFFFF);
#endif
					x1 = x2;
					y1 = y2;
					t1 = t2;
					*step *= 2.0;
					if (*step > maxstep) *step = maxstep;
					break;
				}

				double xh, yh;
				double t3 = t1 + *step/2.0;
				assert(t3<=te);
				conic.getXY(t3, &xh, &yh);
				int x3 = view().u2i(xh);
				int y3 = view().v2j(yh);

				// distance^2 of line 1,2
				double l2 = Sq((double)(x2-x1)) + Sq((double)(y2-y1));
				// perpendicular distance of midpoint 3 wrt to line 1,2
				double d2 = Sq((double)(x3-x1)*(double)(y2-y1)
					      - (double)(y3-y1)*(double)(x2-x1))
					      / (double)l2;
				if (d2<=0.01 && l2<MAXLEN2) {
					if (!width)
						visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
					else
						visible |= painter.lineThick(x1,y1,x2,y2,width,color);
					DUMP(if (debug)
						printf("drawSegment %s [3] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
#if _DEBUG>1
					painter.setPixel(x1,y1,0xFFFFFF);
#endif
					x1 = x2;
					y1 = y2;
					t1 = t2;
					*step *= 2.0;
					if (*step > maxstep) *step = maxstep;
					break;
				} else
				if (d2<=0.1 || l2<MAXLEN2) {
					if (!width)
						visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
					else
						visible |= painter.lineThick(x1,y1,x2,y2,width,color);
					DUMP(if (debug)
						printf("drawSegment %s [4] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
#if _DEBUG>1
					painter.setPixel(x1,y1,0xFFFFFF);
#endif
					x1 = x2;
					y1 = y2;
					t1 = t2;
					break;
				} else
				if (Abs(x3-x1)<=2 && Abs(y3-y1)<=2) {
					if (!width)
						visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
					else
						visible |= painter.lineThick(x1,y1,x2,y2,width,color);
					DUMP(if (debug)
						printf("drawSegment %s [5] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
#if _DEBUG>1
					painter.setPixel(x1,y1,0xFFFFFF);
#endif
					x1 = x3;
					y1 = y3;
					t1 = t3;
					*step *= 2.0;
					break;
				} else {
					x2 = x3;	// half
					y2 = y3;
					t2 = t3;
					*step /= 2.0;
				}
			}
		} while (t1 < te - *step);
		x2 = view().u2i(xe);
		y2 = view().v2j(ye);
		if (!width)
			visible |= painter.lineAntialias(x1,y1,x2,y2,color,true);
		else
			visible |= painter.lineThick(x1,y1,x2,y2,width,color);
		DUMP(if (debug)
			printf("drawSegment %s [6] (%d,%d) .. (%d,%d)\n",body->cname(),x1,y1,x2,y2));
	}
#if _DEBUG>1
	x1 = view().u2i(xs);
	y1 = view().v2j(ys);
	painter.setPixel(x1,y1, 0xFFFF00);
	painter.setPixel(x1-1,y1-1, 0xFFFF00);
	painter.setPixel(x1+1,y1-1, 0xFFFF00);
	painter.setPixel(x1-1,y1+1, 0xFFFF00);
	painter.setPixel(x1+1,y1+1, 0xFFFF00);

	painter.setPixel(x2,y2, 0xFFFF00);
	painter.setPixel(x2-1,y2-1, 0xFFFF00);
	painter.setPixel(x2+1,y2-1, 0xFFFF00);
	painter.setPixel(x2-1,y2+1, 0xFFFF00);
	painter.setPixel(x2+1,y2+1, 0xFFFF00);
#endif
	return visible;
} // drawSegment

/** draw geometry to view */
void D2Layer::drawSegments(Painter& painter, VBody* body, dword mask)
{
	dword maskZone;
	maskZone = mask | Segment::Type::ZONE;

	for (int i=0; i < body->nC; i++) {
		auto& V = body->V[i];
		if (V.size()<2) continue;

		// Initial vertex
		int kp = 0;
		double step = -1.0;

		// Next vertex
		Segment* v = &V[1];
		dword type;
		// WARNING the order is important to avoid segfault when v->in==NULL
		bool zone = mask!=Segment::Type::BODY &&
		            v->type&Segment::Type::ZONE &&
			    v->zone->region()->show()&BIT_SELECT;
		if (zone)
			type = v->type & maskZone;
		else
			type = v->type & mask;

		for (int k=2; k<(int)V.size(); k++) {
			// Check next vertex
			v = &V[k];

			if (zone == (mask!=Segment::Type::BODY &&
			    v->type&Segment::Type::ZONE &&
			    v->zone->region()->show()&BIT_SELECT) &&
			    (v->type&mask)==type) {
#if _DEBUG>1
				int x = view().u2i(v->x);
				int y = view().v2j(v->y);
				painter.setPixel(x,y,0x00FFFF);
#endif
				continue;
			}

			if (zone || (type&mask)) {
				dword color;
				int   width=0;

				if (mask==Segment::Type::BODY) {
					color = body->color();
					width = body->lineWidth();
				} else
				if (type&Segment::Type::ERROR)
					color = viewer.showErrors?
							geometry.errorColor : geometry.regionColor;
				else
				if (type&Segment::Type::ZONE)
					color = geometry.zoneColor;
				else
				if (type&Segment::Type::REGION)
					color = geometry.regionColor;
				else {
					color = body->color();
					width = body->lineWidth();
				}

				if (drawSegment(painter, body, i, kp, k-1,
						color|FLAG_FINAL, width, &step))
					body->visible = true;
			}
#if _DEBUG>1
			else {
				int x1 = view().u2i(V[kp].x);
				int y1 = view().v2j(V[kp].y);
				painter.setPixel(x1,  y1,  0x00FFFF);
				painter.setPixel(x1-1,y1-1,0x00FFFF);
				painter.setPixel(x1+1,y1-1,0x00FFFF);
				painter.setPixel(x1-1,y1+1,0x00FFFF);
				painter.setPixel(x1+1,y1+1,0x00FFFF);
				x1 = view().u2i(v->x);
				y1 = view().v2j(v->y);
				painter.setPixel(x1,  y1,  0x00FFFF);
				painter.setPixel(x1-1,y1-1,0x00FFFF);
				painter.setPixel(x1+1,y1-1,0x00FFFF);
				painter.setPixel(x1-1,y1+1,0x00FFFF);
				painter.setPixel(x1+1,y1+1,0x00FFFF);
			}
#endif

			kp = k-1;
			zone = mask!=Segment::Type::BODY &&
				v->type&Segment::Type::ZONE &&
				v->zone->region()->show()&BIT_SELECT;
			if (zone)
				type = v->type & maskZone;
			else
				type = v->type & mask;
		}

		if (zone || (type&mask)) {
			dword color;
			int   width = 0;

			if (mask==Segment::Type::BODY) {
				color = body->color();
				width = body->lineWidth();
			} else
			if (type&Segment::Type::ERROR)
				color = viewer.showErrors?
						geometry.errorColor : geometry.regionColor;
			else
			if (type&Segment::Type::ZONE)
				color = geometry.zoneColor;
			else
			if (type&Segment::Type::REGION)
				color = geometry.regionColor;
			else {
				color = body->color();
				width = body->lineWidth();
			}

			if (drawSegment(painter, body, i, kp, V.size()-1,
					color|FLAG_FINAL, width, &step))
				body->visible = true;
		}
#if _DEBUG>1
		else {
			int x1 = view().u2i(V[kp].x);
			int y1 = view().v2j(V[kp].y);
			painter.setPixel(x1,y1,0x00FFFF);
			painter.setPixel(x1-1,y1-1,0x00FFFF);
			painter.setPixel(x1+1,y1-1,0x00FFFF);
			painter.setPixel(x1-1,y1+1,0x00FFFF);
			painter.setPixel(x1+1,y1+1,0x00FFFF);
			x1 = view().u2i(v->x);
			y1 = view().v2j(v->y);
			painter.setPixel(x1,y1,0x00FFFF);
			painter.setPixel(x1-1,y1-1,0x00FFFF);
			painter.setPixel(x1+1,y1-1,0x00FFFF);
			painter.setPixel(x1-1,y1+1,0x00FFFF);
			painter.setPixel(x1+1,y1+1,0x00FFFF);
		}
#endif
	}
} // drawSegments

/** drawNodes */
void D2Layer::drawNodes(Painter& painter, VBody* body)
{
	const GBody* gbody = body->body();

	for (int n=0; n<gbody->nodes(); n++) {
		double u, v, w;
		view().xyz2uvw(gbody->node(n), &u, &v, &w);
		if (n==0 || Eq0(w,SMALL3)) {
			int x = view().u2i(u);
			int y = view().v2j(v);
			if (painter.rectangle(x-2,y-2,x+2,y+2, geometry.selectColor))
				body->visible = true;
		}
	}
} // drawNodes

/** draw vertices */
void D2Layer::drawVertices(Painter& painter)
{
	for (size_t ib=0; ib<kernel.nBodies(); ib++) {
		if (stop()) break;
		VBody* body = kernel.getBody(ib);
		if (not (body->show()&BIT_SELFREEZE)) continue;

		for (int i=0; i < body->nC; i++) {
			if (body->V[i].size()<2) continue;
			auto& V = body->V[i];

			int kprev = -1;
			for (int k=0; k<(int)V.size(); k++) {
				Segment& v = V[k];
				if (v.body && v.body->show()&BIT_SELFREEZE) {
					// XXX shows start/end of circles!!!
					int x = view().u2i(v.x);
					int y = view().v2j(v.y);
					painter.setPixel(x,y,  geometry.vertexColor);
					painter.setPixel(x-1,y,geometry.vertexColor);
					painter.setPixel(x+1,y,geometry.vertexColor);
					painter.setPixel(x,y-1,geometry.vertexColor);
					painter.setPixel(x,y+1,geometry.vertexColor);

					// Draw also middle of segment (between two visible vertices!
					// one pixel is enough
					if (body == v.body) {
						if (kprev<0)
							kprev = k;
						else {
							double tm = 0.5 * (v.t + V[kprev].t);
							double xm, ym;
							body->C[i].getXY(tm,&xm,&ym);
							x = view().u2i(xm);
							y = view().v2j(ym);
							painter.setPixel(x,y,geometry.vertexColor);
						}
					}
				}
			}
		}
	}
} // drawVertices

/** draw regions */
void D2Layer::drawRegions(Painter& painter)
{
	int W = painter.width();
	int widthadd = W - (painter.clip().right - painter.clip().left)-1;
	dword* pixelPtr = painter.pixelPtr(painter.clip().left, painter.clip().top);

	double dx = -view().matrix(0,2);
	double dy = -view().matrix(1,2);
	double dz = -view().matrix(2,2);

	geometry.lockRead();	// No need to lock the kernel, the segments are there
	for (int j=painter.clip().top; j<=painter.clip().bottom; j++) {
		if (stop()) break;
		//assert(pixelPtr == painter.pixelPtr(painter._clip.left, j));
		for (int i=painter.clip().left; i<=painter.clip().right; i++, pixelPtr++) {
			// Find a pixel that has at least on background pixel
			// in every direction
			//     b b b
			//     b + b
			//     b b b
			// we start with the up-left corner then check for i+1, j+1
			// 2x2 sometimes it fails
			if (*pixelPtr       != painter.background() ||
			    pixelPtr[1]     != painter.background() ||
			    pixelPtr[2]     != painter.background() ||
			    pixelPtr[W]     != painter.background() ||
			    pixelPtr[W+1]   != painter.background() ||
			    pixelPtr[W+2]   != painter.background() ||
			    pixelPtr[2*W]   != painter.background() ||
			    pixelPtr[2*W+1] != painter.background() ||
			    pixelPtr[2*W+2] != painter.background()) continue;
			double u = view().ic2u(i+1);
			double v = view().jc2v(j+1);

			double x = view().uv2x(u,v);
			double y = view().uv2y(u,v);
			double z = view().uv2z(u,v);

			engine()->incBodyCheckId();

			geometry.lockEdit();
			// Draw selected zones
			if (geometry.editRegion.size()) {
				GZone* zone = geometry.editRegion.inside(x,y,z, dx,dy,dz);
				if (zone) {
					painter.fill(i, j,
						0xFFFFFF,
						(((const GZone*)zone==geometry.editRegion.zones().back())?
							geometry.zoneColor : geometry.visibleColor),
						Painter::Fill::X);
					geometry.unlockEdit();
					continue;
				}
			}
			geometry.unlockEdit();

			VZone* zone = engine()->where2D(x,y,z, dx,dy,dz);
			if (zone) {
				const VRegion* region = zone->region();
				if (region->show() & BIT_SELECT) {
					Color32 col(region->color() | FLAG_REGION);
					col.red()   = col.red()  >0x7F? 0:0xFF;
					col.green() = col.green()>0x7F? 0:0xFF;
					col.blue()  = col.blue() >0x7F? 0:0xFF;
					painter.fill(i, j, region->color() | FLAG_REGION,
						col,
						//(region->color()^geometry.selectColor),
						(Painter::Fill)(region->show()&BIT_SELECT));
				} else
				if (region->type() == GRegion::Type::LATTICE)
					painter.fill(i, j, geometry.latticeColor,
						viewer.lattice.hashColor, Painter::Fill::HASHR);
				else
				if (region->type() == GRegion::Type::VOXEL)
					painter.fill(i, j, geometry.voxelColor,
						viewer.voxel.hashColor, Painter::Fill::HASHR);
				else
				if (viewer.d3.show && region->transparent())
					painter.fill(i, j, geometry.transparentColor());
				else
				if (geometry.lighterLevel)
					painter.fill(i,j,
						Color32::lighter(region->color() | FLAG_REGION,
						geometry.lighterLevel));
				else
					painter.fill(i,j, region->color() | FLAG_REGION);
			} else
			if (viewer.showErrors)
				painter.fill(i,j,
					(viewer.d3.show?
						geometry.transparentColor():
						0xFFFFFF | FLAG_REGION),
					geometry.regionErrorColor,
					Painter::Fill::HASH);
			else
				painter.fill(i,j,
					viewer.d3.show?
						geometry.transparentColor():
						0xFFFFFF | FLAG_REGION);
		}
		pixelPtr += widthadd;
	}
	geometry.unlockRead();
} // drawRegions

/** late draw regions, fill any empty pixel which still has the background color */
void D2Layer::drawRegionsLate(Painter& painter)
{
	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...
	dword* ptr = painter.data();	// data pointer

	double dx = -view().matrix(0,2);
	double dy = -view().matrix(1,2);
	double dz = -view().matrix(2,2);

	geometry.lockRead();
	for (int j=0; j<H; j++) {
		if (stop()) break;
		double v = view().jc2v(j);
		VZone* zone = NULL;
		for (int i=0; i<W; i++, ptr++) {
			if (*ptr != painter.background()) continue;
			double u = view().ic2u(i);
			double x = view().uv2x(u,v);
			double y = view().uv2y(u,v);
			double z = view().uv2z(u,v);
			engine()->incBodyCheckId();
			zone = engine()->where2D(x,y,z, dx,dy,dz, zone);
			// FIXME: maybe check for 3D, lattice or voxel
			if (zone) *ptr = zone->region()->color() | FLAG_REGION;
		}
	}
	geometry.unlockRead();
} // drawRegionsLate

/** flood fill scanning for label
 * @param x,y	I/O	input staring pixel to scan, output average pixel
 * @param pos	ref	return list pos points belonging to this label
 */
void D2Layer::scanLabel(Painter& painter, int* x, int* y, std::vector<IPoint>& pos)
{
#define STACKSIZE 500
	int tos;
	int stackX[STACKSIZE];
	int stackY[STACKSIZE];

	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...

	pos.clear();
	dword startColor = painter(*x, *y);
	long long sx=0, sy=0;

	stackX[0] = *x;
	stackY[0] = *y;
	tos = 1;
	while (tos) {
		int xx = stackX[--tos];
		int yy = stackY[  tos];

		dword* pix = painter.pixelPtr(xx,yy);		// get pointer
		if (*pix != startColor) continue;

		// Move to the left side
		while (xx>0 && pix[-1]==startColor) {
			xx--;
			pix--;
		}

		/* scan lines */
		dword* pixelup = (yy>0)? pix-W : NULL;
		dword* pixeldown = (yy<H-1)? pix+W: NULL;

		int up   = 0;
		int down = 0;
		while (*pix == startColor) {
			*pix |= FLAG_FINAL;
			sx += xx;
			sy += yy;
			pos.push_back(IPoint(xx,yy));

			int pup = (pixelup && *pixelup==startColor);
			if (!up && pup && tos<STACKSIZE) {
				stackX[tos  ] = xx;
				stackY[tos++] = yy-1;
			}
			up = pup;

			int pdown = (pixeldown && *pixeldown==startColor);
			if (!down && pdown && tos<STACKSIZE) {
				stackX[tos  ] = xx;
				stackY[tos++] = yy+1;
			}
			down = pdown;

			if (++xx>=W) break;

			pix++;
			if (pixelup) pixelup++;
			if (pixeldown) pixeldown++;
		}
	}
	assert(pos.size()>0);
	// Find center
	*x = Round((double)sx / double(pos.size()));
	*y = Round((double)sy / double(pos.size()));
} // scanLabel

/** drawLabel if it fits! */
bool D2Layer::drawLabel(Painter& painter, int i, int j, dword color)
{
	const int H = painter.height();	// save variables that might change
	const int W = painter.width();	// during the scanning...

	if (j<=viewer.font.height()*painter.dpiscale() or
	    j>H-viewer.font.height()*painter.dpiscale()) return false;

	// Scan text area if a possible label fits (4 characters wide minimum)
	int dx = viewer.font.width() * 2 + 1;
	if (i<=dx || i>W-dx) return false;

	for (int r=Max(0,j-viewer.font.height()*painter.dpiscale()); r<Min(j+3,H); r++)
		for (int c=Max(0,i-dx); c<Min(i+dx,W); c++)
			if (painter(c,r) != color)
				return false;

	double u = view().ic2u(i);
	double v = view().jc2v(j);

	double x = view().uv2x(u,v);
	double y = view().uv2y(u,v);
	double z = view().uv2z(u,v);

	engine()->incBodyCheckId();
	VZone* zone = engine()->where2D(x,y,z,
				-view().matrix(0,2),
				-view().matrix(1,2),
				-view().matrix(2,2));

	if (zone) {
		painter.pixel(i,j, FLAG_FINAL | 0xFFFFFF);
		const char* name = zone->region()->label;
		i -= painter.measure(viewer.font, name) / 2;
		painter.printf(viewer.font, i, j-viewer.font.height()*painter.dpiscale()-2,
					geometry.labelColor | FLAG_FINAL,
					name);
	}

	return true;
} // drawLabel

/** drawLabels */
void D2Layer::drawLabels(Painter& painter)
{
	srand(314159);

	int H = painter.height();	// save variables that might change
	int W = painter.width();	// during the scanning...
	dword* data = painter.data();	// data pointer
	dword* ptr = data;

	std::vector<IPoint> pos(1024);

	minLabelSize = 10*viewer.font.width();

	kernel.lock();
	// Check previous labels
	for (int i=(int)labels.size(); i--; ) {
		int x = view().u2i(labels[i].x);
		int y = view().v2j(labels[i].y);
		int xx=x, yy=y;
		if (x>=0 && x<W && y>=0 && y<H)
			if ((painter(x,y)&FLAG_INFOMASK) == FLAG_REGION) {
				scanLabel(painter, &x, &y, pos);
				if ((int)pos.size() >= minLabelSize*viewer.font.height()*painter.dpiscale() and
				    drawLabel(painter, xx, yy, painter(xx,yy)))
					continue;
			}
		labels.erase(labels.begin()+i);
	}

	// Scan whole image
	for (int j=0; j<H; j++) {
		if (stop()) break;
		for (int i=0; i<W; i++, ptr++) {
			if ((*ptr&FLAG_INFOMASK) != FLAG_REGION) continue;
			int x=i, y=j;

			scanLabel(painter, &x, &y, pos);

			// Ignore small regions
			if ((int)pos.size() < minLabelSize*viewer.font.height()*painter.dpiscale()) continue;

			if (drawLabel(painter, x, y, *ptr)) {
				Vector2D v;
				v.x = view().ic2u(x),
				v.y = view().jc2v(y);
				labels.push_back(v);
				continue;
			}

			for (int k=0; k<20; k++) {
				int m = rand() % pos.size();
				if (drawLabel(painter, pos[m].x, pos[m].y, *ptr)) {
					Vector2D v;
					v.x = view().ic2u(pos[m].x),
					v.y = view().jc2v(pos[m].y);
					labels.push_back(v);
					break;
				}
			}
		}
	}
	kernel.unlock();
} // drawLabels

/** drawEdge */
void D2Layer::drawEdge(Painter& painter)
{
//#ifdef EXPERIMENTAL
//#if 1
	/* Edge detection if requested */
	// XXX needs speed-up!!!
	int H = painter.height() - 1;	// save variables that might change
	int W = painter.width()  - 1;	// during the scanning...

	for (int j=1; j<H; j++) {
		int jm1 = j-1;
		int jp1 = j+1;
		for (int i=1; i<W; i++) {
			Color32 pixel(painter(i,j));
			if (pixel.alpha() != 0x01) continue;
			int im1 = i-1;
			int ip1 = i+1;

			// Construct the following pixels
			//    +----+----+----+
			//    | mm | 0m | pm |
			//    +----+----+----+
			//    | m0 | ij | p0 |
			//    +----+----+----+
			//    | mp | 0p | pp |
			//    +----+----+----+
			Color32 pmm(painter(im1, jm1));
			if (pmm.alpha() == 0) continue;
			Color32 p0m(painter(i  , jm1));
			if (p0m.alpha() == 0) continue;
			Color32 ppm(painter(ip1, jm1));
			if (ppm.alpha() == 0) continue;

			Color32 pm0(painter(im1, j  ));
			if (pm0.alpha() == 0) continue;
			Color32 pp0(painter(ip1, j  ));
			if (pp0.alpha() == 0) continue;

			Color32 pmp(painter(im1, jp1));
			if (pmp.alpha() == 0) continue;
			Color32 p0p(painter(i  , jp1));
			if (p0p.alpha() == 0) continue;
			Color32 ppp(painter(ip1, jp1));
			if (ppp.alpha() == 0) continue;

			//if (p0m&0x02000000 || pm0&0x02000000)
			//	continue;

			// Find the Sobel gradient in the color space
			int gxR = ( pmm.red()   + 2*p0m.red()    + ppm.red())
			        - ( pmp.red()   + 2*p0p.red()    + ppp.red());
			int gxG = ( pmm.green() + 2*p0m.green()  + ppm.green())
			        - ( pmp.green() + 2*p0p.green()  + ppp.green());
			int gxB = ( pmm.blue()  + 2*p0m.blue()   + ppm.blue())
			        - ( pmp.blue()  + 2*p0p.blue()   + ppp.blue());

			int gyR = (  ppm.red()   + 2*pp0.red()   + ppp.red())
			        - (  pmm.red()   + 2*pm0.red()   + pmp.red());
			int gyG = (  ppm.green() + 2*pp0.green() + ppp.green())
			        - (  pmm.green() + 2*pm0.green() + pmp.green());
			int gyB = (  ppm.blue()  + 2*pp0.blue()  + ppp.blue())
			        - (  pmm.blue()  + 2*pm0.blue()  + pmp.blue());

			int g = (int)isqrt(gxR*gxR + gxG*gxG + gxB*gxB
			                 + gyR*gyR + gyG*gyG + gyB*gyB);

#if 0
			int gx = (  Intensity(painter.pixel(im1, jm1)) +
			          2*Intensity(painter.pixel(i,   jm1)) +
			            Intensity(painter.pixel(ip1, jm1)))
			       - (  Intensity(painter.pixel(im1, jp1)) +
			          2*Intensity(painter.pixel(i,   jp1)) +
			            Intensity(painter.pixel(ip1, jp1)));

			int gy = (  Intensity(painter.pixel(ip1, jm1)) +
			          2*Intensity(painter.pixel(ip1, j  )) +
			            Intensity(painter.pixel(ip1, jp1)))
			       - (  Intensity(painter.pixel(im1, jm1)) +
			          2*Intensity(painter.pixel(im1, j  )) +
			            Intensity(painter.pixel(im1, jp1)));
			int g = (int)isqrt(gx*gx + gy*gy);
#endif

			if (g>200) {
				pixel.alpha(0x02);
				painter.pixel(i,j,pixel);
			}
		}
	}

	for (int j=1; j<H; j++)
		for (int i=1; i<W; i++)
			if (painter.pixel(i,j)&0x02000000)
				painter.pixel(i,j,geometry.zoneColor);
//#endif
//#endif
} // drawEdge

/** draw bodies boundaries */
void D2Layer::draw(Painter& painter)
{
	geometry.lockRead();	// No need to lock the kernel, since the segments/nodes are there
				// only the color is not known

	// 1st pass
	for (auto body : kernel.bodies)
		if (body->show()==0)
			drawSegments(painter, body, Segment::Type::REGION|Segment::Type::ERROR);

	// 2nd pass
	// if exists selected bodies make a second pass to draw them
	if (kernel.bodyShow()) {
		for (auto body : kernel.bodies) {
			if (body->show()==0) continue;
			if (body->show() & BIT_SELFREEZE) {
				drawSegments(painter, body, Segment::Type::BODY);
				if (body->show() & BIT_SELECT)
					drawNodes(painter, body);
			} else
				//drawSegments(painter, body,
				//	Segment::Type::BODY|Segment::Type::REGION|Segment::Type::ERROR);
				drawSegments(painter, body, Segment::Type::REGION|Segment::Type::ERROR);
		}
	}

	geometry.unlockRead();
} // draw

/** post-draw effects*/
void D2Layer::postDraw(Painter& painter)
{
	if (stop()) return;
	painter.correctAntialias();
	//painter.gaussianBlur();
} // postDraw
